﻿"""
TBC Macro to compute line that is defined by adjusting area to specified value
"""
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, UInt32, Math, Byte, String, StringSplitOptions
from System.Text import Encoding
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Image, ImageConverter, Imaging
from System.IO import StreamReader, MemoryStream, Stream, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary
from System.Windows.Controls import StackPanel, ComboBoxItem
from System.Windows.Input import Keyboard

from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar, Point as PointHelper
from Trimble.Vce.Interfaces.Units import AreaType
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox
from Trimble.Vce.Alignment.Linestring import Linestring, Polygon
from Trimble.Vce.Geometry import Point3D, Side, Vector3D, Line2D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import RegionBuilder, Region, Edge

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler, ViewHelper
clr.AddReference("Trimble.Vce.GraphicsEngine2D")
from Trimble.Vce.GraphicsEngine import OverlayBag
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice, UIEvents

try:
    from Trimble.Sdk.UI import InputMethod, MousePosition, CursorStyle    
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod, MousePosition, CursorStyle

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "AdjustArea"
    cmdData.CommandName = "AdjustArea"
    cmdData.Caption = "_Adjust Area"
    cmdData.UIForm = "AdjustArea"
    try:
        cmdData.Version = 1.54
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2019/03/18/tml-adjust-area"
        cmdData.HelpFile=cmdData.MacroInfo
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Adjust Area"
        cmdData.ToolTipTitle = "Adjust Area"
        cmdData.ToolTipTextFormatted = "Create a line (or Polygon) by adjusting a line along a reference line until a specified area is computed."
        cmdData.InsertAfterMenuToolKey = "" # put cmd at start of group
        cmdData.DefaultRibbonToolSize = 3 # make button "Large"
    except:
        pass

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + "Light" + ".png")
        cmdData.ImageLight = b
    except:
        pass
    if False:
        try:
            b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
            cmdData.ImageSmall = b
        except:
            pass
    else:
        # create image from string (so no external file)
        imageAsString = """
        137 80 78 71 13 10 26 10 0 0 0 13 73 72 68 82 0 0 0 38 0 0 0 38 8 6 0 0 0 168 61 
    233 174 0 0 0 4 103 65 77 65 0 0 177 143 11 252 97 5 0 0 0 9 112 72 89 115 0 0 14 
    195 0 0 14 195 1 199 111 168 100 0 0 3 224 73 68 65 84 88 71 197 152 59 104 20 65 
    24 199 55 185 152 139 137 103 244 48 38 26 141 57 47 241 117 38 185 203 67 240 65 
    124 225 3 81 4 43 11 181 176 176 178 211 78 176 17 173 44 197 198 66 176 73 171 22 
    150 134 40 216 40 8 86 62 58 69 16 2 73 80 20 21 244 252 255 135 153 101 230 187 217 
    189 189 7 56 240 203 100 190 253 230 251 126 183 217 219 205 93 176 107 234 120 48 
    53 121 44 68 174 27 137 53 82 171 169 197 170 213 58 249 108 230 47 177 99 196 87 
    203 87 236 8 230 251 54 136 57 235 164 49 185 182 196 156 56 242 166 49 59 30 62 177
     211 152 185 57 4 177 114 61 49 185 62 57 7 177 185 25 223 190 19 152 29 15 37 38 24
     3 95 64 25 188 7 179 205 194 156 49 189 126 11 216 99 30 148 128 227 161 126 216 166
     88 247 99 126 5 248 106 110 233 152 147 147 52 198 117 185 92 14 49 98 118 140 32 47
     111 239 35 190 98 237 152 31 2 138 61 1 109 50 135 36 137 73 129 56 236 125 36 24 217
     185 207 9 232 226 60 83 20 123 13 250 255 139 88 79 207 70 39 160 139 159 5 20 251 12
     38 124 18 227 165 195 65 177 120 80 81 42 30 82 177 166 139 177 160 32 15 120 97 254
     6 167 117 204 33 147 201 6 102 172 95 183 185 226 56 145 205 163 174 49 34 247 6 221
     43 215 168 87 111 76 25 196 188 2 124 4 60 107 215 117 204 33 147 89 173 164 178 217
     190 96 114 226 168 138 137 188 78 217 60 78 204 218 167 80 163 80 216 19 6 116 241 14
     240 20 80 236 177 79 108 40 95 12 82 169 101 65 97 71 197 94 133 175 121 148 152 172 63
     48 176 61 82 172 21 220 6 20 123 231 19 227 155 102 96 227 54 39 102 242 100 99 67 82 49
     94 94 106 12 14 22 194 160 149 116 17 252 2 11 136 109 209 49 7 243 39 52 112 175 108 106
     83 179 88 87 87 183 58 40 152 6 139 224 59 56 165 99 177 200 134 146 56 49 27 71 204 126
     213 250 247 62 192 219 197 31 196 174 234 152 47 79 33 155 73 152 147 228 140 21 199 14 4
     203 151 175 136 21 35 47 1 159 109 252 47 128 79 4 111 158 108 20 69 18 49 94 239 225 104
     109 77 5 91 134 199 43 26 130 187 128 98 47 48 103 172 120 152 39 155 212 139 169 235 136
     113 228 114 35 170 153 224 50 224 141 246 43 232 209 49 7 95 147 122 48 245 188 98 246 153
     208 28 0 63 16 227 255 80 123 25 179 139 201 117 35 176 22 225 101 229 140 116 186 83 29 16
     98 91 193 7 45 118 201 87 176 22 146 92 99 21 98 45 45 173 234 128 16 203 130 231 90 236 158
     44 88 43 62 49 246 169 71 140 60 208 98 179 118 193 122 168 75 140 99 67 255 176 74 18 92 1
     124 3 188 181 11 214 131 20 179 122 40 134 134 74 234 4 85 140 222 222 77 42 193 216 107 142 
    32 198 51 182 4 246 218 141 106 69 138 153 30 166 103 46 183 83 155 136 17 33 198 219 196 55 
    204 124 160 95 176 27 213 138 20 51 84 21 227 141 118 116 100 90 138 117 99 227 27 204 20 187
     43 139 54 11 246 234 235 27 212 38 158 33 63 3 128 52 196 102 48 83 172 225 55 64 20 236 229
     189 190 204 160 24 79 173 224 26 224 195 252 147 175 40 49 185 190 99 73 224 222 88 177 85 171
     214 202 51 198 77 103 48 255 0 11 96 183 44 106 229 57 251 100 94 181 107 44 86 140 31 50 236
     226 4 27 11 152 23 193 79 112 94 199 124 121 222 181 17 144 98 118 222 120 233 16 196 90 180
     133 103 68 136 165 49 127 2 188 206 110 130 150 102 139 241 142 16 59 58 58 186 130 177 209 
    253 225 38 179 17 51 63 149 83 236 17 104 175 69 140 72 49 153 87 85 140 35 191 121 76 37 11
     110 0 62 1 150 192 28 168 248 210 164 10 243 252 166 135 98 144 241 162 219 71 15 138 33 209 
    121 69 152 207 113 179 1 49 62 13 156 194 50 38 215 250 107 168 112 45 209 237 163 7 79 43 18
     165 24 63 41 221 1 230 203 54 231 203 55 95 76 174 27 22 75 165 218 124 98 53 199 228 218 92
     99 118 140 240 222 153 104 80 140 69 155 13 165 124 241 74 177 32 248 7 92 179 11 34 20 169 
    175 50 0 0 0 0 73 69 78 68 174 66 96 130 
        """
        splitOptions = Array[String]([" ", "\n"]) # split so each number is a string
        asStrings = imageAsString.Split(splitOptions, StringSplitOptions.RemoveEmptyEntries)
        # build byte list
        byteList = []
        for b in asStrings:
            b = Byte.Parse(b)
            byteList.append(b)
    
        byteArray = Array[Byte](byteList)
        with MemoryStream(byteArray) as ms:
            cmdData.ImageSmall = Image.FromStream(ms);

class AdjustArea(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        if False: # read from file or text
            with StreamReader(macroFileFolder + r"\AdjustArea.xaml") as s:
                wpf.LoadComponent(self, s)
        else:
            # copy text from *.xaml file and place here
            t = """
<StackPanel  HorizontalAlignment="Stretch"  
    xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
    xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
    xmlns:Wpf="clr-namespace:Trimble.Vce.UI.Controls.Wpf;assembly=Trimble.Vce.UI.Controls">
    <Label Content="Name:" />
    <Wpf:NameEdit x:Name="nameCtl" />
    <Label Content="Layer:" />
    <Wpf:LayerPicker x:Name="layerCtl"/>
    <CheckBox x:Name="makePolygonCtl" Margin="0,5,0,0"  Content="Create &quot;Polygon&quot; instead of line" IsChecked="False" />
    <Label Content="Area:"/>
    <Grid  >
        <Grid.ColumnDefinitions>
            <ColumnDefinition Width="2*"/>
            <ColumnDefinition Width="1*"/>
        </Grid.ColumnDefinitions>
        <Wpf:NumericEdit x:Name="areaCtl" MinValue="0.001" InputType="Double" VerticalAlignment="Center"  Grid.Column="0"/>
        <ComboBox x:Name="areaType" Grid.Column="1" IsEditable="False" VerticalAlignment="Center" />
   </Grid>
    <Label Content="Boundary lines:"  />
    <Wpf:MemberSelection x:Name="objs" UseLocalSelection="True"  ProcessGlobalSelectionChanges="False"/>
    <Label Content="Reference line:"  />
    <Wpf:SelectEntity x:Name="refLine" AutoTab="True" ValueChanged="refLine_ValueChanged" ToolTip="Pick near approx line location"/>
    <Label Content="Approximate location:"/>
    <Wpf:StationEdit x:Name="coordCtl" AutoTab="False" ValueChanged="coordCtl_ValueChanged" ToolTip="Pick side of line to set direction"/>
    <Label Content="Results:" />
    <TextBox x:Name="textBoxEntityInfo" 
                 IsReadOnly="True" Height="90"
                 HorizontalScrollBarVisibility="Auto"  VerticalScrollBarVisibility="Auto"
                 TextAlignment="Left" Margin="3" IsTabStop="False"/>

</StackPanel>
"""
            with StreamReader(MemoryStream(Encoding.UTF8.GetBytes(t))) as s:
                wpf.LoadComponent(self, s)

        self.currentProject = currentProject
        self.macroName = "Adjust Area"
        self.tmlFileFolder = macroFileFolder  

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption
        types = Array[Type](SurfaceTypeLists.AllWithCutFillMap)
        self.uint1 = UInt32.Parse("1") # Python doesn't like unsigned ints so we create one this way...
        self.lType = clr.GetClrType(IPolyseg)
        self.aunits = self.currentProject.Units.Area
        self.lunits = self.currentProject.Units.Linear
        # we don't want the units to be included
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        isMeters = self.lunits.DisplayType == LinearType.Meter
        # get list of area units (first few are only for meters, last ones for feet
        for a in self.aunits.Units:
            au = a.Value
            if au.Type < AreaType.SquareInch:
                if not isMeters:
                    continue
            else:
                if isMeters:
                    break
            cbi = ComboBoxItem()
            cbi.Content = au.Description
            cbi.Tag = au.Type
            self.areaType.Items.Add(cbi)

        self.refLine.AddTypeFilter(clr.GetClrType(IPolyseg))
        self.newLine = 0
        settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
        l = settings.GetUInt32("AdjustArea.Layer", 0)
        if self.currentProject.Concordance.Lookup(l):
            self.layerCtl.SelectedSerialNumber = l
        self.makePolygonCtl.IsChecked = settings.GetBoolean("AdjustArea.MakePolygon", False)
        self.areaCtl.Value = settings.GetDouble("AdjustArea.TargetArea", Double.NaN)
        if isMeters:
            default = 0 # sq meters
            s = "AdjustArea.AreaTypeM"
        else:
            default = 1 # default is sq-ft
            s = "AdjustArea.AreaTypeF"
        self.areaType.SelectedIndex = settings.GetInt32(s, default) 

        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        if isinstance(o, self.lType):
            return True
        return False

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def Compute1Area(self, sta, coord, poly):
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        if not poly.FindPointFromStation(sta, seg, t, pcl, v):
            self.coordCtl.StatusMessage = "Could not project point to ref line"
            return (True, None, None, None, None)
        v = v.Value
        return self.ComputeFromVector(v, coord, poly, pcl.Value)

    def ComputeFromVector(self, v, coord, poly, pcl):
        pt1 = pcl + v * 10000.0
        pt2 = pcl - v * 10000.0
        newPoly = PolySeg()
        newPoly.Add((pt1, pt2))
        
        newPoly.UserDefinedAttributes.Add(self.uint1)
        newPoly.UserDefinedAttributes.Add(self.uint1)
        rb = RegionBuilder()
        rb.Add(newPoly)
        
        for o in self.objs.SelectedMembers(self.currentProject):
            if isinstance(o, self.lType):
                poly = o.ComputePolySeg()
                if not poly or poly.NumberOfNodes < 2:
                    continue
                rb.Add(poly)
        # add any line just created
        if self.newLine:
            l = self.currentProject.Concordance.Lookup(self.newLine)
            if l:
                poly = l.ComputePolySeg()
                if  poly and poly.NumberOfNodes > 1:
                    rb.Add(poly)

        nRegions = rb.Build()
        if nRegions == 0:
            return (True, None, None, None, None)

        foundRegions = []
        for r in rb.Regions():
            for edge in r.Edges():
                polyEdge = edge.Segment
                if edge.Reversed(r):
                    a = edge.LeftProperty
                else:
                    a = edge.RightProperty
                if a == self.uint1:  
                    # picked point must project to line segment
                    (found, pnt, sta2) = polyEdge.FindPointFromPoint(coord)
                    if found: 
                        (area,cg,side) = r.AreaPolyseg().Area()
                        (len2, p1, p2) = Line2D.Length2(cg, coord)
                        foundRegions.Add((len2, area,polyEdge, r))

        if foundRegions:
            if len(foundRegions) == 1:
                # only one found, use that one
                (len2, area,polyEdge, r) = foundRegions[0]
                return (False, area, polyEdge, v, r.AreaPolyseg())
   
            def getKey(item):
                return item[0]
            foundRegions.sort(key=getKey)
            (len2, area,polyEdge, r) = foundRegions[0]
            return (False, area, polyEdge, v, r.AreaPolyseg())
        """
            for (a,cg,p,r) in foundRegions:
                Line2D.Length2(cg, coord)
                (found, pnt, sta)=p.FindPointFromPoint(coord)
                if found:
                    pass
                # if starting point in polyseg, take this one
                if r.AreaPolyseg().PointInPolyseg(coord) == Side.Out:
                    continue
                # take this one
                return (False, a, p, v)

            # we didn't find one that enclosed the pick point, find nearest
            for (a,p,r) in foundRegions:
                seg = clr.Reference[Segment]()
                t = clr.Reference[float]()
                staRef = clr.Reference[float]()
                offset = clr.Reference[float]()
                pcl = clr.Reference[Point3D]()
                v2 = clr.Reference[Vector3D]()
                side = clr.Reference[Side]()
                rtn = p.FindPointFromPoint(coord, seg, t, pcl, staRef, v2,offset, side)
                if not rtn:
                    continue
                return (False, a, p, v)
"""
        return(True, None, None, None, None)

    def coordCtl_ValueChanged(self, ctrl, e):
        # set keyboard focus if change was due to mouse pick
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)
 
    def refLine_ValueChanged(self, sender, e):
        l = self.refLine.Entity
        self.coordCtl.StationProvider = l

    def Interpolate(self, x1, x2, y1, y2, y):
        r = (y2 - y1) / (x2 - x1)
        x = (y - y1) / r 
        x = x + x1
        return x

    def OkClicked(self, sender, e):
        area = self.areaCtl.Value
        if Double.IsNaN(area) or area <= 0:
            self.areaCtl.StatusMessage = "Enter valid area"
            return False
        self.areaCtl.StatusMessage = ""
        at = self.areaType.SelectedValue.Tag
        areaM2 = self.currentProject.Units.Area.Convert(at, area, AreaType.SquareMeter)

        line = self.refLine.SerialNumber
        if not self.IsValid(line):
            self.refLine.StatusMessage = "No reference line selected"
            return
        self.refLine.StatusMessage = ""
        line = self.refLine.Entity
        poly = line.ComputePolySeg()
        if not poly or poly.NumberOfNodes < 2:
            self.refLine.StatusMessage = "Line not defined"
            return
        coord = self.coordCtl.ClickLocation
        if not coord.Is3D and not coord.Is2D :
            self.coordCtl.StatusMessage = "No coordinate defined"
            return
        self.coordCtl.StatusMessage = ""

        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        staRef = clr.Reference[float]()
        offset = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        side = clr.Reference[Side]()
        rtn = poly.FindPointFromPoint(coord, seg, t, pcl, staRef, v,offset, side)
        if not rtn:
            self.coordCtl.StatusMessage = "Could not project point to ref line"
            return
        # check if PI point selected
        pickAtPI = False
        seg = seg.Value
        if t.Value == 0 and seg:
            node = seg.Node
            if node:
                prevNode = node.Previous
                if prevNode:
                    (found, vStart) = seg.Slope(0.0)
                    vStart.Normalize()
                    vStart.Rotate90(Side.Right)
                    segBack = PolySeg.CreateSegment(prevNode)
                    (found, vEnd) = segBack.Slope(1.0)
                    vEnd.Normalize()
                    vEnd.Rotate90(Side.Right)
                    pickAtPI = True

        sta = staRef.Value
        pcl = pcl.Value
        preSta = sta
        preArea = 0
        prevV = v.Value
        self.textBoxEntityInfo.Text = ""
        if pickAtPI:
            (notValid, areaMin, newPoly, v, polyArea) = self.ComputeFromVector(vStart, coord, poly, pcl)
            if notValid:
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute area")
                return
            (notValid, areaMax, newPoly, v, polyArea) = self.ComputeFromVector(vEnd, coord, poly, pcl)
            if notValid:
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute area")
                return
            if max(areaMin, areaMax) < areaM2 or min(areaMin, areaMax) > areaM2:
                s = "Valid area not found at PI\rArea @ Beg:{0}\rArea @ End:{1}\r".format(
                    self.aunits.Format(AreaType.SquareMeter, areaMin, None, at),
                    self.aunits.Format(AreaType.SquareMeter, areaMax, None, at))
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, s)
                return
            azMin = vStart.Angle
            azMax = vEnd.Angle
            increaseWithAngle = False
            if areaMin > areaMax:
                az = azMin
                azMin = azMax
                azMax = az
                a = areaMin
                areaMin = areaMax
                areaMax = a
                increaseWithAngle = True
            v = Vector3D()
            v.Angle = (azMin + azMax) / 2.0
            for i in range(30):  
                (notValid, areaTest, newPoly, v, polyArea) = self.ComputeFromVector(v, coord, poly, pcl)
                if notValid:
                    VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute area")
                    return
                s1 = "Pass {0}: Area:{1}\r".format(i + 1, self.aunits.Format(AreaType.SquareMeter, areaTest, None, at))
                self.textBoxEntityInfo.Text = self.textBoxEntityInfo.Text + s1
                diff = areaTest - areaM2
                if abs(diff) < 0.0001:
                    break
                if diff > 0:
                    #too large
                    if increaseWithAngle:
                        areaMax = areaTest
                        azMax = v.Angle
                    else:
                        areaMin = areaTest
                        azMin = v.Angle
                else:
                    #too small
                    if not increaseWithAngle:
                        areaMax = areaTest
                        azMax = v.Angle
                    else:
                        areaMin = areaTest
                        azMin = v.Angle

                v.Angle = self.Interpolate(azMin, azMax, areaMin, areaMax, areaM2)
            else:
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Did not converge withen tolerance")
                return
        else:
            for i in range(30):  
                (notValid, areaTest, newPoly, v, polyArea) = self.Compute1Area(sta, coord, poly)
                if notValid:
                    VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute area")
                    return
                s1 = "Pass {0}: Area:{1}\r   Loc: {2}\r".format(i + 1,
                     self.aunits.Format(AreaType.SquareMeter, areaTest, None, at), self.lunits.Format(sta, self.lfp))
                self.textBoxEntityInfo.Text = self.textBoxEntityInfo.Text + s1
                diff = areaTest - areaM2
                if abs(diff) < 0.0001:
                    break
                """ debug
                container = ViewHelper.GetContainerForView(self.coordCtl.ClickWindow) # usually WorldView
                l = container.Add(clr.GetClrType(Linestring))
                l.Append(newPoly, None, False, False)
                l.Layer = self.layerCtl.SelectedSerialNumber
                l.Name = str(i)
                """
                defAngle = abs(Vector3D.DeflectionAngle(v, prevV))
                prevV = v
                segLength = newPoly.Length()
                adj = diff / segLength
                # if deflection between this and previous segment more then 20
                # deg, use smaller step
                if defAngle > Math.PI / 9:
                    adj = adj / 10
                # don't do large steps
                if abs(adj) > segLength:
                    if adj < 0:
                        adj = -segLength / 3
                    else:
                        adj = segLength / 3
                if preArea == 0:
                    preArea = areaTest
                    preSta = sta
                    # just use a small adj the first time
                    adj = adj / 2
                    sta = sta + adj
                else:
                    sta2 = self.Interpolate(preSta, sta, preArea, areaTest, areaM2)
                    preArea = areaTest
                    preSta = sta
                    sta = sta2

            else:
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Did not converge withen tolerance")
                return

        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
       
        # the "with" statement will unroll any changes if something go wrong
        with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
            #now create a linestring
            container = line.GetSite() # usually WorldView
            if self.makePolygonCtl.IsChecked:
                l = container.Add(clr.GetClrType(Polygon))
                l.Append(polyArea, None, False, False)
                l.FillTransparency = 80
            else:
                l = container.Add(clr.GetClrType(Linestring))
                l.Append(newPoly, None, False, False)
            l.Layer = self.layerCtl.SelectedSerialNumber
            l.Name = self.nameCtl.SelectedName
            self.newLine = l.SerialNumber
            selectionSet = GlobalSelection.Items(self.currentProject)
            selectionSet.Set([l.SerialNumber])

            # you MUST commit the change to DB or they will all be un-done
            # (inside the "with" scope)
            failGuard.Commit()
        
            settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
            settings.SetUInt32("AdjustArea.Layer", l.Layer)
            settings.SetDouble("AdjustArea.TargetArea", area)
            settings.SetBoolean("AdjustArea.MakePolygon", self.makePolygonCtl.IsChecked)
            isMeters = self.lunits.DisplayType == LinearType.Meter
            if isMeters:
                s = "AdjustArea.AreaTypeM"
            else:
                s = "AdjustArea.AreaTypeF"
            settings.SetInt32(s, self.areaType.SelectedIndex)
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
        if self.nameCtl.SelectedName:
            self.nameCtl.SelectedName = PointHelper.Helpers.Increment(self.nameCtl.SelectedName, None)
        Keyboard.Focus(self.coordCtl)

