﻿"""
Sample TBC mcro to name line objects based on layer name
"""
#region
import clr
# Reference the WPF assemblies
clr.AddReference('IronPython.Wpf')
import wpf
from System import Type, Tuple, Array, Double, UInt32, Math, Byte, String, StringSplitOptions
from System.IO import StreamReader, File
from System.Windows.Controls import StackPanel
from System.Windows import  Visibility
from Microsoft.Win32 import OpenFileDialog
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode, TextJustification
from Trimble.Vce.Core import TransactMethodCall
from Trimble.DiskIO import OptionsManager

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IViewMember, ISettableName, IName, IViewSite, ISnapInHelpers
from Trimble.Vce.Interfaces.Client import CommandGranularity

from Trimble.Vce.ForeignCad import TextEntity, Leader, MText
from Trimble.Vce.Core import GlobalSelection
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
clr.AddReference("Trimble.Vce.UI.Controls")
try:
    from Trimble.Sdk.UI import UIEventArgs   
except:
    from Trimble.Sdk.Interfaces.UI import UIEventArgs
from Trimble.Vce.UI.Controls import VceMessageBox
clr.AddReference("Trimble.Vce.UI.Wrapper")
from Trimble.Vce.UI.Wrapper import CADWrapperHelper

from Trimble.Vce.UI.UIManager import TrimbleOffice
#endregion
# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "AssignNameFromLayer"
    cmdData.CommandName = "AssignNameFromLayer"
    cmdData.Caption = "_Assign Name From Layer"
    cmdData.UIForm = "AssignNameFromLayer"

    try:
        cmdData.Version = 1.53
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2018/11/29/tml-assignnamefromlayer"
        cmdData.HelpFile=cmdData.MacroInfo
    except:
        pass

    try:    
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Name from Layer"
        cmdData.ToolTipTitle = "Assign Names from Layer"
        cmdData.ToolTipTextFormatted = "Tool to name line object based on name of their layer."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class AssignNameFromLayer(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\AssignNameFromLayer.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        self.objs.IsEntityValidCallback=self.IsValid
        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu
        self.lType = clr.GetClrType(IPolyseg)
        self.useFileCB.IsChecked = OptionsManager.GetBool("AssignNameFromLayer:useFileCB", False)
        self.fileNameCtl.Text = OptionsManager.GetString("AssignNameFromLayer:fileNameCtl", "")
        self.ignoreNamedCtl.IsChecked = OptionsManager.GetBool("AssignNameFromLayer:ignoreNamedCtl", True)
        self.nameMaskCtl.Text = OptionsManager.GetString("AssignNameFromLayer:nameMaskCtl", "{L}")
        self.labelLinesCtl.IsChecked = OptionsManager.GetBool("AssignNameFromLayer:labelLinesCtl", False)        
        self.labelLinesCtl.Checked += self.labelLinesCtlChanged
        self.labelLinesCtl.Unchecked += self.labelLinesCtlChanged
        self.labelLinesCtlChanged(None, None)
        self.layerNameMap = {}

    def useFileCB_Checked(self, sender, args):
        if self.useFileCB.IsChecked:
            self.useMapFile.Visibility = Visibility.Visible
            self.useLayerMap.Visibility = Visibility.Collapsed
        else:
            self.useMapFile.Visibility = Visibility.Collapsed
            self.useLayerMap.Visibility = Visibility.Visible

    def Button_Click(self, sender, args):
        dlg = OpenFileDialog()
        dlg.DefaultExt = ".txt"
        res = dlg.ShowDialog()
        if res:
            self.fileNameCtl.Text = dlg.FileName

    def labelLinesCtlChanged(self, sender, args):
        if self.labelLinesCtl.IsChecked:
            self.labelingControls.Visibility = Visibility.Visible
        else:
            self.labelingControls.Visibility = Visibility.Collapsed

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o=self.currentProject.Concordance.Lookup(serial)
        if isinstance(o, self.lType):
            return True
        return False

    def LoadLayerMapFile(self, fileName):
        if not File.Exists(fileName):
            return True
        self.layerNameMap = {}
        splitOptions = Array[String]([",", "\n"]) # split so each number is a string
        with File.OpenText(fileName) as fh:
            while True:
                line = fh.ReadLine()
                if not line:
                    break
                a = line.Split(splitOptions, StringSplitOptions.RemoveEmptyEntries)
                if len(a) > 1:
                    l = a[0].lower()
                    if not self.layerNameMap.ContainsKey(l):
                        self.layerNameMap.Add(l, a[1])
        return False

    def MapNameByFile(self, layer):
        l = layer.Name.lower()
        if self.layerNameMap.ContainsKey(l):
            name = self.layerNameMap[l]
            name = name.TrimEnd()
            return name
        return None

    def OkClicked(self, cmd, e):
        if self.useFileCB.IsChecked:
            if self.LoadLayerMapFile(self.fileNameCtl.Text):
                VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not read map file")
                return
        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        wv = self.currentProject[Project.FixedSerial.WorldView]
        UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())
        # pause the graphics so every change doesn't draw
        wv.PauseGraphicsCache(True)
        try:
            self.worldPlotscale = wv.PlotScale
            layers = self.currentProject[Project.FixedSerial.LayerContainer]
            # the "with" statement will unroll any changes if something go wrong
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                for o in self.objs.SelectedMembers(self.currentProject):
                    if isinstance(o, self.lType) and isinstance(o, clr.GetClrType(ISettableName)):
                        name = IName.Name.__get__(o)
                        if not name or not self.ignoreNamedCtl.IsChecked:
                            l = layers.Lookup(o.Layer)
                            if l:
                                if self.useFileCB.IsChecked:
                                    newName = self.MapNameByFile(l)
                                    if newName:
                                        o.SetName = newName
                                else:
                                    newName = self.nameMaskCtl.Text
                                    newName = newName.Replace("{L}", l.Name)
                                    newName = newName.Replace("{N}", name)
                                    o.SetName = newName
                                    if self.labelLinesCtl.IsChecked:
                                        self.AddLineLabel(o, newName)

                # you MUST commit the change to DB or they will all be un-done (inside the "with" scope)
                failGuard.Commit()
            OptionsManager.SetValue("AssignNameFromLayer:useFileCB", self.useFileCB.IsChecked)
            OptionsManager.SetValue("AssignNameFromLayer:fileNameCtl", self.fileNameCtl.Text)
            OptionsManager.SetValue("AssignNameFromLayer:ignoreNamedCtl", self.ignoreNamedCtl.IsChecked)
            OptionsManager.SetValue("AssignNameFromLayer:nameMaskCtl", self.nameMaskCtl.Text)
            OptionsManager.SetValue("AssignNameFromLayer:labelLinesCtl", self.labelLinesCtl.IsChecked)            

        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
            UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
            wv.PauseGraphicsCache(False)

        cmd.CloseUICommand()

    def AddLineLabel(self, line, name):
        polySeg = line.ComputePolySeg()
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        midStation = (polySeg.FirstNode.Station + polySeg.LastNode.Station) / 2.0
        if polySeg.FindPointFromStation(midStation, seg, t, pcl, v):
            container = line.GetSite()
            textObj = container.Add(clr.GetClrType(MText))
            textObj.Point0 = Point3D(pcl.X, pcl.Y)
            textStyle = self.labelStyleCtl.SelectedSnapIn
            textObj.TextString = name
            v.Rotate90(Side.Left)
            textObj.Rotation = v.Angle
            textObj.Layer = self.layerCtl.SelectedSerialNumber
            textObj.TextStyleSerial = textStyle.SerialNumber
            textObj.WidthFactor = textStyle.WidthFactor
            textObj.AddWhiteOut = textStyle.WhiteOut
            textObj.AutoFlip = textStyle.AutoFlip
            textObj.ObliqueAngle = textStyle.ObliqueAngle
            textObj.Height = self.ComputeTextHeight(textStyle, container)
            textObj.AttachPoint = CADWrapperHelper.ConvertTextJustificationToAttachmentPoint(textStyle.Justification)

    def ComputeTextHeight(self, textStyle, viewContainer):
        height = textStyle.Height
        if isinstance(viewContainer, clr.GetClrType(IViewSite)):
            plotScale = viewContainer.PlotScale
        else:
            csd,view = ISnapInHelpers.GetViewSite(viewContainer)
            plotScale = view.PlotScale
        
        if plotScale < 1.1:
            # sheet view
            if textStyle.HeightMode == TextHeightMode.Ground:
                height = height / self.worldPlotscale
            else:
                height /= 1000.0
        else:
            if textStyle.HeightMode == TextHeightMode.Paper:
                height = height * plotScale / 1000
        return height