﻿"""
Sample TBC macro to compute pad locations from lot boundaries
"""
#region
import clr
# Reference the WPF assemblies
clr.AddReference('IronPython.Wpf')
import wpf
from System import Type, Tuple, Array, Double, Math, String, Action, Func
from System.Text import Encoding
from System.IO import StreamReader, MemoryStream, Stream, Path, File
from System.Collections.Generic import List
from System.Windows.Controls import StackPanel
from System.Windows import  Visibility
from System.Windows.Input import Keyboard
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Point as GdiPoint

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode, TextJustification
from Trimble.Vce.Core import TransactMethodCall

from Trimble.DiskIO import OptionsManager

from Trimble.Vce.Alignment import SuperelevationView, ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Intersections
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Node import Arc as NodeArc
from Trimble.Vce.Geometry.PolySeg.Segment import Segment

from Trimble.Vce.Geometry.Region import RegionBuilder, Region, Edge

from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IViewMember, ISettableName, IName, IViewSite, ISnapInHelpers
from Trimble.Vce.Interfaces.Client import CommandGranularity
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, GestureDispatcher

try:
    from Trimble.Sdk.UI import InputMethod, I2DProjection
    from Trimble.Sdk.UI import LineDrawingInstruction, PolylineDrawingInstruction
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod, I2DProjection
    from Trimble.Sdk.Interfaces.UI import LineDrawingInstruction, PolylineDrawingInstruction
from Trimble.Sdk.UI import GdiCursorBase

from Trimble.Vce.UI.Controls.Adapters import InputSettingsAdapter

from Trimble.Vce.ForeignCad import TextEntity, Leader, MText

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice, UIEvents
clr.AddReference("Trimble.Vce.UI.Wrapper")
from Trimble.Vce.UI.Wrapper import CADWrapperHelper
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "ComputeLotSetback"
    cmdData.CommandName = "ComputeLotSetback"
    cmdData.Caption = "_Compute Lot Setback"
    cmdData.UIForm = "ComputeLotSetback"

    try:
        cmdData.Version = 1.01
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo =  r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2019/11/04/compute-lot-setback-line"
        cmdData.HelpFile=cmdData.MacroInfo
    except:
        pass

    try:    
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Compute Lot Setback"
        cmdData.ToolTipTitle = "Compute Lot Setback"
        cmdData.ToolTipTextFormatted = "Compute lot setback line and optional pad from lot geometry"

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class ComputeLotSetback(StackPanel): # this inherits from the WPF StackPanel control
    Xaml="""
<StackPanel  HorizontalAlignment="Stretch"  
    xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
    xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
    xmlns:Wpf="clr-namespace:Trimble.Vce.UI.Controls.Wpf;assembly=Trimble.Vce.UI.Controls">

    <GroupBox BorderThickness="2">
        <GroupBox.Header>
            <StackPanel Orientation="Horizontal">
                <CheckBox x:Name="setbacksDefinedCB"  Content="Define setbacks" IsChecked="False" Checked="setbacksDefinedCB_Checked" Unchecked="setbacksDefinedCB_Checked"/>
                <Expander x:Name="expandSetbackInfo" IsExpanded="True" Expanded="expandSetbackInfo_Expanded" Collapsed="expandSetbackInfo_Expanded" ToolTip="Click here to hide setback properties"/>
            </StackPanel>            
        </GroupBox.Header>

        <StackPanel x:Name="setbackInfo" Visibility="Collapsed" >
            <Label Content="Setback line layer:" />
            <Grid >
                <Grid.ColumnDefinitions>
                    <ColumnDefinition Width="22"/>
                    <ColumnDefinition Width="*"/>
                </Grid.ColumnDefinitions>
                <CheckBox x:Name="createSetbackLineCtl" VerticalAlignment="Center" Grid.Column="0"  ToolTip="Check this if setback line should be created" />
                <Wpf:LayerPicker x:Name="layerSetbackCtl"  Grid.Column="1"/>
            </Grid>

            <Label Content="Front:"  />
            <Wpf:DistanceEdit x:Name="frontSetbackCtl" />
            <Label Content="Left:"  />
            <Wpf:DistanceEdit x:Name="leftSetbackCtl" />
            <Label Content="Right:"  />
            <Wpf:DistanceEdit x:Name="rightSetbackCtl" />
            <Label Content="Back:"  />
            <Wpf:DistanceEdit x:Name="backSetbackCtl"  />
            <Label Content="Side (Frontage):"  />
            <Wpf:DistanceEdit x:Name="sideSetbackCtl" />
        </StackPanel>
    </GroupBox>

    <GroupBox BorderThickness="2">
        <GroupBox.Header>
            <StackPanel Orientation="Horizontal">
                <CheckBox x:Name="drawPadCB"  Content="Draw Pad" IsChecked="True" Checked="drawPadCB_Checked"  Unchecked="drawPadCB_Checked"/>
                <Expander x:Name="expandPadInfo" IsExpanded="True" Expanded="expandPadInfo_Expanded" Collapsed="expandPadInfo_Expanded" ToolTip="Click here to hide pad properties"/>
            </StackPanel>
        </GroupBox.Header>
        <StackPanel x:Name="padInfo" Visibility="Visible" >
            <GroupBox Header="Pad size"  BorderThickness="2">
                <StackPanel>
                    <Label  Content="Width:" />
                    <Wpf:DistanceEdit x:Name="padWidthCtl" />
                    <Label  Content="Depth:" />
                    <Wpf:DistanceEdit x:Name="padDepthCtl" />
                </StackPanel>
            </GroupBox>
            <Label Content="Name:" />
            <Wpf:NameEdit x:Name="nameCtl" />
            <Label Content="Pad layer:" />
            <Wpf:LayerPicker x:Name="layerCtl" />
        </StackPanel>
    </GroupBox>
    <ComboBox x:Name="lotBoundaryType" SelectionChanged="lotBoundaryType_SelectionChanged" >
        <ComboBoxItem>Pick lot line</ComboBoxItem>
        <ComboBoxItem>Select lot boundary lines</ComboBoxItem>
    </ComboBox>
    <Label x:Name="lineLabelCtl" Content="Reference line:" />
    <Wpf:SelectEntity x:Name="lineCtl" ValueChanged="lineCtl_ValueChanged" AutoTab="False"/>
    <Label x:Name="objsLabelCtl" Content="Boundary lines:" />
    <Wpf:MemberSelection x:Name="objs" ValueChanged="objs_ValueChanged"/>
    <Label x:Name="regionLabelCtl" Content="Pick Lot:"  />
    <Wpf:CoordinateEdit x:Name="regionLocationCtl" ValueChanged="regionLocationCtl_ValueChanged" GotKeyboardFocus="regionLocationCtl_GotFocus" AutoTab="False"/>

    <GroupBox x:Name="padPlacement" BorderThickness="2" Header="Pad placement">
        <StackPanel  Visibility="Visible" >
            <CheckBox x:Name="autoPadPlacementCB"  Content="Auto" IsChecked="True" Checked="autoPadPlacementCB_Checked" Unchecked="autoPadPlacementCB_Checked"/>
            <Label Content="Pad direction:" />
            <Wpf:BearingEdit x:Name="padDirectionCtl" IsEnabled="False" AutoTabExtSkip="False" AutoTabExtVisible="True"/>
            <Label Content="Pad location:" />
            <Wpf:CoordinateEdit x:Name="padLocationCtl" IsEnabled="False" ValueChanged="padLocationCtl_ValueChanged"/>
        </StackPanel>
    </GroupBox>
</StackPanel>
"""
    def __init__(self, currentProject, macroFileFolder):
        xamlFile = macroFileFolder + r"\ComputeLotSetback.xaml"
        if File.Exists(xamlFile):
            with StreamReader(xamlFile) as s:
                wpf.LoadComponent(self, s) 
        else:
            with StreamReader(MemoryStream(Encoding.UTF8.GetBytes(self.Xaml))) as s:
                wpf.LoadComponent(self, s)  

        self.currentProject = currentProject
        self.macroName = "Compute Lot Setback"
        self.tmlFileFolder = macroFileFolder

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.cmd = cmd
        self.Caption = cmd.Command.Caption
        self.lType = clr.GetClrType(IPolyseg)
        self.lineCtl.AddTypeFilter(self.lType)

        settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
        l = settings.GetUInt32("ComputeLotSetback.Layer", 0)
        if self.currentProject.Concordance.Lookup(l):
            self.layerCtl.SelectedSerialNumber = l
        l = settings.GetUInt32("ComputeLotSetback.SetbackLayer", 0)
        if self.currentProject.Concordance.Lookup(l):
            self.layerSetbackCtl.SelectedSerialNumber = l

        self.LastPickedLine = 0
        self.LastSetbackLineCreated = 0
        self.LastPadLineCreated = 0
        self.SavedRegions = None
        self.SetbackPolyseg = None
        self.padWidthCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:padWidthCtl", Double.NaN)
        self.padDepthCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:padDepthCtl", Double.NaN)
        self.setbacksDefinedCB.IsChecked = OptionsManager.GetBool("ComputeLotSetback:setbacksDefinedCB", True)
        self.createSetbackLineCtl.IsChecked = OptionsManager.GetBool("ComputeLotSetback:createSetbackLineCtl", True)
        self.drawPadCB.IsChecked = OptionsManager.GetBool("ComputeLotSetback:drawPadCB", True)
        self.autoPadPlacementCB.IsChecked = OptionsManager.GetBool("ComputeLotSetback:autoPadPlacementCB", True)
        self.frontSetbackCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:frontSetbackCtl", Double.NaN)
        self.leftSetbackCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:leftSetbackCtl", Double.NaN)
        self.rightSetbackCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:rightSetbackCtl", Double.NaN)
        self.backSetbackCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:backSetbackCtl", Double.NaN)
        self.sideSetbackCtl.Distance = OptionsManager.GetDouble("ComputeLotSetback:sideSetbackCtl", Double.NaN)
        self.padDirectionCtl.AutoTabExtSkip = OptionsManager.GetBool("ComputeLotSetback:padDirectionCtl", True)

        self.objsLabelCtl.Visibility = Visibility.Collapsed
        self.objs.Visibility = Visibility.Collapsed
        self.lotBoundaryType.SelectedIndex = OptionsManager.GetInt("ComputeLotSetback:lotBoundaryType", 0)
        #self.lotBoundaryType.Visibility = Visibility.Collapsed
        self.padLocationCtl.ShowGdiCursor += self.DrawCustomCursor

        def SetFocusToControl():
            Keyboard.Focus(self.lineCtl)
        Dispatcher.BeginInvoke(Dispatcher.CurrentDispatcher, Action(SetFocusToControl))

        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        if isinstance(o, self.lType):
            return True
        return False

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def lineCtl_ValueChanged(self, sender, e):
        sn = self.lineCtl.SerialNumber
        if sn:
            GlobalSelection.Items(self.currentProject).Set([sn])
        else:
            GlobalSelection.Clear()
        if e.Cause == InputMethod.Mouse:
            # check if picked the same line again, delete old results
            deletePrev = sn and sn == self.LastPickedLine
            self.LastPickedLine = sn
            self.SetbackPolyseg = None
            if self.ValidateData(True, False):
                return
            pickLocation = self.lineCtl.PickPointWorld
            if self.setbacksDefinedCB.IsChecked:
                poly = self.lineCtl.Entity.ComputePolySeg()
                poly = self.SaveSetbackLine(poly, self.lineCtl.PickPointWorld, deletePrev)
                if self.drawPadCB.IsChecked:
                    if poly:
                        if self.autoPadPlacementCB.IsChecked:
                            poly = self.SavePad(poly, pickLocation, self.lineCtl.PickPointWorld, deletePrev)
                            Keyboard.Focus(self.setbacksDefinedCB)
                            Keyboard.Focus(self.lineCtl)                        
                        else:
                            self.SetUpPad(poly, pickLocation)
                            if self.padDirectionCtl.AutoTabExtSkip:
                                Keyboard.Focus(self.padLocationCtl)
                            else:
                                Keyboard.Focus(self.padDirectionCtl)
            else:
                if self.drawPadCB.IsChecked:
                    poly = self.lineCtl.Entity.ComputePolySeg()
                    if self.autoPadPlacementCB.IsChecked:                        
                        if poly:
                            poly = self.SavePad(poly, pickLocation, self.lineCtl.PickPointWorld, deletePrev)
                        Keyboard.Focus(self.setbacksDefinedCB)
                        Keyboard.Focus(self.lineCtl)
                    else:
                        if poly:
                            self.SetUpPad(poly, pickLocation)
                        if not self.padDirectionCtl.AutoTabExtSkip:
                            Keyboard.Focus(self.padDirectionCtl)                            
                        else:
                            Keyboard.Focus(self.padLocationCtl)

    def objs_ValueChanged(self, sender, e):
        self.SavedRegions = None

    def padLocationCtl_ValueChanged(self, sender, e):
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)
  
    def regionLocationCtl_GotFocus(self, sender, e):
        if self.SavedRegions:
            return
        rb = RegionBuilder()       
        for o in self.objs.SelectedMembers(self.currentProject):
            if isinstance(o, self.lType):
                poly = o.ComputePolySeg()
                if not poly or poly.NumberOfNodes < 2:
                    continue
                rb.Add(poly)

        nRegions = rb.Build()
        self.SavedRegions = rb

    def regionLocationCtl_ValueChanged(self, sender, e):
        if not self.SavedRegions:
            return
        if self.ValidateData(True, False):
            return
        pickLocation = self.regionLocationCtl.ClickLocation
        if pickLocation.IsUndefined:
            return
        for r in self.SavedRegions.Regions():
            poly = r.AreaPolyseg()
            if poly.PointInPolyseg(pickLocation) == Side.Out:
                continue
            self.LastPickedLine = 0
            self.SetbackPolyseg = None
            poly = self.SaveSetbackLine(poly, pickLocation, False)
            if self.drawPadCB.IsChecked:
                if poly:
                    if self.autoPadPlacementCB.IsChecked:
                        poly = self.SavePad(poly, pickLocation, pickLocation, False)                     
                    else:
                        self.SetUpPad(poly, pickLocation)
                        if self.padDirectionCtl.AutoTabExtSkip:
                            Keyboard.Focus(self.padLocationCtl)
                        else:
                            Keyboard.Focus(self.padDirectionCtl)        

    def ValidateData(self, updateStatusMessage, includePadControls):
        if  self.lotBoundaryType.SelectedIndex  == 0:
            line = self.lineCtl.Entity
            if not line:
                if updateStatusMessage:
                    self.lineCtl.StatusMessage = "No line selected"
                return True
            if updateStatusMessage:
                self.lineCtl.StatusMessage = ""

            poly1 = self.lineCtl.Entity.ComputePolySeg()
            if not poly1 or poly1.NumberOfNodes < 2 or not poly1.IsClosed:
                if updateStatusMessage:
                    self.lineCtl.StatusMessage = "Lot boundary not valid (,ust be closed)"
                return True
            if updateStatusMessage:
                self.lineCtl.StatusMessage = ""

        if self.drawPadCB.IsChecked:
            padWidth = self.padWidthCtl.Distance
            if Double.IsNaN(padWidth):
                if updateStatusMessage:
                    self.padWidthCtl.StatusMessage = "Enter pad width"
                    Keyboard.Focus(self.padWidthCtl)
                return True
            if updateStatusMessage:
                self.padWidthCtl.StatusMessage = ""

            padDepth = self.padDepthCtl.Distance
            if Double.IsNaN(padDepth):
                if updateStatusMessage:
                    self.padDepthCtl.StatusMessage = "Enter pad width"
                    Keyboard.Focus(self.padDepthCtl)
                return True
            if updateStatusMessage:
                self.padDepthCtl.StatusMessage = ""

            if not self.autoPadPlacementCB.IsChecked and includePadControls:
                b = self.padDirectionCtl.Direction
                if Double.IsNaN(b):
                    if updateStatusMessage:
                        self.padDirectionCtl.StatusMessage = "Enter pad direction"
                        Keyboard.Focus(self.padDirectionCtl)
                    return True
                if updateStatusMessage:
                    self.padDirectionCtl.StatusMessage = ""
                padLocation = self.padLocationCtl.CoordinateWorld 
                if padLocation.IsUndefined:
                    if updateStatusMessage:
                        self.padLocationCtl.StatusMessage = "Enter pad location"
                        Keyboard.Focus(self.padLocationCtl)
                    return True
                if updateStatusMessage:
                    self.padLocationCtl.StatusMessage = ""

        if self.setbacksDefinedCB.IsChecked:
            frontSetback = self.frontSetbackCtl.Distance
            if Double.IsNaN(frontSetback) or frontSetback < 0:
                if updateStatusMessage:
                    self.frontSetbackCtl.StatusMessage = "Define setback"
                    Keyboard.Focus(self.frontSetbackCtl)
                return True
            if updateStatusMessage:
                self.frontSetbackCtl.StatusMessage = ""

            leftSetback = self.leftSetbackCtl.Distance
            if Double.IsNaN(leftSetback) or leftSetback < 0:
                self.leftSetbackCtl.StatusMessage = "Define setback"
                Keyboard.Focus(self.leftSetbackCtl)
                return True
            self.leftSetbackCtl.StatusMessage = ""

            rightSetback = self.rightSetbackCtl.Distance
            if Double.IsNaN(rightSetback) or rightSetback < 0:
                if updateStatusMessage:
                    self.rightSetbackCtl.StatusMessage = "Define setback"
                    Keyboard.Focus(self.rightSetbackCtl)
                return True
            if updateStatusMessage:
                self.rightSetbackCtl.StatusMessage = ""

            backSetback = self.backSetbackCtl.Distance
            if Double.IsNaN(backSetback) or backSetback < 0:
                if updateStatusMessage:
                    self.backSetbackCtl.StatusMessage = "Define setback"
                    Keyboard.Focus(self.backSetbackCtl)
                return True
            if updateStatusMessage:
                self.backSetbackCtl.StatusMessage = ""

            sideSetback = self.sideSetbackCtl.Distance
            if Double.IsNaN(sideSetback) or sideSetback < 0:
                if updateStatusMessage:
                    self.sideSetbackCtl.StatusMessage = "Define setback"
                    Keyboard.Focus(self.sideSetbackCtl)
                return True
            if updateStatusMessage:
                self.sideSetbackCtl.StatusMessage = ""
        return False

    def CancelClicked(self, cmd, args):
        OptionsManager.SetValue("ComputeLotSetback:padWidthCtl", self.padWidthCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:padDepthCtl", self.padDepthCtl.Distance)            
        OptionsManager.SetValue("ComputeLotSetback:setbacksDefinedCB", self.setbacksDefinedCB.IsChecked)
        OptionsManager.SetValue("ComputeLotSetback:createSetbackLineCtl", self.createSetbackLineCtl.IsChecked)
        OptionsManager.SetValue("ComputeLotSetback:autoPadPlacementCB", self.autoPadPlacementCB.IsChecked)
        OptionsManager.SetValue("ComputeLotSetback:drawPadCB", self.drawPadCB.IsChecked)
        OptionsManager.SetValue("ComputeLotSetback:frontSetbackCtl", self.frontSetbackCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:leftSetbackCtl",  self.leftSetbackCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:rightSetbackCtl", self.rightSetbackCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:backSetbackCtl", self.backSetbackCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:sideSetbackCtl", self.sideSetbackCtl.Distance)
        OptionsManager.SetValue("ComputeLotSetback:padDirectionCtl", self.padDirectionCtl.AutoTabExtSkip)
        OptionsManager.SetValue("ComputeLotSetback:lotBoundaryType", self.lotBoundaryType.SelectedIndex)
        cmd.CloseUICommand()

    def SaveSetbackLine(self, poly, linePick, deletePrev):
        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        try:
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                wv = self.currentProject[Project.FixedSerial.WorldView]
                polys = self.BreakPolyseg(poly, linePick, self.frontSetbackCtl.Distance)

                if not polys:
                    VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute setback line.")
                    return None
                poly = self.ComputeSetbackLine(polys)
                if poly:
                    ls = wv.Add(clr.GetClrType(Linestring))
                    ls.Layer = self.layerSetbackCtl.SelectedSerialNumber 
                    #ls.Name = self.nameCtl.SelectedName
                    ls.Append(poly, None, False, False)
                    if deletePrev:
                        if wv.Lookup(self.LastSetbackLineCreated):
                            wv.Remove(self.LastSetbackLineCreated)
                    self.LastSetbackLineCreated = ls.SerialNumber
                    failGuard.Commit()
            settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
            settings.SetUInt32("ComputeLotSetback.SetbackLayer", self.layerSetbackCtl.SelectedSerialNumber)
        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
        return poly

    def SetUpPad(self, poly, fromPoint):
        polys = self.BreakPolyseg(poly, fromPoint, self.frontSetbackCtl.Distance)
        self.SetbackPolyseg = polys
        if not polys:
            return None
        (pOnLine, side, tMax, v) = self.GetDefaultPadLocation(polys, fromPoint, self.padWidthCtl.Distance, self.padDepthCtl.Distance)
        if pOnLine:
            self.padDirectionCtl.Direction = v.Angle

    def SavePad(self, poly, fromPoint, linePick, deletePrev):
        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        try:
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                wv = self.currentProject[Project.FixedSerial.WorldView]
                polys = self.BreakPolyseg(poly, fromPoint, self.frontSetbackCtl.Distance)

                if not polys:
                    VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute pad.")
                    return None

                (pOnLine, side, tMax, v) = self.GetDefaultPadLocation(polys, linePick, self.padWidthCtl.Distance, self.padDepthCtl.Distance)
                polySetback = polys[0].Clone()
                extend = 50.0
                polySetback.Extend(extend)
                pw = self.padWidthCtl.Distance
                if self.autoPadPlacementCB.IsChecked:
                    pwL = pw/2
                    pwR = pw/2
                else:
                    if side == Side.Left:
                        pwL=0
                        pwR=pw
                    elif side == Side.Right:
                        pwL = pw
                        pwR = 0
                    else:
                        pwL = pw/2
                        pwR = pw/2
                (lineLen, pMid) = self.PadWidthAtT(polys, pOnLine, tMax, v)
                if pMid:
                    tMax = self.AdjPadToWidthLimits(polySetback, pMid, v, pwL, pwR)
                    (lineLen, p) = self.PadWidthAtT(polys, pMid, tMax, v)
                    if lineLen < self.padWidthCtl.Distance:
                        t2 = tMax + 1
                        (lineLen2, p) = self.PadWidthAtT(polys, pMid, t2, v)
                        if lineLen != lineLen2:
                            tMax = self.Interpolate(tMax, t2, lineLen, lineLen2, self.padWidthCtl.Distance)
                    polyRtn = self.BuildPadPoly(pMid, tMax, v, pwL, pwR, self.padDepthCtl.Distance)
                    ls = wv.Add(clr.GetClrType(Linestring))
                    ls.Layer = self.layerCtl.SelectedSerialNumber 
                    ls.Name = self.nameCtl.SelectedName
                    ls.Append(polyRtn, None, False, False)
                    if deletePrev:
                        if wv.Lookup(self.LastPadLineCreated):
                            wv.Remove(self.LastPadLineCreated)
                    self.LastPadLineCreated = ls.SerialNumber
                    failGuard.Commit()

            settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
            settings.SetUInt32("ComputeLotSetback.Layer", self.layerCtl.SelectedSerialNumber)
        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

    def OkClicked(self, cmd, e):
        if self.ValidateData(True, True):
            return

        pw = self.padWidthCtl.Distance
        pwL = pw / 2
        pwR = pw / 2
        b = self.padDirectionCtl.Direction
        v = Vector3D(1,0)
        v.Angle = b

        polys = self.SetbackPolyseg
        if not polys:
            return
        padLocation = self.padLocationCtl.CoordinateWorld
        tMax = self.AdjPadToWidthLimits(polys[0], padLocation, v, pwL, pwR)
        (lineLen, p) = self.PadWidthAtT(polys, padLocation, tMax, v)
        if lineLen < self.padWidthCtl.Distance:
            t2 = tMax + 1
            (lineLen2, p) = self.PadWidthAtT(polys, padLocation, t2, v)
            if lineLen != lineLen2:
                tMax = self.Interpolate(tMax, t2, lineLen, lineLen2, self.padWidthCtl.Distance)
        polyRtn = self.BuildPadPoly(padLocation, tMax, v, pwL, pwR, self.padDepthCtl.Distance)
        pts = List[GdiPoint]() 
        if polyRtn:
            self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
            try:
                with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                    wv = self.currentProject[Project.FixedSerial.WorldView]
                    ls = wv.Add(clr.GetClrType(Linestring))
                    ls.Layer = self.layerCtl.SelectedSerialNumber 
                    ls.Name = self.nameCtl.SelectedName
                    ls.Append(polyRtn, None, False, False)
                    failGuard.Commit()

                settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
                settings.SetUInt32("ComputeLotSetback.Layer", self.layerCtl.SelectedSerialNumber)
            finally:
                self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        if self.lotBoundaryType.SelectedIndex == 0:
            Keyboard.Focus(self.lineCtl)
        else:
            Keyboard.Focus(self.regionLocationCtl)

    def DrawCustomCursor(self, sender, e):
        if not isinstance(e.MessagingView, clr.GetClrType(I2DProjection)):
            return
        if self.ValidateData(False, False):
            return

        cursorLocation = e.MessagingView.ConvertPageToProject(Point3D(e.MousePosition.X, e.MousePosition.Y))  
        if self.lotBoundaryType.SelectedIndex == 0:
            poly1 = self.lineCtl.Entity.ComputePolySeg()
        else:
            poly1 = None
            if not self.SavedRegions:
                return
            if self.ValidateData(True, False):
                return
            coord = self.regionLocationCtl.ClickLocation
            if coord.IsUndefined:
                return
            for r in self.SavedRegions.Regions():
                poly = r.AreaPolyseg()
                if poly.PointInPolyseg(cursorLocation) == Side.Out:
                    continue
                poly1 = poly
                break

        if not poly1 or poly1.NumberOfNodes < 2:
            return

        b = self.padDirectionCtl.Direction
        if Double.IsNaN(b):
            return
                      
        pw = self.padWidthCtl.Distance
        if Double.IsNaN(pw):
            return
        pwL = pw / 2
        pwR = pw / 2
        v = Vector3D(1,0)
        v.Angle = b

        polys = self.SetbackPolyseg
        if not polys:
            return
        tMax = self.AdjPadToWidthLimits(polys[0], cursorLocation, v, pwL, pwR)
        (lineLen, p) = self.PadWidthAtT(polys, cursorLocation, tMax, v)
        if lineLen < self.padWidthCtl.Distance:
            t2 = tMax + 1
            (lineLen2, p) = self.PadWidthAtT(polys, cursorLocation, t2, v)
            if lineLen != lineLen2:
                tMax = self.Interpolate(tMax, t2, lineLen, lineLen2, self.padWidthCtl.Distance)
        polyRtn = self.BuildPadPoly(cursorLocation, tMax, v, pwL, pwR, self.padDepthCtl.Distance)
        pts = List[GdiPoint]() 
        if polyRtn:
            for node in polyRtn.Nodes():
                pagePt = e.MessagingView.ConvertProjectToPage(node.Point)
                g = GdiPoint(pagePt.X, pagePt.Y)
                pts.Add(g)

        #pagePt = e.MessagingView.ConvertProjectToPage(pcl)
        #gdiPoint = GdiPoint(pagePt.X, pagePt.Y)
        cursor = MyCustomCursor(e.MousePosition, pts, True, e.MousePosition)
        GestureDispatcher.PushVisibleCursor(cursor.Show(None, e.MessagingView, e.MousePosition));

    def Interpolate(self, x1, x2, y1, y2, y):
        r = (y2 - y1) / (x2 - x1)
        x = (y - y1) / r 
        x = x + x1
        return x

    def GetDefaultPadLocation(self, polys, fromPoint, padWidth, padDepth):
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        staRef = clr.Reference[float]()
        offset = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v2 = clr.Reference[Vector3D]()
        side = clr.Reference[Side]()
        polyFront = polys[0].Clone()
        extend = 50.0
        polyFront.Extend(extend)
        found = polyFront.FindPointFromPoint(fromPoint, seg, t, pcl, staRef, v2,offset, side)
        if not found:
            return (None, None, None, None)
        along = (staRef.Value - polyFront.FirstNode.Station - extend) / \
            (polyFront.LastNode.Station - polyFront.FirstNode.Station - 2 * extend)
        if not polys[1].FirstSegment:
            along = 0.0
        if not polys[3].LastSegment:
            along = 1.0
        if along < 0.25: # use left side direction
            side = Side.Left
            pOnLine = polys[3].LastSegment.EndPoint   
            v = polys[3].FirstSegment.BeginPoint - polys[3].LastSegment.EndPoint
        elif along > 0.75: # use right side
            side = Side.Right
            pOnLine = polys[1].FirstSegment.BeginPoint
            v = polys[1].LastSegment.EndPoint - polys[1].FirstSegment.BeginPoint
        else:
            side = Side.On
            seg = clr.Reference[Segment]()
            t = clr.Reference[float]()
            pcl = clr.Reference[Point3D]()
            v = clr.Reference[Vector3D]()
            sta = (polyFront.FirstNode.Station + polyFront.LastNode.Station) / 2.0
            if not polyFront.FindPointFromStation(sta, seg, t, pcl, v):
                return (None, None, None, None)
            pOnLine = pcl.Value
            if False:
                # use direction at mid point
                v = v.Value
                v.Rotate180()                
            else:
                vL = polys[3].FirstSegment.BeginPoint - polys[3].LastSegment.EndPoint 
                vR = polys[1].LastSegment.EndPoint - polys[1].FirstSegment.BeginPoint
                vR.To2D()
                vR.Normalize()
                vL.To2D()
                vL.Normalize()
                v = (vL + vR) / 2.0

        v.To2D()
        v.Normalize()
        tMax = -1e10

        for node in polys[0].Nodes():
            (t, p, p, p) = Line2D.TAtPoint(pOnLine, v, node.Point)
            if t > tMax:
                tMax = t

        return (pOnLine, side, tMax, v)

    def AdjPadToWidthLimits(self, polyFront, pStart, v, padWidthLeft, padWidthRight):
        # compute t parameter along vector that best matches line
        v2 = Vector3D(v)
        v2.Rotate90(Side.Right)
        p = polyFront.Clone()
        pBeg = pStart - 1000 * v
        p1 = pBeg - v2 * padWidthLeft
        p2 = pBeg + v2 * padWidthRight
        pEnd = pStart + 1000 * v
        p3 = pEnd + v2 * padWidthLeft
        p4 = pEnd - v2 * padWidthRight
        clipPoly = PolySeg()
        clipPoly.Add(Array[Point3D]([p1, p2, p3, p4, p1]))
        p.Clip(clipPoly, Side.Out)
        tMax = -1e10
        for node in p.Nodes():
            (t, p0, p1, p2) = Line2D.TAtPoint(pStart, v, node.Point)
            if t > tMax:
                tMax = t
            if isinstance(node, clr.GetClrType(NodeArc)) and node.Side == Side.Right:
                if not node.Visible:
                    continue
                testPt = node.RadiusPoint + v * node.Radius
                (p0, p1, p2, dist2, sideR) = Line2D.Distance2ToPoint(pBeg, pEnd,	 testPt)
                if sideR == Side.Right:
                    if dist2 > padWidthRight * padWidthRight:
                        continue
                if sideR == Side.Left:
                    if dist2 > padWidthLeft * padWidthLeft:
                        continue
                (t, p0, p1, p2) = Line2D.TAtPoint(pStart, v, testPt)
                if t > tMax:
                    seg = p.CreateSegment(node)
                    (rtn, ts, pcl, side) = seg.PerpendicularPoint(testPt)
                    if ts > 0 and ts < 1:
                        tMax = t
        return tMax

    def PadWidthAtT(self, polys, pStart, tTest, v):
        testPt = pStart + tTest * v
        v2 = Vector3D(v)
        v2.Rotate90(Side.Right)
        acrossPoly = PolySeg()
        acrossPoly.Add(Array[Point3D]([testPt - 1000 * v2, testPt + 1000 * v2]))
        ints = Intersections()
        polyLeft = polys[3].Clone()
        if polyLeft.NumberOfNodes:
            polyLeft.Extend(500)
        else:
            polyLeft = polys[0].Clone()
            polyLeft += polys[2].Clone()      
        polyRight = polys[1].Clone()
        if polyRight.NumberOfNodes:
            polyRight.Extend(500)
        else:
            polyRight = polys[0].Clone()
            polyRight += polys[2].Clone()
        pMin = None
        tMin = 1e10
        mMax = None
        tMax = -1e10
        if polyLeft.Intersect(acrossPoly, True, ints):
            for i in ints:
                if i.T2 < tMin:
                    pMin = i.Point
                    tMin = i.T2
                if i.T2 > tMax:
                    pMax = i.Point
                    tMax = i.T2
            if polyRight.Intersect(acrossPoly, True, ints):
                for i in ints:
                    if i.T2 < tMin:
                        pMin = i.Point
                        tMin = i.T2
                    if i.T2 > tMax:
                        pMax = i.Point
                        tMax = i.T2
                lineLen = (pMax - pMin).Length
                pMid = 0.5 * (pMax + pMin)
                return (lineLen, pMid)

        return (Double.NaN, None)

    def BuildPadPoly(self, pStart, tTest, v, padWidthLeft, padWidthRight, padDepth):
        pStart = pStart + tTest * v
        v2 = Vector3D(v)
        v2.Rotate90(Side.Right)
        p1 = pStart - v2 * padWidthLeft
        p2 = pStart + v2 * padWidthRight
        p3 = p2 + v * padDepth
        p4 = p1 + v * padDepth
        poly = PolySeg()
        poly.Add(Array[Point3D]([p1, p2, p3, p4, p1]))
        return poly

    def ComputeSetbackLine(self, polys):
        extend = Math.Max(self.frontSetbackCtl.Distance, self.backSetbackCtl.Distance)
        rb = RegionBuilder()
        p = polys[0].Clone() # front
        (rtn, p) = p.Offset(Side.Left, self.frontSetbackCtl.Distance)
        p.Extend(extend, True)
        rb.Add(p)

        p = polys[1].Clone() # right
        if p.FirstNode:
            if p.FirstNode.Description == "_@Side":
                (rtn, p) = p.Offset(Side.Left, self.sideSetbackCtl.Distance)
            else:
                (rtn, p) = p.Offset(Side.Left, self.rightSetbackCtl.Distance)
            #p.Extend(extend, True)
            rb.Add(p)

        p = polys[2].Clone() # back
        (rtn, p) = p.Offset(Side.Left, self.backSetbackCtl.Distance)
        p.Extend(extend, True)
        rb.Add(p)

        p = polys[3].Clone() # left
        if p.FirstNode:
            if p.FirstNode.Description == "_@Side":
                (rtn, p) = p.Offset(Side.Left, self.sideSetbackCtl.Distance)
            else:
                (rtn, p) = p.Offset(Side.Left, self.leftSetbackCtl.Distance)
            #p.Extend(extend, True)
            rb.Add(p)
        
        nRegions = rb.Build()
        if nRegions == 0:
            return None
        rb.SortRegions() # make sure largest ones first
        return rb.Regions()[0].AreaPolyseg()

    def CheckDeflection(self, v1,v2):
        if v1.Is2D:
            v1.Z = 0
        if v2.Is2D:
            v2.Z = 0
        v1.Normalize()
        v2.Normalize()
        (dot, v1, v2) = Vector3D.DotProduct(v1, v2)
        # check for round-off
        if dot > 1.0:
            dot = 1.0
        if dot < -1.0:
            dot = -1.0
        angle = Math.Acos(dot)
        joinIt = angle <= Math.PI / 6 # join if deflection less the 30 deg
        return joinIt

    def BreakPolyseg(self, polySeg, fromPoint, frontSetback):
        if polySeg.IsClockWise():
            polySeg = polySeg.Clone()
            polySeg.Reverse()
        lastSeg = None
        newPoly = None
        polys = []
        for node in polySeg.Nodes():
            seg = polySeg.CreateSegment(node)
            if seg:
                if lastSeg:
                    (b,vLast) = lastSeg.Slope(1.0) # get direction at end of last segment
                    (b,v) = seg.Slope(0.0)
                    if self.CheckDeflection(vLast, v):
                        newPoly.Add(seg)
                        lastSeg = seg
                        continue
                if newPoly:
                    polys.Add(newPoly)
                newPoly = PolySeg()
                newPoly.Add(seg)      
                lastSeg = seg
        if newPoly:
            (b,vLast) = polys[0].FirstSegment.Slope(0.0)
            (b,v) = newPoly.LastSegment.Slope(1.0)
            if self.CheckDeflection(v, vLast):
                polys[0] = newPoly + polys[0] # append end and start of polyseg
            else:
                polys.Add(newPoly)

        polys = self.SortSegments(polys, fromPoint)
        if not polys:
            return None
        # polysegs sorted front-right-back-left
        if len(polys) == 4:
            return polys
        if len(polys) == 3:
            # likely one side is tangent to front.
            return self.BreakFrontSegment(polys, fromPoint, frontSetback)
        if len(polys) > 4:
            # likely two segments along back
            return self.MergeBackSegments(polys)
        return None

    def SortSegments(self, polys, fromPoint):
        # put segments in order of front, right, back, left
        numPolys = len(polys)
        frontIndex = -1
        bestOffset = 1000.0
        for i in range(numPolys):
            seg = clr.Reference[Segment]()
            t = clr.Reference[float]()
            staRef = clr.Reference[float]()
            offset = clr.Reference[float]()
            pcl = clr.Reference[Point3D]()
            v2 = clr.Reference[Vector3D]()
            side = clr.Reference[Side]()
            polyTest = polys[i].Clone()
            polyTest.Extend(1.0)
            found = polyTest.FindPointFromPoint(fromPoint, seg, t, pcl, staRef, v2,offset, side)
            if found and offset.Value < bestOffset:
                frontIndex = i
                bestOffset = offset.Value

        if frontIndex == -1:
            return None
        newPolys = []
        for i in range(numPolys):
            ii = i + frontIndex
            if ii >= numPolys:
                ii -= numPolys
            newPolys.Add(polys[ii])
        return newPolys

    def BreakFrontSegment(self, polys, fromPoint, frontSetback):
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        staRef = clr.Reference[float]()
        offset = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v2 = clr.Reference[Vector3D]()
        side = clr.Reference[Side]()
        polyFront = polys[0]
        found = polyFront.FindPointFromPoint(fromPoint, seg, t, pcl, staRef, v2,offset, side)
        if not found:
            return None
        seg = seg.Value
        node = seg.Node
        # search for node where we can break polyseg. The small arc goes on the "side" segment
        foundNode = None
        while node:
            if isinstance(node, clr.GetClrType(NodeArc)):
                if node.Radius < 1.5 * frontSetback and node.Length > frontSetback / 2:
                    foundNode = node
                    break
            node = node.Next
        if not foundNode:
            node = seg.Node
            while node:
                if isinstance(node, clr.GetClrType(NodeArc)):
                    if node.Radius < 1.5 * frontSetback and node.Length > frontSetback / 2:
                        foundNode = node.Next
                        break
                node = node.Previous
        if not foundNode:
            newPolys = []
            newPolys.Add(polyFront)
            newPolys.Add(polys[1])
            newPolys.Add(polys[2])
            newPolys.Add(PolySeg())
            return self.SortSegments(newPolys, fromPoint)

        (split, p1) = polyFront.Split(foundNode)
        newPolys = []
        # name first node so we can do special handeling of this side (different offset)
        polyFront.FirstNode.Description = "_@Side"
        p1.FirstNode.Description = "_@Side"
        newPolys.Add(polyFront)
        newPolys.Add(p1)
        newPolys.Add(polys[1])
        newPolys.Add(polys[2])
        return self.SortSegments(newPolys, fromPoint)

    def MergeBackSegments(self, polys):
        # merge the 3rd and 4th polyseg
        newPolys = []
        newPolys.Add(polys[0])
        newPolys.Add(polys[1])
        num = len(polys)
        p = polys[2]
        for i in range(num-4):
            p = p + polys[i + 3]
        newPolys.Add(p)
        newPolys.Add(polys[num-1])
        return newPolys

    def VectorIntersectLine(self, fromPoint, directionVector, polyToIntersect, polyToIntersectVert):
        acrossPoly = PolySeg()
        acrossPoly.Add(Array[Point3D]([fromPoint - 1000.0 * directionVector, fromPoint + 1000.0 * directionVector]))
        ints = Intersections()
        if not polyToIntersect.Intersect(acrossPoly, True, ints):
            return (Point3D.Undefined, Double.NaN)
        # we could get more than one intersection.
        # Find nearest to middle (t value of 0.5)
        bestT = 1e10
        bestI = 0
        for i in range(ints.Count):
            if Math.Abs(ints[i].T2 - 0.5) < bestT:
                bestT = ints[i].T2
                bestI = i
        i = ints[bestI]
        staRef = i.Segment1.Station(i.T1)
        (ret, pRef) = i.Segment1.ComputePoint(i.T1)
        # check if ref line has vertical polyseg
        if polyToIntersectVert:
            (found, segment, p, isOn) = polyToIntersectVert.LocateX(Point3D(staRef, 0.0))
            if found:
                pRef.Z = p.Y
        return (pRef, staRef)
    
    def setbacksDefinedCB_Checked(self, sender, e):
        if self.setbacksDefinedCB.IsChecked:
            self.expandSetbackInfo.IsExpanded = True
            self.setbackInfo.Visibility = Visibility.Visible
        else:
            self.setbackInfo.Visibility = Visibility.Collapsed

    def expandSetbackInfo_Expanded(self, sender, e):        
        if self.expandSetbackInfo.IsExpanded:
            self.setbackInfo.Visibility = Visibility.Visible
        else:
            self.setbackInfo.Visibility = Visibility.Collapsed
    
    def drawPadCB_Checked(self, sender, e):
        if self.drawPadCB.IsChecked:
            self.expandPadInfo.IsExpanded = True
            self.padInfo.Visibility = Visibility.Visible
            self.padPlacement.Visibility = Visibility.Visible            
        else:
            self.padInfo.Visibility = Visibility.Collapsed
            self.padPlacement.Visibility = Visibility.Collapsed
 
    def expandPadInfo_Expanded(self, sender, e):        
        if self.expandPadInfo.IsExpanded:
            self.padInfo.Visibility = Visibility.Visible
        else:
            self.padInfo.Visibility = Visibility.Collapsed

    def autoPadPlacementCB_Checked(self, sender, e):
        self.padDirectionCtl.IsEnabled = not self.autoPadPlacementCB.IsChecked
        self.padLocationCtl.IsEnabled = not self.autoPadPlacementCB.IsChecked

        if self.lotBoundaryType.SelectedIndex == 0:
            Keyboard.Focus(self.lineCtl)
        else:
            Keyboard.Focus(self.regionLocationCtl)

    def lotBoundaryType_SelectionChanged(self, sender, e):
        index = e.OriginalSource.SelectedIndex
        if index == 0:
            self.lineLabelCtl.Visibility = Visibility.Visible
            self.lineCtl.Visibility = Visibility.Visible
            self.objsLabelCtl.Visibility = Visibility.Collapsed
            self.objs.Visibility = Visibility.Collapsed
            self.regionLocationCtl.Visibility = Visibility.Collapsed
            self.regionLabelCtl.Visibility = Visibility.Collapsed  
        elif index == 1:
            self.lineLabelCtl.Visibility = Visibility.Collapsed
            self.lineCtl.Visibility = Visibility.Collapsed
            self.objsLabelCtl.Visibility = Visibility.Visible
            self.objs.Visibility = Visibility.Visible
            self.regionLocationCtl.Visibility = Visibility.Visible  
            self.regionLabelCtl.Visibility = Visibility.Visible  
            if not self.setbacksDefinedCB.IsChecked:
                self.setbacksDefinedCB.IsChecked = True

class MyCustomCursor(GdiCursorBase):
    def __new__(self, cursorLocation, points, closed, poly2Pol):
        return GdiCursorBase.__new__(self, InputSettingsAdapter())

    def __init__(self, cursorLocation, points, closed, poly2Pol):
        self.cursorLocation = cursorLocation
        self.points = points
        self.closed = closed
        self.poly2Pol = poly2Pol

    def GenerateDrawingInstructions(self, viewCollection, messagingView, mousePosition):
        width = int(self.m_InputSettings.PickAperture * 0.66)
        instru = PolylineDrawingInstruction(messagingView, self.points, self.closed)
        instru.SetComputedColor(self.CursorColor)
        instru.Width = width
        self.PushInstruction(instru.Xor())
        if self.poly2Pol:
            instru = LineDrawingInstruction(messagingView, self.cursorLocation, self.poly2Pol)
            instru.SetComputedColor(self.CursorColor)
            instru.Width = width
            self.PushInstruction(instru.Xor())
