﻿"""
Sample TBC mcro to export the start/end coordinates of a drill pile
"""
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring
from Trimble.Vce.Drillhole import Drillhole

from Trimble.Vce.ForeignCad import TextEntity, MText, Text

from Trimble.Vce.Gem import Model3D

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember
from Trimble.Vce.Interfaces.Point import IPoint
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client

clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, SurfaceTypeLists
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "DrillPlanExport"
    cmdData.CommandName = "DrillPlanExport"
    cmdData.Caption = "_Drill Plan Export"
    cmdData.UIForm = "DrillPlanExport"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxxx"

    try:
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Drill Plan Export"
        cmdData.ToolTipTitle = "Drill Plan Export"
        cmdData.ToolTipTextFormatted = "Export the start/end coordinates of a drill pile."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class DrillPlanExport(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\DrillPlanExport.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        self.dType = clr.GetClrType(Drillhole)
        self.lunits = self.currentProject.Units.Linear
        # we don't want the units to be included
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        # needs to be a point type
        if isinstance(obj, self.dType):
            return True
        return False

    def ConvertToDisplay(self, value):
        # this will convert the number to the current display units (all values
        # are stored as meters)
        return self.lunits.Format(value, self.lfp)

    def OkClicked(self, cmd, e):
        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        wv = self.currentProject[Project.FixedSerial.WorldView]
        layers = self.currentProject[Project.FixedSerial.LayerContainer]


        # points are stored in meters, get conversion factor
        self.unitFactor = self.currentProject.Units.Linear.Convert(1.0, LinearType.Display)
        self.cultureInfo = CultureInfo.CurrentCulture.TextInfo.ListSeparator
        self.coordsInNE = self.currentProject.Units.Coordinate.CoordinateOrder == CoordinateOrder.NEZ

        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.objs.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.objs.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)

        fileName = Path.ChangeExtension(fileName, ".txt")
        if File.Exists(fileName):
            File.Delete(fileName)
        with open(fileName, 'w') as f:   
            self.file = f
            sComma = "\t" # use tabs
            sCsv = "StartX" + sComma + "StartY" + sComma + "StartZ" + sComma + "StartX" + \
                sComma + "StartY" + sComma + "StartZ"
            sCsv = sCsv + "\n"
            self.file.write(sCsv)
            for o in self.objs.SelectedMembers(self.currentProject):
                if not self.IsValidObj(o):
                    continue
                p = o.StartPoint
                sCsv = self.ConvertToDisplay(p.X) + sComma + self.ConvertToDisplay(p.Y) + sComma + self.ConvertToDisplay(p.Z) + sComma

                p = o.EndPoint
                sCsv = sCsv + self.ConvertToDisplay(p.X) + sComma + self.ConvertToDisplay(p.Y) + sComma + self.ConvertToDisplay(p.Z)
                sCsv = sCsv + "\n"
                self.file.write(sCsv)

        cmd.CloseUICommand()

