﻿"""
TBC Macro to explode a IFC shell into lines and points
"""
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from System.IO import StreamReader
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity

from Trimble.DiskIO import OptionsManager
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import SurfaceTypeLists

from Trimble.Vce.Gem import Model3D, ModelBoundaries

from Trimble.Vce.Alignment.Linestring import Linestring,ElementFactory, IStraightSegment, IXYZLocation, IPointIdLocation

from Trimble.Vce.ForeignCad import Point as CadPoint

from Trimble.Vce.Data.Construction.IFC import IFCMesh

from Trimble.Vce.Geometry import Point3D, Plane3D, Vector3D, Line2D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.Region import RegionBuilder
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
try:
    from Trimble.Sdk.UI import UIEventArgs
except:
    from Trimble.Sdk.Interfaces.UI import UIEventArgs
        
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler, ViewHelper

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "ExplodeIFCShell"
    cmdData.CommandName = "ExplodeIFCShell"
    cmdData.Caption = "_Explode IFC Shell"
    cmdData.UIForm = "ExplodeIFCShell"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"

    try:
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "Surface"
        cmdData.ShortCaption = "Explode IFC Shell"

        cmdData.Version = 0.31
        cmdData.MacroAuthor = "Trimble"
        #cmdData.MacroInfo =
        #r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2019/04/19/tml-explode-surface"
        
        cmdData.ToolTipTitle = "Explode IFC Shell"
        cmdData.ToolTipTextFormatted = "Explode a IFC shell into lines and points."
    except:
        pass
    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

    
class ExplodeIFCShell(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\ExplodeIFCShell.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Explode IFC Shell"
        self.tmlFileFolder = macroFileFolder

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption
        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu
        self.mType = clr.GetClrType(IFCMesh)
        self.faceDirectionType.SelectedIndex = OptionsManager.GetInt("ExplodeIFCShell:faceDirectionType", 0)
        self.lineLayer.Text = OptionsManager.GetString("ExplodeIFCShell:lineLayer", "{L}-Lines")
        self.createPoints.IsChecked = OptionsManager.GetBool("ExplodeIFCShell:createPoints", True)

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        return isinstance(o, self.mType)

    def TriangleSlope(self, facePts):
        l = len(facePts)
        # check till we get resonable deflection angle
        bestA = 0
        bestV = Vector3D.Invalid
        for i in range(l - 2):
            v1 = facePts[i] - facePts[i + 1]
            v2 = facePts[i + 1] - facePts[i + 2]
            a = Vector3D.ThreeDimensionsAngleBetween(v1,v2)
            if a > bestA:
                bestA = a
                (v,v1,v2) = Vector3D.CrossProduct(v1,v2)
                v.Normalize()  
                bestV = v
        return bestV

    def CreateLineWithPoints(self, wv, layerSerial, p1, p2):
        l = wv.Add(clr.GetClrType(Linestring))
        e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IPointIdLocation))
        e.LocationSerialNumber = p1 
        l.AppendElement(e)
        e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IPointIdLocation))
        e.LocationSerialNumber = p2
        l.AppendElement(e)       
        l.Layer = layerSerial

    def CreateLine(self, wv, layerSerial, p1, p2):
        l = wv.Add(clr.GetClrType(Linestring))
        e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
        e.Position = p1 
        l.AppendElement(e)
        e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
        e.Position = p2
        l.AppendElement(e)       
        l.Layer = layerSerial

    def IncludeFace(self, v):
        if self.faceDirectionType.SelectedIndex == 0:
            return True # include all
        if self.faceDirectionType.SelectedIndex == 1:
            return v.h > 0
        if self.faceDirectionType.SelectedIndex == 2:
            return v.h < 0
        return Math.Abs(v.h) < 0.000001

    def CheckIfAlreadCreated(self, lines, v1, v2):
        if lines.has_key(v1):
            l = lines[v1]
            if l.has_key(v2):
                return True
            l.Add(v2, v2)
            lines[v1] = l
        else:
            l = {}
            l.Add(v2, v2)
            lines.Add(v1, l)
        return False

    def GetVertexCadPoint(self, v, wv, pnts, transform, vertices, layerSerial):
        if pnts.has_key(v):
            return pnts[v]

        cadPoint = wv.Add(clr.GetClrType(CadPoint))
        p = transform.TransformPoint(vertices[v])
        cadPoint.Point0 = p
        if layerSerial:
            cadPoint.Layer = layerSerial
        # map surface index to serial number
        pnts.Add(v, cadPoint.SerialNumber)
        return cadPoint.SerialNumber

    def MakeVertexInOrder(self, v1, v2):
        if v2 > v1:
            return (v1, v2)
        return (v2, v1)

    def GetLayer(self, layerSerial, layers, layerString):
        layerObj = layers.Lookup(layerSerial)
        if not layerObj:
            return Project.FixedSerial.LayerZero
        newLayerName = layerString.Replace("{L}", layerObj.Name)
        layerNew = layers.GetLayer(newLayerName)
        if layerNew:
            return layerNew.SerialNumber
        # layer doesn't exist
        newLayer = layers.Add(clr.GetClrType(Layer))
        newLayer.Name = newLayerName
        newLayer.DefaultColor = layerObj.DefaultColor
        newLayer.LineWeight = layerObj.LineWeight
        newLayer.LineStyle = layerObj.LineStyle
        newLayer.LayerGroupSerial = layerObj.LayerGroupSerial
        return newLayer.SerialNumber

    def ExplodeMesh(self, mesh, wv, layerSerial):
        createPoints = self.createPoints.IsChecked
        transform = mesh.Transformation
        pnts = {}
        lines = {}
        vertices = mesh.GetVertex()
        faces = mesh.GetFaceList()
        n = len(faces)
        i = 0
        while i < n:
            nV = faces[i]
            i+=1
            if nV < 0: # ignore holes
                i -= nV
                continue

            facePts = []
            for v in range(nV):
                index = faces[i + v]
                facePts.Add(transform.TransformPoint(vertices[index]))
                          
            v = self.TriangleSlope(facePts)
            if self.IncludeFace(v):
                for v in range(nV - 1):
                    i0 = faces[i + v]
                    i1 = faces[i + v + 1]
                    (i0, i1) = self.MakeVertexInOrder(i0, i1)
                    if self.CheckIfAlreadCreated(lines, i0, i1):
                        continue
                    if createPoints:
                        s0 = self.GetVertexCadPoint(i0, wv, pnts, transform, vertices, layerSerial)
                        s1 = self.GetVertexCadPoint(i1, wv, pnts, transform, vertices, layerSerial)
                        self.CreateLineWithPoints(wv, layerSerial, s0, s1)
                    else:
                        self.CreateLine(wv, layerSerial, transform.TransformPoint(vertices[i0]), transform.TransformPoint(vertices[i1]))
                i0 = faces[i + nV - 1]
                i1 = faces[i]
                (i0, i1) = self.MakeVertexInOrder(i0, i1)
                if not self.CheckIfAlreadCreated(lines, i0, i1):
                    if createPoints:
                        s0 = self.GetVertexCadPoint(i0, wv, pnts, transform, vertices, layerSerial)
                        s1 = self.GetVertexCadPoint(i1, wv, pnts, transform, vertices, layerSerial)
                        self.CreateLineWithPoints(wv, layerSerial, s0, s1)
                    else:
                        self.CreateLine(wv, layerSerial, transform.TransformPoint(vertices[i0]), transform.TransformPoint(vertices[i1]))
            i+=nV

    def OkClicked(self, sender, e):           
        wv = self.currentProject[Project.FixedSerial.WorldView]
        layers = self.currentProject[Project.FixedSerial.LayerContainer]
        UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())
        # pause the graphics so every change doesn't draw
        wv.PauseGraphicsCache(True)
        try:
            self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard: 
                for o in self.objs.SelectedMembers(self.currentProject):
                    if isinstance(o, self.mType):
                        layerSerial = self.GetLayer(o.Layer, layers, self.lineLayer.Text)
                        self.ExplodeMesh(o, wv, layerSerial)
                failGuard.Commit()
        finally:
            OptionsManager.SetValue("ExplodeIFCShell:faceDirectionType", self.faceDirectionType.SelectedIndex)
            OptionsManager.SetValue("ExplodeIFCShell:lineLayer", self.lineLayer.Text)
            OptionsManager.SetValue("ExplodeIFCShell:createPoints", self.createPoints.IsChecked)
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
            UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
            wv.PauseGraphicsCache(False)

        Keyboard.Focus(self.objs)

