﻿"""
TBC macro to adjust the global selection depending on if linework "follows" the alignment
"""
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func, DateTime
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys
from System.Windows.Forms import DialogResult

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,ElementFactory, IStraightSegment, IXYZLocation, IPointIdLocation
from Trimble.Vce.ForeignCad import TextEntity, MText, Text
from Trimble.Vce.Gem import Model3D
from Trimble.Vce.Coordinates import PointCollection, NEECoordinate, Point
from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember, TransformData, ResultsOfTransform
from Trimble.Vce.Interfaces.Point import IPoint, IPointLocation
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, SurfaceTypeLists
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "FilterByAlignment"
    cmdData.CommandName = "FilterByAlignment"
    cmdData.Caption = "_Filter Selection By Alignment"
    cmdData.UIForm = "FilterByAlignment"

    try:
        cmdData.Version = 0.61
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Filter By Alignment"
        cmdData.ToolTipTitle = "Filter By Alignment"
        cmdData.ToolTipTextFormatted = 'Adjust the global selection depending on if linework "follows" the alignment.'

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class FilterByAlignment(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\FilterByAlignment.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Filter By Alignment"
        self.tmlFileFolder = macroFileFolder        

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption 
        self.tObj1.AddTypeFilter(clr.GetClrType(IPolyseg))
        self.lType = clr.GetClrType(IPolyseg)
        self.incAlongCB.IsChecked = OptionsManager.GetBool("FilterByAlignment:incAlongCB", True)
        self.maxDeflectionAngle.Angle = OptionsManager.GetDouble("FilterByAlignment:maxDeflectionAngle", Math.PI / 4)

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand() 

    def OkClicked(self, cmd, e):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        h = self.tObj1.SerialNumber
        if not h:
            self.tObj1.StatusMessage = "Select reference geometry"
            Keyboard.Focus(self.tObj1)
            return False
        self.tObj1.StatusMessage = ""

        maxAngle = self.maxDeflectionAngle.Angle
        if Double.IsNaN(maxAngle):
            self.maxDeflectionAngle.StatusMessage = "Enter valid angle"
            return False
        self.maxDeflectionAngle.StatusMessage = ""

        hal = self.tObj1.Entity
        alignPolyseg = hal.ComputePolySeg().Clone()
        alignPolyseg.Extend(10000.0)

        selectionSet = GlobalSelection.Items(self.currentProject)
        foundLines = []
        for obj in selectionSet:
            if not isinstance(obj, self.lType):
                continue
            poly = obj.ComputePolySeg()
            if not poly or poly.NumberOfNodes < 2:
                continue
            (found, pnt, sta)=alignPolyseg.FindPointFromPoint(poly.FirstNode.Point)
            if not found:
                continue
            offset = (poly.FirstNode.Point - pnt).Length2D
            (found, pnt, sta2)=alignPolyseg.FindPointFromPoint(poly.LastNode.Point)
            if not found:
                continue
            offset2 = (poly.LastNode.Point - pnt).Length2D
            dx = abs(sta2 - sta)
            if not dx:
                dx = 0.0001
            dy = abs(offset - offset2)
            ang = Math.Atan2(dy, dx)
            isAlong = ang < maxAngle
            if self.incAlongCB.IsChecked:
                if isAlong:
                    foundLines.Add(obj.SerialNumber)
            else:
                if not isAlong:
                    foundLines.Add(obj.SerialNumber)
        if foundLines:
            GlobalSelection.Items(self.currentProject).Set(foundLines)

        OptionsManager.SetValue("FilterByAlignment:maxDeflectionAngle", self.maxDeflectionAngle.Angle)
        OptionsManager.SetValue("FilterByAlignment:incAlongCB", self.incAlongCB.IsChecked)
        cmd.CloseUICommand()