"""
TBC MAcro to compute surface areas
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult, Control, Keys
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Color
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard


from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode
from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings
from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.ForeignCad import MText

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Triangle3D, Triangle2D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
try:
    from Trimble.Sdk.UI import InputMethod
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod

from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox, DisplayWindow, ExtensionMethods
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice

clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

clr.AddReference("Trimble.Vce.GraphicsEngine2D")
from Trimble.Vce.GraphicsEngine import OverlayBag
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice
from System import Guid
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "FindSurfaceArea"
    cmdData.CommandName = "FindSurfaceArea"
    cmdData.Caption = "_Find Surface Area"
    cmdData.UIForm = "FindSurfaceArea"
    try:
        cmdData.Version = 1.54
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2018/11/04/tml-findsurfacearea"
        cmdData.HelpFile = cmdData.MacroInfo
    except:
        pass

    try:
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "Surface"
        cmdData.ShortCaption = "Find Surface Area"
        cmdData.ToolTipTitle = "Find Surface Area"
        cmdData.ToolTipTextFormatted = "Computes the surface area of trianges bounded by breaklins or texture edges."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class FindSurfaceArea(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\FindSurfaceArea.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroFileFolder = macroFileFolder
        self.macroName = "Find Surface Area"
        self.tmlFileFolder = macroFileFolder

        self.aunits = currentProject.Units.Area
        self.PlanViewOverlay = Guid.NewGuid()
        self.overlayBag = OverlayBag(self.PlanViewOverlay)
        self.limitToBL.IsChecked = OptionsManager.GetBool("FindSurfaceArea:limitToBL", True)
        self.limitToDrapeline.IsChecked = OptionsManager.GetBool("FindSurfaceArea:limitToDrapeline", False)

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption
        types = Array[Type](SurfaceTypeLists.AllWithCutFillMap)
        self.surface.FilterByEntityTypes = types
        self.area = 0
        self.slopeArea = 0
        self.text = ""
        
    def surface_ValueChanged(self, sender, e):
        surfaceModel = self.surface.SelectedEntity()
        if surfaceModel:
            surfaceModel.Gem.FlagTriangles(False)
        if e.Cause == InputMethod.Mouse:     
            Keyboard.Focus(self.coordCtl)

    def coordCtl_ValueChanged(self, sender, e):
        self.shift = Control.ModifierKeys & Keys.Shift
        # run cmd if change was due to mouse pick
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)

    def CancelClicked(self, thisCmd, args):
        thisCmd.CloseUICommand()

    def OkClicked(self, thisCmd, e):
        if self.surface.SelectedSerial == 0:
            self.surface.StatusMessage = "Enter a surface"
            return False

        coord = self.coordCtl.Coordinate
        if not coord.Is3D and not coord.Is2D :
            self.coordCtl.StatusMessage = "No coordinate defined"
            return
        self.coordCtl.StatusMessage = ""

        self.surfaceModel = self.surface.SelectedEntity()
        self.gem = self.surfaceModel.Gem
        self.gem.SetAllBreaklinesAsClipedge()
        locator = self.surfaceModel.GetLocator()
        self.textBoxEntityInfo.Text = ""
        TrimbleOffice.TheOffice.MainWindow.AppViewManager.RemoveOverlayGeometry(self.PlanViewOverlay)

        found,primitive,iTri,iCorner,p = self.surfaceModel.PickSurface(locator, coord)# get point from surface

        if found and not self.gem.GetTriangleFlag(iTri): # ignore if triangle already "flagged"
            iMaterial = self.surfaceModel.GetTriangleMaterial(iTri)
            nTri, isContained = self.gem.FlagTriangleArea(-1, True, iTri, True, self.limitToBL.IsChecked)
                    
            TrimbleOffice.TheOffice.MainWindow.AppViewManager.RemoveOverlayGeometry(self.PlanViewOverlay)
            pts = self.gem.Boundary.ToArray()
            if pts:
                area,slopeArea = self.ComputeSurfaceAreas()
                if self.shift:
                    self.area = self.area + area
                    self.slopeArea = self.slopeArea + slopeArea
                    t = "Total Area: " + self.aunits.Format(self.area) + "\n" + \
                        "Total Surface Area: " + self.aunits.Format(self.slopeArea)
                    self.text = t
                    t =  "\n" + "Plan Area: " + self.aunits.Format(area) + "\n" + \
                        "Surface Area: " + self.aunits.Format(slopeArea) + "\n" + t
                else:
                    t = "Plan Area: " + self.aunits.Format(area) + "\n" + \
                        "Surface Area: " + self.aunits.Format(slopeArea)
                    self.area = area
                    self.slopeArea = slopeArea
                    self.text = t
                    t = "Number of Triangles: " + str(nTri) + "\n" + t

                self.textBoxEntityInfo.Text = t
                if not self.shift:
                    self.overlayBag = OverlayBag(self.PlanViewOverlay)
                self.overlayBag.AddPolygon(0, Color.Red.ToArgb(), Color.Yellow.ToArgb(), pts, pts.Count, 250)
                array = Array[Guid]([DisplayWindow.HoopsPlanViewGUID])
                TrimbleOffice.TheOffice.MainWindow.AppViewManager.AddOverlayGeometry(array, self.overlayBag)
            else:
                self.area = 0
                self.slopeArea = 0
        else:
            self.gem.FlagTriangles(False)

        Keyboard.Focus(self.coordCtl)

    def Dispose(self, thisCmd, disposing):
        OptionsManager.SetValue("FindSurfaceArea:limitToBL", self.limitToBL.IsChecked)
        OptionsManager.SetValue("FindSurfaceArea:limitToDrapeline", self.limitToDrapeline.IsChecked)
        TrimbleOffice.TheOffice.MainWindow.AppViewManager.RemoveOverlayGeometry(self.PlanViewOverlay)

    def ComputeSurfaceAreas(self):
        area = 0
        slopeArea = 0
        for i in range(self.gem.NumberOfTriangles):
            if not self.gem.IsTriangleMaterialPresent(i):
                continue
            if not self.gem.GetTriangleFlag(i):
                continue
            
            p0 = self.gem.GetTrianglePoint(i, 0)
            p1 = self.gem.GetTrianglePoint(i, 1)
            p2 = self.gem.GetTrianglePoint(i, 2)
            rtn = Triangle3D.Area(p0, p1, p2)
            slopeArea = slopeArea + rtn[0]
            rtn = Triangle2D.Area(p0, p1, p2)
            area = area + rtn[0]
            pass
       
        return area,slopeArea

    def createTextBtn_Click(self, sender, e):
        coord = self.coordCtl.Coordinate
        if not coord.Is3D and not coord.Is2D :
            self.coordCtl.StatusMessage = "No coordinate defined"
            return
       
        try:
            form = FindSurfaceAddTextDialog(self).ShowDialog()
        except Exception as ex:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, str(ex))
            return

class FindSurfaceAddTextDialog(Window): # this inherits from the WPF window control
    def __init__(self, cmd):
        with StreamReader(cmd.macroFileFolder + r"\FindSurfaceArea.AddText.xaml") as s:
            wpf.LoadComponent(self, s) 

        self.cmd = cmd
        self.textBoxCtl.Text = self.cmd.text
        self.CancelBtn.Click += self.Button_Click_Cancel
        self.OkBtn.Click += self.Button_Click_Ok

        settings = ConstructionCommandsSettings.ProvideObject(self.cmd.currentProject)
        l = settings.GetUInt32("FindSurfaceArea.Layer", 0)
        if self.cmd.currentProject.Concordance.Lookup(l):
            self.layerCtl.SelectedSerialNumber = l
        l = settings.GetUInt32("FindSurfaceArea.TextStyle", 0)
        if self.cmd.currentProject.Concordance.Lookup(l):
            self.textStyleCtl.SelectedSerialNumber = l

    def Button_Click_Cancel(self, sender, e):
        self.Close()

    def Button_Click_Ok(self, sender, e):
        text = self.textBoxCtl.Text
        if not text:
            return
                
        self.cmd.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.cmd.Caption)
       
        # the "with" statement will unroll any changes if something go wrong
        with TransactMethodCall(self.cmd.currentProject.TransactionCollector) as failGuard:
            wv = self.cmd.currentProject[Project.FixedSerial.WorldView]
            textObj = wv.Add(clr.GetClrType(MText))
            textObj.TextString = text
            textObj.Layer = self.layerCtl.SelectedSnapIn.SerialNumber
            textStyle = self.textStyleCtl.SelectedSnapIn
            textObj.TextStyleSerial = textStyle.SerialNumber
            h = ExtensionMethods.GetWorldHeight(textStyle, self.cmd.currentProject)
            textObj.Height = h
            textObj.Point0 = self.cmd.coordCtl.Coordinate

            # you MUST commit the change to DB or they will all be un-done
            # (inside the "with" scope)
            failGuard.Commit()

            settings = ConstructionCommandsSettings.ProvideObject(self.cmd.currentProject)
            settings.SetUInt32("FindSurfaceArea.Layer", textObj.Layer)
            settings.SetUInt32("FindSurfaceArea.TextStyle", textObj.TextStyleSerial)

            self.cmd.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        self.Close()
        pass

    

