"""
TBC Macro to compute grid volumes from two surfaces and list of boundaries
"""
import os
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func
from System.Windows.Threading import DispatcherPriority
from System.Globalization import CultureInfo

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult, Control, Keys, Application
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Color
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IName
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode
from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings, SurfaceCutFillMap, DtmVolumes, ModelBoundaries
from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.ForeignCad import MText

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Triangle3D, Triangle2D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities

from Trimble.Vce.Data.Construction import SurfaceVolumeGrid
from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
try:
    from Trimble.Sdk.UI import UIEventArgs
except:
    from Trimble.Sdk.Interfaces.UI import UIEventArgs
          
from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox, DisplayWindow, ExtensionMethods

from Trimble.Vce.UI.UIManager import TrimbleOffice

clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

clr.AddReference("Trimble.Vce.GraphicsEngine2D")
from Trimble.Vce.GraphicsEngine import OverlayBag

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
from System import Guid
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "GridVolumeFromBoundaries"
    cmdData.CommandName = "GridVolumeFromBoundaries"
    cmdData.Caption = "_Grid Volume From Boundaries"
    cmdData.UIForm = "GridVolumeFromBoundaries"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"

    try:
        cmdData.Version = 1.05
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""
    except:
        pass   

    try:
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "Surface"
        cmdData.ShortCaption = "Grid Volume"
        cmdData.ToolTipTitle = "Compute Cut/Fill Volume by boundaries"
        cmdData.ToolTipTextFormatted = "Computes grid based volumes."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class MySortObject(object):
    def __init__(self, sta, offset, line):
        self.sta = sta
        self.offset = offset
        self.line = line

    def __cmp__(self, other):
        # if CG withen 1 meter of "other", use offset to sort
        if Math.Abs(self.sta - other.sta) > 1:
            return self.sta.__cmp__(other.sta)
        return self.offset.__cmp__(other.offset)

class GridVolumeFromBoundaries(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\GridVolumeFromBoundaries.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroFileFolder = macroFileFolder
        self.macroName = "Grid Volume From Boundaries"
        self.tmlFileFolder = macroFileFolder
        # get units
        self.lunits = self.currentProject.Units.Linear
        # we don't want the units to be included
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        self.staUnits = self.currentProject.Units.Station
        self.lunits = currentProject.Units.Linear
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        self.aunits = currentProject.Units.Area
        self.vunits = currentProject.Units.Volume
        self.afp = self.aunits.Properties.Copy()
        self.afp.AddSuffix = False
        self.vfp = self.vunits.Properties.Copy()
        self.vfp.AddSuffix = False
        self.cultureInfo = CultureInfo.CurrentCulture.TextInfo.ListSeparator

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption
        self.refSurfaceCtl.FilterByEntityTypes = Array[Type](SurfaceTypeLists.All)
        self.finalSurfaceCtl.FilterByEntityTypes = Array[Type](SurfaceTypeLists.All)
        self.alignmentCtl.FilterByEntityTypes = Array[Type]([clr.GetClrType(ProfileView)])
        self.alignmentCtl.SelectIndex(0) # assume "none"
        self.lType = clr.GetClrType(IPolyseg)
        self.gridSizeCtl.Distance = OptionsManager.GetDouble("GridVolumeFromBoundaries:gridSizeCtl", Double.NaN)
        self.brgCtl.Direction = OptionsManager.GetDouble("GridVolumeFromBoundaries:brgCtl", 0.0)
        self.saveGridsCB.IsChecked = OptionsManager.GetBool("GridVolumeFromBoundaries:saveGridsCB", False)
        self.alignCB.IsChecked = OptionsManager.GetBool("GridVolumeFromBoundaries:alignCB", False)

        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

    def CancelClicked(self, thisCmd, args):
        OptionsManager.SetValue("GridVolumeFromBoundaries:gridSizeCtl", self.gridSizeCtl.Distance)
        OptionsManager.SetValue("GridVolumeFromBoundaries:brgCtl", self.brgCtl.Direction)
        OptionsManager.SetValue("GridVolumeFromBoundaries:saveGridsCB", self.saveGridsCB.IsChecked)
        OptionsManager.SetValue("GridVolumeFromBoundaries:alignCB", self.alignCB.IsChecked)
        thisCmd.CloseUICommand()

    def ComputeFirstLastStation(self, alignmentPolyseg, bndyPolyseg):
        firstSta = 1e20
        lastSta = -1e20
        for node in bndyPolyseg.Nodes():
            rtn = alignmentPolyseg.FindPointFromPoint(node.Point, None, None)
            if rtn[0]:
                station = rtn[4]
                if station < firstSta:
                    firstSta = station
                if station > lastSta:
                    lastSta = station
        return (firstSta, lastSta)

    def OkClicked(self, thisCmd, e):
        if self.refSurfaceCtl.SelectedSerial == 0:
            self.refSurfaceCtl.StatusMessage = "Enter a surface"
            return False
        if self.finalSurfaceCtl.SelectedSerial == 0:
            self.finalSurfaceCtl.StatusMessage = "Enter a surface"
            return False

        rotation = self.brgCtl.Direction
        if Double.IsNaN(rotation):
            self.brgCtl.StatusMessage = "Enter valid grid direction"
            return False
        self.brgCtl.StatusMessage = ""

        gridSize = self.gridSizeCtl.Distance
        if Double.IsNaN(gridSize) or gridSize < 0.1:
            self.gridSizeCtl.StatusMessage = "Enter valid grid size"
            return False
        self.gridSizeCtl.StatusMessage = ""

        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.refSurfaceCtl.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.refSurfaceCtl.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)
        fileName = Path.ChangeExtension(fileName, ".csv")
        if File.Exists(fileName):
            File.Delete(fileName)

        alignment = self.alignmentCtl.SelectedEntity()
        alignmentPolyseg = None
        if alignment != None:
            alignmentPolyseg = alignment.ComputePolySeg()
        existingSurface = self.refSurfaceCtl.SelectedEntity()
        proposedSurface = self.finalSurfaceCtl.SelectedEntity()
        #switch order if surface Classification is set
        (hasOrder,switchOrder)=Model3D.CheckClassificationOrder(existingSurface.Classification, proposedSurface.Classification)
        if switchOrder:
            tmp = existingSurface
            existingSurface = proposedSurface
            proposedSurface = tmp

        ProgressBar.TBC_ProgressBar.Title = self.Caption
        wv = self.currentProject[Project.FixedSerial.WorldView]
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())
        # pause the graphics so every change doesn't draw
        wv.PauseGraphicsCache(True)
        myList = []
        try:
            # the "with" statement will unroll any changes if something go
            # wrong
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                gridSerial = 0
                n = self.objs.SelectedMembers(self.currentProject).Count
                i = 0
                with open(fileName, 'w') as f:   
                    sComma = str(self.cultureInfo) # could be period in some regions
                    q = '"'
                    # write header info
                    sCsv = "Existing surface:" + existingSurface.Name + "\n"
                    f.write(sCsv)
                    sCsv = "Proposed surface:" + proposedSurface.Name + "\n"
                    f.write(sCsv)
                    sCsv = "Object name" + sComma
                    if alignmentPolyseg:
                        sCsv+= "Station" + sComma + "Offset" + sComma
                    sCsv += "Cut Volume" + sComma + "Fill Volume" + sComma + "Net Volume" + sComma + "Ave Depth" + sComma
                    sCsv += "Cut Area" + sComma + "Fill Area" + sComma + "Zero Volume Area" + sComma + "Total Area" + sComma + "Boundary Area" + "\n"
                    f.write(sCsv)
                    volGrid = None
                    for o in self.objs.SelectedMembers(self.currentProject):
                        i+=1
                        if isinstance(o, self.lType):
                            polySeg = o.ComputePolySeg()
                            if polySeg:
                                if polySeg.IsClosed:
                                    name = ""
                                    if isinstance(o, clr.GetClrType(IName)):
                                        name = IName.Name.__get__(o)
                                    # make sure name doesn't have " character
                                    if name.IndexOf(q, 0):
                                        name = name.Replace(q, "_")
                                    if name.IndexOf(sComma, 0):
                                        name = name.Replace(sComma, "_")
                                    
                                    if not volGrid:
                                        volGrid = wv.Add(clr.GetClrType(SurfaceVolumeGrid))
                                        gridSerial = volGrid.SerialNumber
                                    volGrid.InitialSurfaceSerialNumber = existingSurface.SerialNumber
                                    volGrid.FinalSurfaceSerialNumber = proposedSurface.SerialNumber
                                    volGrid.Rotation = rotation
                                    volGrid.Resolution = gridSize
                                    volGrid.ResetGridVolume() # we do this so only one refresh call
                                    boundaryCollection = ModelBoundaries.GetModelBoundaries(self.currentProject, volGrid.SerialNumber, True)
                                    boundaryCollection.Clear()
                                    boundaryCollection.Add(o.SerialNumber)
                                    volGrid.ComputeGridVolume()

                                    area = volGrid.AreaCut + volGrid.AreaFill + volGrid.AreaZero
                                    areaB,cg,dir = polySeg.Area()
                                    volCut = volGrid.VolumesCut
                                    volFill = volGrid.VolumesFill
                                    netVol = volFill - volCut
                                    aveDepth = netVol / areaB
                                    lineV = self.vunits.Format(volCut, self.vfp) + sComma + self.vunits.Format(volFill, self.vfp) + sComma + \
                                        self.vunits.Format(netVol, self.vfp) + sComma + \
                                        self.lunits.Format(aveDepth, self.lfp)
                                    lineA = self.aunits.Format(volGrid.AreaCut, self.afp) + sComma + \
                                        self.aunits.Format(volGrid.AreaFill, self.afp) + sComma + \
                                        self.aunits.Format(volGrid.AreaZero, self.afp) + sComma + \
                                        self.aunits.Format(area, self.afp) + sComma + \
                                        self.aunits.Format(areaB, self.afp)

                                    # kind of a pain to update UI while
                                    # computing.
                                    # Just use dispatcher..
                                    self.textBoxCtl.Text = str(i) + ": " + lineV
                                    def DoNothing():
                                        pass
                                    self.textBoxCtl.Dispatcher.Invoke(DispatcherPriority.Render, Action(DoNothing))

                                    sCsv = q + name + q + sComma
                                    intSta = 0
                                    intOffset = 0
                                    if alignmentPolyseg:
                                        firstSta, lastSta = self.ComputeFirstLastStation(alignmentPolyseg, polySeg)
                                        rtn = alignmentPolyseg.FindPointFromPoint(cg, None, None)
                                        if rtn[0]:
                                            station = self.staUnits.Format(firstSta, alignment) +">"+ \
                                                self.staUnits.Format(lastSta, alignment)
                                            offset = self.lunits.Format(rtn[5], self.lfp)
                                            sCsv += station + sComma + offset + sComma
                                            intSta = int(rtn[4] + 0.5)
                                            intOffset = int(rtn[5]+0.5)

                                    sCsv += lineV + sComma + lineA + "\n"
                                    #f.write(sCsv)
                                    myList.Add(MySortObject(intSta, intOffset, sCsv))

                                    if self.saveGridsCB.IsChecked:
                                        volGrid = None

                                    if ProgressBar.TBC_ProgressBar.Cancelled:
                                        break
                    myList.sort()
                    for o in myList:
                        f.write(o.line);
                failGuard.Commit()
                if not self.saveGridsCB.IsChecked and gridSerial:
                    wv.Remove(gridSerial)
        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
            UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
            wv.PauseGraphicsCache(False)

        self.textBoxCtl.Text = ""
        ProgressBar.TBC_ProgressBar.Title = ""
        # start excel if we can
        try:
            os.system('start excel.exe "%s"' % (fileName,))
        except Exception as ex:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, str(ex))
