﻿"""
TBC Macro to compare one surface to another
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings, GemVertexType, GemCorner

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities

from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "LinestringImport"
    cmdData.CommandName = "LinestringImport"
    cmdData.Caption = "_Linestring Import"
    cmdData.UIForm = "LinestringImport"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"
    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class LinestringImport(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\LinestringImport.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def OkClicked(self, sender, e):
        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.name.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.name.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)

        #fileName = Path.ChangeExtension(fileName, ".xyz")
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        ProgressBar.TBC_ProgressBar.Title = "Computing Linestring"
           
        try:
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                if True:
                    self.ReadLinestringFile(fileName)
                else:
                    self.WriteSurfaceFile(fileName)
                failGuard.Commit()

        except Exception as ex:
            VceMessageBox.ShowError(str(ex), "Error:" + fileName)

        self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        sender.CloseUICommand()

    def WriteSurfaceFile(self, fileName):
        if File.Exists(fileName):
            File.Delete(fileName)
        sComma = ","
        with open(fileName, 'w') as f: 
            nVertices = self.SurfaceModel.NumberOfVertices
            nTri = self.SurfaceModel.NumberOfTriangles
            f.write(str(nVertices) + "\n")
            f.write(str(nTri) + "\n")
            for i in range(nVertices):
                p = self.SurfaceModel.GetVertexPoint(i)
                s = str(i) + sComma + format(p.X, "0.5f") + sComma + format(p.Y, "0.5f") + \
                    sComma + format(p.Z, "0.5f") + "\n"
                f.write(s)
            for i in range(nTri):  
                iVertex0 = self.SurfaceModel.GetTriangleIVertex(i, 0)
                iVertex1 = self.SurfaceModel.GetTriangleIVertex(i, 1)
                iVertex2 = self.SurfaceModel.GetTriangleIVertex(i, 2)

                if self.SurfaceModel.GetTriangleOuterSide(i, 0):
                    iTriangleLeft0 = -1
                else:
                    iTriangleLeft0, corner = self.SurfaceModel.GetTriangleLeftCorner(i, 0)

                if self.SurfaceModel.GetTriangleOuterSide(i, 1):
                    iTriangleLeft1 = -1
                else:
                    iTriangleLeft1, corner = self.SurfaceModel.GetTriangleLeftCorner(i, 1)

                if self.SurfaceModel.GetTriangleOuterSide(i, 2):
                    iTriangleLeft2 = -1
                else:
                    iTriangleLeft2, corner = self.SurfaceModel.GetTriangleLeftCorner(i, 2)

                s = str(i) + sComma + str(iVertex0) + sComma + str(iVertex1) + sComma + \
                    str(iVertex2) + sComma + \
                    str(iTriangleLeft0) + sComma + \
                    str(iTriangleLeft1) + sComma + \
                    str(iTriangleLeft2) + "\n"
                f.write(s)

    def CornerFromVertex(self, triangles, tri,  iv):
        for c in range(3):
            if triangles.VertexIndex(tri, c) == iv:
                return c
        raise Exception("Vertex not in triangle")

    def ReadLinestringFile(self, fileName):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        linestring = wv.Add(clr.GetClrType(Linestring))
        l = None
        with StreamReader(fileName) as s:
            while(True):
                # Read a line.
                line = s.ReadLine()
                # When readline returns an empty string, the file is fully read.
                if line == "" or not line:
                    break

                splits = line.Split(' ')
                if len(splits) != 3:
                    continue
                x = float(splits[0]) / 1000
                y = float(splits[1]) / 1000
                z = float(splits[2]) / 1000
                crd = Point3D(x,y,z)
                if not l:
                    #now create a linestring
                    l = wv.Add(clr.GetClrType(Linestring))
                    e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                    e.Position = crd
                    l.AppendElement(e)                
                else:
                    # check if dup location
                    e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                    e.Position = crd 
                    l.AppendElement(e)
                
            #set name/layer each time to get changes
            l.Name = self.nameCtl.SelectedName
