﻿"""
Sample TBC macro to move selected points to nearest rounded station
"""
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func, DateTime
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys
from System.Windows.Forms import DialogResult

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring

from Trimble.Vce.ForeignCad import TextEntity, MText, Text

from Trimble.Vce.Gem import Model3D

from Trimble.Vce.Coordinates import PointCollection, NEECoordinate, Point

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember, TransformData, ResultsOfTransform
from Trimble.Vce.Interfaces.Point import IPoint, IPointLocation
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces import ProgressBar

clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
try:
    from Trimble.Sdk.UI import UIEventArgs
except:
    from Trimble.Sdk.Interfaces.UI import UIEventArgs
from Trimble.Vce.Core import GlobalSelection     
from Trimble.Vce.UI.Controls import VceMessageBox, SurfaceTypeLists

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "MovePointsToStation"
    cmdData.CommandName = "MovePointsToStation"
    cmdData.Caption = "_Move Points To Station"
    cmdData.UIForm = "MovePointsToStation"

    try:
        cmdData.Version = 0.52
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Move Points To Station"
        cmdData.ToolTipTitle = "Move Points To Station"
        cmdData.ToolTipTextFormatted = "Move points to nearest even station."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class MovePointsToStation(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\MovePointsToStation.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Move Points To Station"
        self.tmlFileFolder = macroFileFolder        

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"

        self.Caption = cmd.Command.Caption 
        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd
        self.objs.IsEntityValidCallback=self.IsValid
        optionMenu = SelectionContextMenuHandler()
        self.pType = clr.GetClrType(IPointLocation)
        
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

        self.stationDeltaCtl.Distance = OptionsManager.GetDouble("MovePointsToStation:stationDeltaCtl", Double.NaN)
        self.staTolCtl.Distance = OptionsManager.GetDouble("MovePointsToStation:staTolCtl", Double.NaN)
        self.lunits = self.currentProject.Units.Linear
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o=self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        # needs to be a point type
        if isinstance(obj, self.pType):
            return True
        return False

    def refAlignment_ValueChanged(self, sender, e):
        if self.refAlignment.SelectedSerial:
            l = self.refAlignment.SelectedEntity()
            self.stationDefinedCtl.StationProvider = l
    
    def moveNearestCtl_Checked(self, sender, e):
        if self.moveNearestCtl.IsChecked:
            self.roundStationLabelCtl.Visibility = Visibility.Visible
            self.stationDeltaCtl.Visibility = Visibility.Visible
            self.stationDefinedLabelCtl.Visibility = Visibility.Collapsed
            self.stationDefinedCtl.Visibility = Visibility.Collapsed
        else:
            self.roundStationLabelCtl.Visibility = Visibility.Collapsed
            self.stationDeltaCtl.Visibility = Visibility.Collapsed
            self.stationDefinedLabelCtl.Visibility = Visibility.Visible
            self.stationDefinedCtl.Visibility = Visibility.Visible

    def Round(self, val, delta):
        if val > 0:
            v =  int((val + delta/2) / delta) * delta
        else:
            v  = int((val - delta/2) / delta) * delta
        return v

    def OkClicked(self, cmd, e):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        self.pointCollection = PointCollection.ProvidePointCollection(self.currentProject)
        h = self.refAlignment.SelectedSerialNumber()
        if not h:
            self.refAlignment.StatusMessage = "Select reference alignment"
            Keyboard.Focus(self.refAlignment)
            return False
        self.refAlignment.StatusMessage = ""

        if self.moveNearestCtl.IsChecked:
            staDelta = self.stationDeltaCtl.Distance
            if Double.IsNaN(staDelta) or staDelta <= 0.0:
                self.stationDeltaCtl.StatusMessage = "Enter valid station rounding"
                return False
            self.stationDeltaCtl.StatusMessage = ""
        else:
            staDelta = self.stationDefinedCtl.Distance
            if Double.IsNaN(staDelta) or staDelta <= 0.0:
                self.stationDefinedCtl.StatusMessage = "Enter valid station location"
                return False
            self.stationDefinedCtl.StatusMessage = ""

        staTol = self.staTolCtl.Distance
        if Double.IsNaN(staTol) or staTol <= 0.0:
            self.staTolCtl.StatusMessage = "Enter valid station tolerance"
            return False
        self.staTolCtl.StatusMessage = ""

        self.hal = self.refAlignment.SelectedEntity()
        self.alignPolyseg = self.hal.ComputePolySeg().Clone()
        self.alignPolyseg.Extend(10000.0)

        UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())
        # pause the graphics so every change doesn't draw
        wv.PauseGraphicsCache(True)
        try:
            self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
            # the "with" statement will unroll any changes if something go wrong
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:   
                staTable = self.hal.StationTable
                numIgnored = 0
                numIgnoredDueToStaEq = 0
                numOutOfTol = 0
                nPts = 0
                totalDist = 0
                maxDist = 0
                for o in self.objs.SelectedMembers(self.currentProject):
                    if not self.IsValidObj(o) or not o.CanTransform:
                        numIgnored += 1
                        continue
                    p = o.Position
                    rtn = self.alignPolyseg.FindPointFromPoint(p, None, None)
                    if not rtn[0]:
                        numIgnored += 1
                        continue
                    station = rtn[4]
                    offset =  rtn[5]                    
                    if self.moveNearestCtl.IsChecked:
                        # adjust for sta-eq
                        s = staTable.DistanceToStation(station)
                        station = self.Round(s.Value, staDelta)
                        s.Value = station
                        if not staTable.IsValid(s, True):
                            numIgnoredDueToStaEq += 1
                            continue
                        station = staTable.StationToDistance(s)
                    else:
                        station = staDelta
                    seg = clr.Reference[Segment]()
                    t = clr.Reference[float]()
                    pcl = clr.Reference[Point3D]()
                    v = clr.Reference[Vector3D]()
                    if not self.alignPolyseg.FindPointFromStation(station, seg, t, pcl, v):
                        numIgnored += 1
                        continue
                    coord = pcl.Value + v.Value * offset
                    coord.Z = 0 # we just want 2D move
                    p.Z = 0
                    dist = (coord - p).Length2D
                    if dist > staTol:
                        numOutOfTol += 1
                        continue
                    nPts += 1
                    totalDist += dist
                    if dist > maxDist:
                        maxDist = dist

                    transformData = TransformData(Matrix4D.BuildTransformMatrix(p, coord, 0, 1, 1, 1), None)
                    o.Transform(transformData)
                            
                failGuard.Commit()

            info = "Points Moved: " + str(nPts)
            if numIgnored:
                info = info + "\rIgnored: " + str(numIgnored)
            if numOutOfTol:
                info = info + "\rOut of Tolerance: " + str(numOutOfTol)
            if numIgnoredDueToStaEq:
                info = info + "\rSta Eq rounding error: " + str(numIgnoredDueToStaEq)

            if nPts:
                info = info + "\rAve distance moved: " +self.lunits.Format(totalDist / nPts, self.lfp)
                info = info + "\rMax distance moved: " +self.lunits.Format(maxDist, self.lfp)
            VceMessageBox.Show(None, info, "Move Points Results")

        finally:
            OptionsManager.SetValue("MovePointsToStation:stationDeltaCtl", self.stationDeltaCtl.Distance)
            OptionsManager.SetValue("MovePointsToStation:staTolCtl", self.staTolCtl.Distance)
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
            UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
            wv.PauseGraphicsCache(False)
        Keyboard.Focus(self.objs)

