﻿"""
Sample TBC macro to compute offset line while elevation defined by slope
"""
#region
import clr
# Reference the WPF assemblies
clr.AddReference('IronPython.Wpf')
import wpf
from System import Type, Tuple, Array, Double, Math, String
from System.IO import StreamReader
from System.Collections.Generic import List
from System.Windows.Controls import StackPanel
from System.Windows import  Visibility
from System.Windows.Input import Keyboard
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Point as GdiPoint

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode, TextJustification
from Trimble.Vce.Core import TransactMethodCall

from Trimble.DiskIO import OptionsManager
from Trimble.Vce.Alignment import SuperelevationView, ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Intersections
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IViewMember, ISettableName, IName, IViewSite, ISnapInHelpers
from Trimble.Vce.Interfaces.Client import CommandGranularity
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, GestureDispatcher
try:
    from Trimble.Sdk.UI import InputMethod, I2DProjection
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod, I2DProjection

from Trimble.Vce.UI.Controls.Adapters import InputSettingsAdapter

from Trimble.Vce.ForeignCad import TextEntity, Leader, MText

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice, UIEvents
clr.AddReference("Trimble.Vce.UI.Wrapper")
from Trimble.Vce.UI.Wrapper import CADWrapperHelper
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
clr.AddReference("Trimble.Vce.ViewModel")
try:
    from Trimble.Vce.ViewModel.CommandLine.CommandLineCommands import OffsetLineCalculator
except:
    clr.AddReference("Trimble.Sdk.UI.Commands")
    from Trimble.Sdk.UI.Commands.CommandLine import OffsetLineCalculator

#endregion
# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "OffsetSlope"
    cmdData.CommandName = "OffsetSlope"
    cmdData.Caption = "_Offset Slope"
    cmdData.UIForm = "OffsetSlope"

    try:
        cmdData.Version = 0.93
        cmdData.MacroAuthor = "Trimble"
        #cmdData.MacroInfo =
        #r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2018/11/29/tml-assignnamefromlayer"
        #cmdData.HelpFile=cmdData.MacroInfo
    except:
        pass

    try:    
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "Surface"
        cmdData.ShortCaption = "Offset Slope"
        cmdData.ToolTipTitle = "Offset Slope"
        cmdData.ToolTipTextFormatted = "Offset a line and set elevation based on slope"

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class OffsetSlope(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\OffsetSlope.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Offset Slope"
        self.tmlFileFolder = macroFileFolder
        

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.cmd = cmd
        self.Caption = cmd.Command.Caption
        self.lType = clr.GetClrType(IPolyseg)
        self.line1Ctl.AddTypeFilter(self.lType)
        self.line1RefCtl.AddTypeFilter(self.lType)

        settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
        l = settings.GetUInt32("OffsetSlope.Layer", 0)
        if self.currentProject.Concordance.Lookup(l):
            self.layerCtl.SelectedSerialNumber = l
        self.line1SlopeOptionCtl.IsChecked = OptionsManager.GetBool("OffsetSlope:line1SlopeOptionCtl", True)
        self.line1VOffsetCtl.Distance = OptionsManager.GetDouble("OffsetSlope:line1VOffsetCtl", 0.0)
        self.line1VOffsetTypeCtl.IsChecked = OptionsManager.GetBool("OffsetSlope:line1VOffsetTypeCtl", False)
        self.line1SlopeOptionCtl_Click(None, None)
        self.line1VOffsetMakeLineCtl.IsChecked = OptionsManager.GetBool("OffsetSlope:line1VOffsetMakeLineCtl", False)

        self.slope1Ctl.Angle = OptionsManager.GetDouble("OffsetSlope:slope1Ctl", Double.NaN)
        self.samplingDistanceCtl.Distance = OptionsManager.GetDouble("OffsetSlope:samplingDistanceCtl", Double.NaN)
        self.offsetDistanceCtl.Distance = OptionsManager.GetDouble("OffsetSlope:offsetDistanceCtl", Double.NaN)

        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def line1Ctl_ValueChanged(self, sender, e):
        if not self.refAlignment.SelectedSerial:
            l = self.line1Ctl.Entity
            self.startStaCtl.StationProvider = l
            self.endStaCtl.StationProvider = l
        if self.line1SlopeOptionCtl.IsChecked and e.Cause == InputMethod.Mouse:
            Keyboard.Focus(self.line1RefCtl)

    def refAlignment_ValueChanged(self, sender, e):
        if self.refAlignment.SelectedSerial:
            l = self.refAlignment.SelectedEntity()
            self.startStaCtl.StationProvider = l
            self.endStaCtl.StationProvider = l

    def OkClicked(self, cmd, e):
        if self.ValidateInput(self.line1Ctl, self.line1SlopeOptionCtl, self.slope1Ctl, self.line1RefCtl, self.line1VOffsetCtl):
            return

        sampDist = self.samplingDistanceCtl.Distance
        if Double.IsNaN(sampDist):
            self.samplingDistanceCtl.StatusMessage = "Enter sampling distance"
            Keyboard.Focus(self.samplingDistanceCtl)
            return

        poly1 = self.line1Ctl.Entity.ComputePolySeg()
        if not poly1 or poly1.NumberOfNodes < 2:
            self.line1Ctl.StatusMessage = "Line not defined"
            return

        offset = self.offsetDistanceCtl.Distance
        if Double.IsNaN(offset):
            self.offsetDistanceCtl.StatusMessage = "Enter offset"
            Keyboard.Focus(self.offsetDistanceCtl)
            return

        begSta = self.startStaCtl.Distance
        if Double.IsNaN(begSta):
            begSta = 0.0
        if begSta < poly1.FirstNode.Station:
            self.startStaCtl.StatusMessage = "Enter start location"
            Keyboard.Focus(self.startStaCtl)
            return
        endSta = self.endStaCtl.Distance
        if Double.IsNaN(endSta):
            endSta = poly1.LastNode.Station
        if endSta <= begSta:
            self.endStaCtl.StatusMessage = "Enter end location"
            Keyboard.Focus(self.endStaCtl)
            return
        
        lineDeltaZ = self.line1VOffsetCtl.Distance
        lineMeasurePerpendicular = self.line1VOffsetTypeCtl.IsChecked
        if self.line1SlopeOptionCtl.IsChecked:
            e = self.line1RefCtl.Entity
            # if object in super view, then it defines the slope
            isSlope = isinstance(e.GetSite(), SuperelevationView)    
            polyRef = e.ComputePolySeg()
            polyRefVert = e.ComputeVerticalPolySeg()
            self.SlopeOptions1 = (True, lineDeltaZ, lineMeasurePerpendicular, isSlope, polyRef, polyRefVert)
        else:
            self.SlopeOptions1 = (False, lineDeltaZ, lineMeasurePerpendicular, Math.Tan(Math.PI / 2.0 - self.slope1Ctl.Angle))


        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        try:
            self.debug = self.nameCtl.SelectedName.StartsWith("__@d")
            if self.debug:
                self.textBoxEntityInfo.Visibility = Visibility.Visible
                self.textBoxEntityInfo.Text = ""

            refHalPoly = None
            if self.refAlignment.SelectedSerial:
                refHalPoly = self.refAlignment.SelectedEntity().ComputePolySeg()
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                self.ComputeOffsetSlopeIntersect(refHalPoly, offset, poly1, self.line1Ctl.Entity.ComputeVerticalPolySeg(), 
                        begSta, endSta, sampDist)
                failGuard.Commit()
            settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
            settings.SetUInt32("OffsetSlope.Layer", self.layerCtl.SelectedSerialNumber)

            OptionsManager.SetValue("OffsetSlope:line1SlopeOptionCtl", self.line1SlopeOptionCtl.IsChecked)
            OptionsManager.SetValue("OffsetSlope:line1VOffsetCtl", self.line1VOffsetCtl.Distance)
            OptionsManager.SetValue("OffsetSlope:line1VOffsetTypeCtl", self.line1VOffsetTypeCtl.IsChecked)

            OptionsManager.SetValue("OffsetSlope:line1VOffsetMakeLineCtl", self.line1VOffsetMakeLineCtl.IsChecked)

            OptionsManager.SetValue("OffsetSlope:slope1Ctl", self.slope1Ctl.Angle)
            OptionsManager.SetValue("OffsetSlope:samplingDistanceCtl", self.samplingDistanceCtl.Distance)
            OptionsManager.SetValue("OffsetSlope:offsetDistanceCtl", self.offsetDistanceCtl.Distance)

        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
        Keyboard.Focus(self.line1Ctl)

    def ValidateInput(self, lineCtl, option, slopeCtl, refLineCtl, lineVOffsetCtl):
        line = lineCtl.Entity
        if not line:
            lineCtl.StatusMessage = "No line selected"
            Keyboard.Focus(lineCtl)
            return True
        lineCtl.StatusMessage = ""
        refLineCtl.StatusMessage = ""
        slopeCtl.StatusMessage = ""
        if option.IsChecked:
            line = refLineCtl.Entity
            if not line:
                refLineCtl.StatusMessage = "No line selected"
                Keyboard.Focus(refLineCtl)
                return True
        else:
            v = slopeCtl.Angle
            if Double.IsNaN(v):
                slopeCtl.StatusMessage = "Enter valid slope"
                Keyboard.Focus(slopeCtl)
                return True
        vo = lineVOffsetCtl.Distance
        if Double.IsNaN(vo):
            lineVOffsetCtl.StatusMessage = "Enter vertical offset distance"
            Keyboard.Focus(lineVOffsetCtl)
            return True
        return False

    def line1SlopeOptionCtl_Click(self, sender, e):
        if self.line1SlopeOptionCtl.IsChecked:
            self.slope1NameLabel.Visibility = Visibility.Collapsed
            self.slope1Ctl.Visibility = Visibility.Collapsed
            self.line1RefLabel.Visibility = Visibility.Visible
            self.line1RefCtl.Visibility = Visibility.Visible
        else:
            self.slope1NameLabel.Visibility = Visibility.Visible
            self.slope1Ctl.Visibility = Visibility.Visible
            self.line1RefLabel.Visibility = Visibility.Collapsed
            self.line1RefCtl.Visibility = Visibility.Collapsed
    
    def endStaCtl_ValueChanged(self, sender, e):
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)

# methods to calculate slope intersect
    def ComputeOffsetSlopeIntersect(self, refHalPoly, offset, poly1, vPoly1, begSta, endSta, deltaSta):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        if refHalPoly:
            if begSta < refHalPoly.FirstNode.Station:
                begSta = refHalPoly.FirstNode.Station
            if endSta > refHalPoly.LastNode.Station:
                endSta = refHalPoly.LastNode.Station
        else:
            if begSta < poly1.FirstNode.Station:
                begSta = poly1.FirstNode.Station
            if endSta > poly1.LastNode.Station:
                endSta = poly1.LastNode.Station
        pts = List[Point3D]() 
        pts1 = List[Point3D]() 
        pts1R = List[Point3D]() 
        if deltaSta > 0.01:
            sta = begSta
            while sta <= endSta:
                # if near end, set to end
                if Math.Abs(endSta - sta) < deltaSta / 2:
                    sta = endSta
                staToUse = sta
                if False and not refHalPoly:
                    # try to adj stations (not used now)
                    seg = clr.Reference[Segment]()
                    t = clr.Reference[float]()
                    pcl = clr.Reference[Point3D]()
                    v = clr.Reference[Vector3D]()
                    if poly1.FindPointFromStation(sta, seg, t, pcl, v):
                        t = t.Value
                        if t < 0.5 and t > 0:
                            if t * seg.Value.Length2D < Math.Abs(offset):
                                staToUse = seg.Value.BeginStation
                        if t > 0.5 and t < 1:
                            if (1.0-t) * seg.Value.Length2D < Math.Abs(offset):
                                staToUse = seg.Value.EndStation

                (p, pcl1, pRef1) = self.ComputeOffsetSlopeIntersectAtStation(refHalPoly, staToUse, offset, poly1, vPoly1)
                if p.Is3D:
                    if p.Z < -1.0e10:
                        break
                    pts.Add(p)
                    pts1.Add(pcl1)
                    pts1R.Add(pRef1)
                elif pts and not self.debug:
                    # if we can't find solution and we did before, stop
                    break            
                sta += deltaSta

                # if near end, set to end
                if Math.Abs(endSta - sta) < deltaSta / 2:
                    sta = endSta
        else:
            # compute at poly1 nodes
            for node in poly1.Nodes():
                if refHalPoly:
                    (found, pnt, sta)=refHalPoly.FindPointFromPoint(node.Point)
                    if not found:
                        if pts:
                            break
                        continue
                else:
                    sta = node.Station
                if sta < begSta:
                    continue
                if sta > endSta:
                    break
                (p, pcl1, pRef1) = self.ComputeOffsetSlopeIntersectAtStation(refHalPoly, sta, offset, poly1, vPoly1)
                if p.Is3D:
                    if p.Z < -1.0e10:
                        break
                    pts.Add(p)
                    pts1.Add(pcl1)
                    pts1R.Add(pRef1)
                elif pts and not self.debug:
                    # if we can't find solution and we did before, stop
                    break    
        # check if any points computed
        if not pts:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute.")
            return False

        side = Side.Right
        if offset < 0:
            offset = -offset
            side = Side.Left
        polyseg = OffsetLineCalculator.ComputeOffsetPolyseg(self.line1Ctl.Entity, offset, side, False, begSta, endSta, 1.0)
        polyseg.Scale(Vector3D(1.0, 1.0))
        ls = wv.Add(clr.GetClrType(Linestring))
        ls.Layer = self.layerCtl.SelectedSerialNumber 
        ls.Name = self.nameCtl.SelectedName
        ls.Append(polyseg, None, False, False)
        lastSta = -1
        for p in pts:
            (found, pnt, sta)=polyseg.FindPointFromPoint(p)
            if found:
                if abs(lastSta-sta) > Point3D.Tolerance:
                    e = VerticalElementFactory.Create(clr.GetClrType(IVerticalNoCurve), clr.GetClrType(IStationVerticalLocation))
                    e.StationDistance = sta
                    e.Elevation = p.Z
                    ls.AddVerticalElement(e)
                lastSta = sta
        # check if we offset the segment
        if self.SlopeOptions1[1] and self.line1VOffsetMakeLineCtl.IsChecked:
            if offset:
                polyseg = PolySeg()
                polyseg.Add(pts1)
                ls = wv.Add(clr.GetClrType(Linestring))
                ls.Layer = self.layerCtl.SelectedSerialNumber 
                ls.Name = self.nameCtl.SelectedName
                ls.Append(polyseg, None, False, False)

            if self.SlopeOptions1[0]:
                polyseg = PolySeg()
                polyseg.Add(pts1R)
                ls = wv.Add(clr.GetClrType(Linestring))
                ls.Layer = self.layerCtl.SelectedSerialNumber 
                ls.Name = self.nameCtl.SelectedName
                ls.Append(polyseg, None, False, False)

    def ComputeOffsetSlopeIntersectAtStation(self, refHalPoly, staOnRefAlignment, offset, poly1, vPoly1):
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        if refHalPoly:
            if not refHalPoly.FindPointFromStation(staOnRefAlignment, seg, t, pcl, v):
                return (Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
            v1 = v.Value
            v1.To2D()
            v1.Normalize()
            pclOn1 = pcl.Value
            (pclOn1, staOn1, side) = self.VectorIntersectLine(pclOn1, v1, poly1, None)
        else:
            staOn1 = staOnRefAlignment
            if not poly1.FindPointFromStation(staOn1, seg, t, pcl, v):
                return (Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
            v1 = v.Value
            v1.To2D()
            v1.Normalize()
            pclOn1 = pcl.Value
           
        (el, pcl1, pRef1) = self.ComputeOffsetElev(staOn1, pclOn1, offset, vPoly1, self.SlopeOptions1, v1)
        if Double.IsNaN(el):
            return (Point3D(0, 0, -1.1e10), Point3D.Undefined, Point3D.Undefined)
        pNew = pclOn1 + v1 * offset
        pNew.Z = el
        return (pNew, pcl1, pRef1)

    def ComputeOffsetElev(self, sta, pcl, offset, polyVert, slopeOptions, v):
        z = pcl.Z
        pRef = Point3D.Undefined
        if polyVert:
            (found, segment, p, isOn) = polyVert.LocateX(Point3D(sta, 0.0))
            if found:
                z = p.Y
        if Double.IsNaN(z):
            return (z, pcl, pRef)
        slope = Double.NaN
        zAdj = slopeOptions[1]
        adjPerp = slopeOptions[2]
        if slopeOptions[0]:

            if slopeOptions[3]:
                # slope table used
                (found, segment, p, isOn) = slopeOptions[4].LocateX(Point3D(sta, 0.0))
                if found:
                    slope = p.Y / 100
            else:
                # find intersection of vector from line to ref line
                (pRef,s, side) = self.VectorIntersectLine(pcl, v, slopeOptions[4], slopeOptions[5])
                (dist, p, p) = Line2D.Length(pcl, pRef)
                if dist:
                    if side == Side.Left:
                        dist = -dist
                    slope = (pRef.Z - z) / dist
        else:
            slope = slopeOptions[3]
           
        if Double.IsNaN(slope):
            return (z, pcl, pRef)
        
        deltaZ = offset * slope
        # check if we should ofset the line
        pcl_ = Point3D(pcl)
        pcl_.Z  = z
        if zAdj:
            if adjPerp:
                zAdj = zAdj / Math.Cos(Math.Atan(slope))
            deltaZ += zAdj
            pcl_.Z += zAdj
            if pRef.Is3D:
                pRef.Z += zAdj
        z = z + deltaZ
        return (z, pcl_, pRef)

    def VectorIntersectLine(self, fromPoint, directionVector, polyToIntersect, polyToIntersectVert):
        acrossPoly = PolySeg()
        acrossPoly.Add(Array[Point3D]([fromPoint - 1000.0 * directionVector, fromPoint + 1000.0 * directionVector]))
        ints = Intersections()
        if not polyToIntersect.Intersect(acrossPoly, True, ints):
            return (Point3D.Undefined, Double.NaN, Side.On)
        # we could get more than one intersection.  
        # Find nearest to middle (t value of 0.5)
        bestT = 1e10
        bestI = 0
        for i in range(ints.Count):
            if Math.Abs(ints[i].T2 - 0.5) < bestT:
                bestT = ints[i].T2
                bestI = i
        i = ints[bestI]
        staRef = i.Segment1.Station(i.T1)
        (ret, pRef) = i.Segment1.ComputePoint(i.T1)
        # check if ref line has vertical polyseg
        if polyToIntersectVert:
            (found, segment, p, isOn) = polyToIntersectVert.LocateX(Point3D(staRef, 0.0))
            if found:
                pRef.Z = p.Y
        s = Side.Left
        if i.T2 > 0.5:
            s = Side.Right
        return (pRef, staRef, s)