"""
TBC Macro to show how to create or insert a CAD block
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult, Control, Keys
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Color
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Core.Components import BlockView

from Trimble.Vce.ForeignCad import MText, PolyLine, BlockReference

try:
    from Trimble.Sdk.UI import InputMethod
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod

from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox, DisplayWindow, ExtensionMethods
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import ViewHelper
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice
from System import Guid
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "PlaceBlock"
    cmdData.CommandName = "PlaceBlock"
    cmdData.Caption = "_Place Block"
    cmdData.UIForm = "PlaceBlock"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"

    try:
        cmdData.Version = 1.01
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""
    except:
        pass  

    try: 
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Place Block"
        cmdData.ToolTipTitle = "Place Plock"
        cmdData.ToolTipTextFormatted = "Inserts a block at a user defined location."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class PlaceBlock(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\PlaceBlock.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroFileFolder = macroFileFolder

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        self.blockCollection = self.currentProject[Project.FixedSerial.BlockCollection]
        self.CreateSampleBlock()
        self.block.SearchContainer = self.blockCollection.SerialNumber
        self.block.UseSelectionEngine = False
        types = Array[Type]([clr.GetClrType(BlockView)])
        self.block.FilterByEntityTypes = types

    def coordCtl_ValueChanged(self, sender, e):
        # run cmd if change was due to mouse pick
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)

    def CancelClicked(self, thisCmd, args):
        thisCmd.CloseUICommand()

    def OkClicked(self, thisCmd, e):
        if self.block.SelectedSerial == 0:
            self.block.StatusMessage = "Select a block"
            return False

        coord = self.coordCtl.Coordinate
        if not coord.Is3D and not coord.Is2D :
            self.coordCtl.StatusMessage = "No coordinate defined"
            return
        self.coordCtl.StatusMessage = ""

        block = self.block.SelectedEntity()
        self.currentProject.TransactionManager.AddBeginMark(Client.CommandGranularity.Command, self.Caption)
        try:
            # unroll any changes if something goes wrong
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                container = ViewHelper.GetContainerForView(self.coordCtl.ClickWindow)
                blk = container.Add(clr.GetClrType(BlockReference))
                blk.InsertionPoint = coord
                blk.BlockSerial = block.SerialNumber

                # you MUST commit the change or they will all be un-done
                failGuard.Commit()
        finally:
            self.currentProject.TransactionManager.AddEndMark(Client.CommandGranularity.Command)

        Keyboard.Focus(self.coordCtl)

    def CreateSampleBlock(self):
        if self.blockCollection.Contains("TestBlock"):
            return
        # To create block, add container (BlockView) to collection and
        # add typical cad objects to collection. Remember, the
        # coordinates are relative to insertion point (so near 0,0)
        newBlock = self.blockCollection.Add(clr.GetClrType(BlockView))
        newBlock.Name = "TestBlock"

        # now add objects to block
        pline = newBlock.Add(clr.GetClrType(PolyLine))
        n = PolyLine.PolyLineNode()
        n.X = 0
        n.Y = 0
        pline.AddNode(n)
        n.X = 1.0
        n.Y = 0
        pline.AddNode(n)
        n.X = 1.0
        n.Y = 1.0
        pline.AddNode(n)
        n.X = 0
        n.Y = 1.0
        pline.AddNode(n)
        pline.Closed = True



    

