import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Uri
from System.Windows.Controls import StackPanel, Image
from System.Windows.Media.Imaging import BitmapImage, BitmapCacheOption
from System.Windows.Media  import Stretch 
from System.IO import StreamReader, Path, File, FileMode, MemoryStream, FileStream, FileAccess
from System.Windows.Input import Keyboard
from System import Type, Tuple, Array, Double
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle, BlockCollection, BlockView

from Trimble.Vce.Gem import Model3D, ModelBoundaries

from Trimble.Vce.Data.Scanning import PointCloudRegion
try:
    oldVersion = False
    from Trimble.Vce.Data.Scanning import PointCloudDatabaseContainer, ExposedPointCloudRegion
except:
    oldVersion = True
    from Trimble.Vce.Data.Scanning import PointCloudRegionContainer, DefaultPointCloudRegion
    from Trimble.Vce.Scanning import ClippingManager

clr.AddReference("Sde.Net")
from Trimble.Sde import Cloud
from Trimble.Vce.Geometry import Point3D, Intersections
from Trimble.Vce.Geometry.PolySeg import PolySeg

from Trimble.Vce.Interfaces import Client

from Trimble.Vce.UI.Controls import VceMessageBox
from Trimble.Vce.UI.Controls.Wpf import CoordinateEdit, DistanceEdit
clr.AddReference("Trimble.Vce.UI.Hoops")
from Trimble.Vce.UI.Hoops import HoopsBlockEditorView
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "PointCloudRegion"
    cmdData.Caption = "_Point Cloud Region"
    cmdData.UIForm = "PointCloudRegionUI"

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class PointCloudRegionUI(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\PointCloudRegion.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroFolder = macroFileFolder
        if oldVersion:
            rc = PointCloudRegionContainer.GetExposedUniqueInstance(self.currentProject)
        else:
            rc =  PointCloudDatabaseContainer.GetInstance(self.currentProject)
        #self.regions.SearchContainer = rc.SerialNumber;
        #self.regions.UseSelectionEngine = False
        self.regions.AllowNone = True
        # this allows both the default and user created regions to be listed.
        if oldVersion:
            self.regions.SetEntityType(Array[Type]([clr.GetClrType(PointCloudRegion),
                     clr.GetClrType(DefaultPointCloudRegion)]), currentProject)
        else:
            self.regions.SetEntityType(Array[Type]([clr.GetClrType(PointCloudRegion),
                 clr.GetClrType(ExposedPointCloudRegion)]), currentProject)
        self.regions.SelectIndex(0) # start with "none"

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption

    def region_ValueChanged(self, sender, e):
        pass

    def CancelClicked(self, thisCmd, args):
        thisCmd.CloseUICommand()

    def OkClicked(self, thisCmd, e):
        r = self.regions.SelectedEntity()
        if r:
            self.currentProject.TransactionManager.AddBeginMark(Client.CommandGranularity.Command, self.Caption)
            wv = self.currentProject[Project.FixedSerial.WorldView]
            surName = r.Name
            surName = Model3D.GetUniqueName(surName, None, wv) #make sure name is unique
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                if self.createSurfaceCtl.IsChecked:
                    newSurface = wv.Add(clr.GetClrType(Model3D))
                    newSurface.Name = surName
                    builder = newSurface.GetGemBatchBuilder()
                else:
                    builder=None
                try:  
                    if oldVersion:
                        t = str(r.GetNumberOfPoints())
                        ClippingManager.PauseOn()
                        sde = r.SDECloud # the ID in the point cloud DB
                        if not sde:
                            return
                        with Cloud(sde) as cloud:
                            # get the points...
                            count = 0
                            for p in cloud:
                                count += 1
                                if count > 10000: # limit to fewer points for testing....
                                    break
                                if builder:
                                    builder.AddVertex(Point3D(p.Coordinates.X, p.Coordinates.Y, p.Coordinates.Z))
                                t = t + "\n" + str(p.Coordinates.X) + " " +  str(p.Coordinates.Y) + " " +  str(p.Coordinates.Z)
                    else:
                        # new version
                        t = str(r.Integration.PointCount)
                        count = 0
                        for p in r.Integration.GetPoints():
                            count += 1
                            if count > 10000: # limit to fewer points for testing....
                                break
                            if builder:
                                builder.AddVertex(Point3D(p.X, p.Y, p.Z))
                            t = t + "\n" + str(p.X) + " " +  str(p.Y) + " " +  str(p.Z)

                    self.regionCount.Text = t
                    if builder:
                        builder.Construction()
                        builder.Commit()
                    
                    failGuard.Commit()

                except Exception as ex:
                    VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, str(ex))
                    return
                finally:
                    if oldVersion:
                        ClippingManager.PauseOff()
            self.currentProject.TransactionManager.AddEndMark(Client.CommandGranularity.Command)

            # can now get triangle info from surface.
        else:
            self.regionCount.Text = ""