﻿"""
Sample TBC macro to measure the distance from a point to nearest vertex in point cloud
"""
import os
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func, DateTime
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys
from System.Windows.Forms import DialogResult

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring

from Trimble.Vce.ForeignCad import TextEntity, MText, Text
from Trimble.Vce.Gem import Model3D

from Trimble.Vce.Coordinates import PointCollection, NEECoordinate, Point

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember, TransformData, ResultsOfTransform
from Trimble.Vce.Interfaces.Point import IPoint, IPointLocation
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces import ProgressBar, Point as PointHelper
from Trimble.Vce.Graphics import LeftMouseModeType
from Trimble.Vce.Units import AngularFormatProperties
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, SurfaceTypeLists
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
clr.AddReference("Sde.Net")
from Trimble.Sde import Cloud
from Trimble.Vce.Data.Scanning import PointCloudRegion
try:
    oldVersion = True
    from Trimble.Vce.Data.Scanning import DefaultPointCloudRegion    
except:
    oldVersion = False

clr.AddReference("Trimble.Vce.UI.Hoops")
from Trimble.Vce.UI.Hoops import Hoops2dView
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice

#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "PointsToCloudReport"
    cmdData.CommandName = "PointsToCloudReport"
    cmdData.Caption = "_Point To Cloud Report"
    cmdData.UIForm = "PointsToCloudReport"

    try:
        cmdData.Version = 0.63
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Points To Cloud"
        cmdData.ToolTipTitle = "Measure Points To Point Cloud"
        cmdData.ToolTipTextFormatted = "Measure distance from point to nearest vertex in point cloud."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class PointsToCloudReport(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\PointsToCloudReport.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Points To Cloud"
        self.tmlFileFolder = macroFileFolder        

    def OnLoad(self, cmd, buttons, event):
        self.okBtn = buttons[0]
        self.Caption = cmd.Command.Caption 
        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        self.pType = clr.GetClrType(IPointLocation)
        self.regionType = clr.GetClrType(PointCloudRegion)  
        self.pointCollection = PointCollection.ProvidePointCollection(self.currentProject)
        
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

        self.searchTolCtl.Distance = OptionsManager.GetDouble("PointsToCloudReport:searchTolCtl", Double.NaN)
        self.numPtsToUseCtl.Value = OptionsManager.GetInt("PointsToCloudReport:numPtsToUseCtl", 1)
        self.nameCtl.SelectedName = OptionsManager.GetString("PointsToCloudReport.Name", "1")
        self.lunits = self.currentProject.Units.Linear
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        self.aunits = self.currentProject.Units.Azimuth
        self.afp = self.aunits.Properties.Copy()
        self.afp.AddSuffix = False
        self.afp.DMSRoundingMode = AngularFormatProperties.DMSRounding.NearestDegree
        self.afp.Show0Seconds = False
        self.afp.Show0Minutes = False
        self.cultureInfo = CultureInfo.CurrentCulture.TextInfo.ListSeparator
        self.coordsInNE = self.currentProject.Units.Coordinate.CoordinateOrder == CoordinateOrder.NEZ
        self.unitFactor = self.currentProject.Units.Linear.Convert(1.0, LinearType.Display)

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        # needs to be a point type
        if isinstance(obj, self.pType):
            return True
        return False

    def ConvertToDisplay(self, value):
        # this will convert the number to the current display units (all values
        # are stored as meters)
        return self.lunits.Format(value, self.lfp)

    def OkClicked(self, cmd, e):
        Keyboard.Focus(self.okBtn)
        activeForm = TrimbleOffice.TheOffice.MainWindow.AppViewManager.ActiveView
        if not isinstance(activeForm, clr.GetClrType(Hoops2dView)):
            self.objs.StatusMessage = "Active view must be a 2D view"
            return
        if not activeForm.View.LeftMouseOperation == LeftMouseModeType.PolygonSelection:
            self.objs.StatusMessage = "You must be in 'Polygon Select mode'"
            return

        wv = self.currentProject[Project.FixedSerial.WorldView]
        self.pointCollection = PointCollection.ProvidePointCollection(self.currentProject)

        searchTol = self.searchTolCtl.Distance
        if Double.IsNaN(searchTol) or searchTol <= 0.0:
            self.searchTolCtl.StatusMessage = "Enter valid search tolerance"
            return False
        self.searchTolCtl.StatusMessage = ""
       
        self.currentProject.TransactionManager.AddBeginMark(Client.CommandGranularity.Command, self.Caption)
        ProgressBar.TBC_ProgressBar.Title = "Finding scan points"
        fileName = Path.Combine(SearchPath.ExportAppDataPath,"PointsToCloud.csv")
        if File.Exists(fileName):
            try:
                File.Delete(fileName)
            except:
                VceMessageBox.ShowWarning(TrimbleOffice.TheOffice.MainWindow.Form, "Could not delete file. Close Excel if file open", "Delete Existing File")
                return
        with open(fileName, 'w') as f:
            sComma = str(self.cultureInfo) # could be period in some regions
            q = '"'
            # write header info
            sCsv = "Point ID" + sComma
            if self.coordsInNE:
                sCsv += "North" + sComma + "East" + sComma + "Elevation" + sComma
            else:
                sCsv += "East" + sComma + "North" + sComma + "Elevation" + sComma
            sCsv += "Num Scan Pts" + sComma + "Delta XY" + sComma + "Delta Z" + sComma + "Direction" + "\n"
            f.write(sCsv)

             # gete the current list so we can change global selection
            entities = Array[ISnapIn](self.objs.SelectedMembers(self.currentProject))
            count = entities.Count
            i = 1
            for o in entities:
                percent = i * 100 // count
                i += 1
                if ProgressBar.TBC_ProgressBar.SetProgress(percent):
                    break
                if not self.IsValidObj(o):
                    continue
                p = IPointLocation.Position.__get__(o)
                if p.Is2D:
                    p.Z = 0.0
                name = ""
                if isinstance(o, clr.GetClrType(IName)):
                    name = o.Name
                # make sure name doesn't have " character
                if name.IndexOf(sComma, 0) >= 0:
                    name = name.Replace(sComma, "_")
                if self.coordsInNE:
                    sCsv = name + sComma + self.ConvertToDisplay(p.Y) + sComma + self.ConvertToDisplay(p.X)
                else:
                    sCsv = name + sComma + self.ConvertToDisplay(p.X) + sComma + self.ConvertToDisplay(p.Y)
                if Double.IsNaN(p.Z):
                    sCsv += sComma
                else:
                    sCsv += sComma + self.ConvertToDisplay(p.Z)

                ptsList = List[Point3D]() # we need a generic list, not an array
                newPoint = Point3D(p.X - searchTol, p.Y - searchTol, 0)
                ptsList.Add(newPoint)
                newPoint = Point3D(p.X + searchTol, p.Y - searchTol, 0)
                ptsList.Add(newPoint)
                newPoint = Point3D(p.X + searchTol, p.Y + searchTol, 0)
                ptsList.Add(newPoint)
                newPoint = Point3D(p.X - searchTol, p.Y + searchTol, 0)
                ptsList.Add(newPoint)
                newPoint = Point3D(p.X - searchTol, p.Y - searchTol, 0)
                ptsList.Add(newPoint)

                op = activeForm.View.CurrentOperator # should be PolygonSelectionOperator
                op.PolygonStarted = True
                op.PolygonPoints = ptsList
                activeForm.View.PolygonSelectionStarted = True # need to fake a mouse gesture
                op.PerformSelection()         
                selectionSet = GlobalSelection.Items(self.currentProject)
                numScanPts = 0
                foundPoints = []
                for obj in selectionSet:
                    if not isinstance(obj, self.regionType):
                        continue
                    if oldVersion:
                        id = obj.GetSelectedCloudId()
                        if id:
                            with Cloud(id) as cloud:
                                # get the points...
                                for v in cloud:
                                    vp = Point3D(v.Coordinates.X, v.Coordinates.Y, v.Coordinates.Z)
                                    vector = vp - p
                                    dist = vector.Length2D
                                    if dist <= searchTol:
                                        numScanPts += 1
                                        if numScanPts > 1000:
                                            break
                                        dz = vp.Z - p.Z
                                        foundPoints.Add((dist, vector.Angle, dz))
                    else:
                        for v in obj.Integration.GetPointInfos():
                            vp = Point3D(v.Coordinates.X, v.Coordinates.Y, v.Coordinates.Z)
                            vector = vp - p
                            dist = vector.Length2D
                            if dist <= searchTol:
                                numScanPts += 1
                                if numScanPts > 1000:
                                    break
                                dz = vp.Z - p.Z
                                foundPoints.Add((dist, vector.Angle, dz))

                if not numScanPts:
                    # no scan points found
                    sCsv += sComma + str(numScanPts) + "\n"
                    f.write(sCsv)
                elif numScanPts > 1000:
                    sCsv += "More than 1000 scan points found. Reduce search tolerance"
                    f.write(sCsv)
                else:
                    def getKey(item):
                        return -item[0] # use distance to sort

                    foundPoints.sort(key=getKey)
                    foundPoints.reverse()
                    if self.createPointsCB.IsChecked:
                        (dist, az, dz) = foundPoints[0]
                        p.Z = dz
                        self.CreatePoint(p)

                    n = self.numPtsToUseCtl.Value
                    if n == 1:
                        (dist, az, dz) = foundPoints[0]
                        az = self.aunits.Format(az, self.afp)
                        sCsv += sComma + str(numScanPts) + sComma + self.ConvertToDisplay(dist) + sComma + self.ConvertToDisplay(dz) + sComma + az + "\n"
                        f.write(sCsv)
                    else:
                        isFirst = True
                        i = 0
                        for (dist, az, dz) in foundPoints:
                            az = self.aunits.Format(az, self.afp)
                            if isFirst:
                                sCsv += sComma + str(numScanPts) + sComma + self.ConvertToDisplay(dist) + sComma + self.ConvertToDisplay(dz) + sComma + az + "\n"
                                f.write(sCsv)
                                isFirst = False
                            else:
                                sCsv = sComma + sComma + sComma + sComma + sComma + self.ConvertToDisplay(dist) + sComma + self.ConvertToDisplay(dz) + sComma + az + "\n"
                                f.write(sCsv)
                            i+=1
                            if i >= n:
                                break
                        f.write("\n")

            ProgressBar.TBC_ProgressBar.Title = ""
            self.currentProject.TransactionManager.AddEndMark(Client.CommandGranularity.Command)
            OptionsManager.SetValue("PointsToCloudReport:searchTolCtl", self.searchTolCtl.Distance)
            OptionsManager.SetValue("PointsToCloudReport:numPtsToUseCtl", int(self.numPtsToUseCtl.Value))
            OptionsManager.SetValue("PointsToCloudReport.Name", self.nameCtl.SelectedName)

        # start excel if we can
        try:
            os.system('start excel.exe "%s"' % (fileName,))
        except Exception as ex:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, str(ex))
        cmd.CloseUICommand()

    def createPointsCB_Checked(self, sender, e):
        if self.createPointsCB.IsChecked:
            self.expandPointInfo.IsExpanded = True
            self.pointInfo.Visibility = Visibility.Visible
        else:
            self.pointInfo.Visibility = Visibility.Collapsed

    def expandPointInfo_Expanded(self, sender, e):        
        if self.expandPointInfo.IsExpanded:
            self.pointInfo.Visibility = Visibility.Visible
        else:
            self.pointInfo.Visibility = Visibility.Collapsed

    def CreatePoint(self, coord):
        p = self.pointCollection.Add(clr.GetClrType(Point))
        p.PointID = self.nameCtl.SelectedName
        self.nameCtl.SelectedName = PointHelper.Helpers.Increment(self.nameCtl.SelectedName, None)
        p.AddCoordinate(NEECoordinate(coord))
        p.Layer = self.layerCtl.SelectedSerialNumber
