﻿"""
Sample TBC mcro to compute sta/offset and delta elevation
of selected points
"""
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring

from Trimble.Vce.ForeignCad import TextEntity, MText, Text

from Trimble.Vce.Gem import Model3D

from Trimble.Vce.Data import MediaFolder

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember
from Trimble.Vce.Interfaces.Point import IPoint
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client

clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
from Trimble.Vce.Core import GlobalSelection  
from Trimble.Vce.UI.Controls import VceMessageBox, SurfaceTypeLists

clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "PointsToCsv"
    cmdData.CommandName = "PointsToCsv"
    cmdData.Caption = "_Points To Csv"
    cmdData.UIForm = "PointsToCsv"
    cmdData.HelpFile="Macros.chm"
    cmdData.HelpTopic="xxxx"

    try:
        cmdData.Version = 1.0
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Points To Csv"
        cmdData.ToolTipTitle = "Points To Csv"
        cmdData.ToolTipTextFormatted = "Create CSV file containing station, offset, and delta elevation of selected points."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class PointsToCsv(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\PointsToCsv.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroFileFolder = macroFileFolder
        self.macroName = "Export Points to CSV file"

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        types = Array[Type](SurfaceTypeLists.AllWithCutFillMap)
        self.refSurface.FilterByEntityTypes = types   
        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd
        self.objs.IsEntityValidCallback=self.IsValid
        optionMenu = SelectionContextMenuHandler()
        self.pType = clr.GetClrType(IPoint)
        self.mfType = clr.GetClrType(MediaFolder)
        self.lunits = self.currentProject.Units.Linear
        # we don't want the units to be included
        self.lfp = self.lunits.Properties.Copy()
        self.lfp.AddSuffix = False
        self.staUnits = self.currentProject.Units.Station
        
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

        self.incFC.IsChecked = OptionsManager.GetBool("PointsToCvs:incFC", True)
        self.incD1.IsChecked = OptionsManager.GetBool("PointsToCvs:incD1", False)
        self.incD2.IsChecked = OptionsManager.GetBool("PointsToCvs:incD2", False)
        self.incMF.IsChecked = OptionsManager.GetBool("PointsToCvs:incMF", False)
        self.incLayer.IsChecked = OptionsManager.GetBool("PointsToCvs:incLayer", False)

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o=self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        # needs to be a point type
        if isinstance(obj, self.pType):
            return True
        return False

    def ConvertToDisplay(self, value):
        # this will convert the number to the current display units (all values are stored as meters)
        return self.lunits.Format(value, self.lfp)

    def OkClicked(self, cmd, e):
        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        wv = self.currentProject[Project.FixedSerial.WorldView]
        layers = self.currentProject[Project.FixedSerial.LayerContainer]

        self.refSurfaceModel = self.refSurface.SelectedEntity()
        self.alignPolyseg = None
        self.hal = None
        if self.refAlignment.SelectedEntity():
            self.alignPolyseg = self.refAlignment.SelectedEntity().ComputePolySeg()
            self.hal = self.refAlignment.SelectedEntity()

        # points are stored in meters, get conversion factor
        self.unitFactor = self.currentProject.Units.Linear.Convert(1.0, LinearType.Display)
        self.cultureInfo = CultureInfo.CurrentCulture.TextInfo.ListSeparator
        self.coordsInNE = self.currentProject.Units.Coordinate.CoordinateOrder == CoordinateOrder.NEZ

        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.refSurface.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.refSurface.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)

        fileName = Path.ChangeExtension(fileName, ".csv")
        if File.Exists(fileName):
            File.Delete(fileName)
        with open(fileName, 'w') as f:   
            self.file = f
            sComma = str(self.cultureInfo) # could be period in some regions
            for o in self.objs.SelectedMembers(self.currentProject):
                if not self.IsValidObj(o):
                    continue
                p = o.Position
                name = ""
                if isinstance(o, clr.GetClrType(IName)):
                    name = o.Name
                # make sure name doesn't have " character
                if name.IndexOf(sComma, 0):
                    name = name.Replace(sComma, "_")
                if self.coordsInNE:
                    sCsv = name + sComma + self.ConvertToDisplay(p.Y) + sComma + self.ConvertToDisplay(p.X)
                else:
                    sCsv = name + sComma + self.ConvertToDisplay(p.X) + sComma + self.ConvertToDisplay(p.Y)
                if Double.IsNaN(p.Z):
                    sCsv = sCsv + sComma
                else:
                    sCsv = sCsv + sComma + self.ConvertToDisplay(p.Z)

                if self.incFC.IsChecked:
                    fCode = ""
                    if isinstance(o, clr.GetClrType(IFeatureCodeHost)):
                        # we are only getting the first FC
                        for fc in IFeatureCodeHost.FeatureCodes.GetValue(o):
                            fCode = fc.Code
                            break
                        if fCode.IndexOf(sComma, 0):
                            fCode = fCode.Replace(sComma, "_")
                    sCsv = sCsv + sComma + fCode

                if self.incD1.IsChecked:
                    d1 = ""
                    if isinstance(o, clr.GetClrType(IDescriptionFields)):
                        d1 = o.Description1
                        if d1.IndexOf(sComma, 0):
                            d1 = d1.Replace(sComma, "_")
                    sCsv = sCsv + sComma + d1

                if self.incD2.IsChecked:
                    d2 = ""
                    if isinstance(o, clr.GetClrType(IDescriptionFields)):
                        d2 = o.Description2
                        if d2.IndexOf(sComma, 0):
                            d2 = d2.Replace(sComma, "_")
                    sCsv = sCsv + sComma + d2

                if self.incLayer.IsChecked:
                    layer = ""
                    if isinstance(o, clr.GetClrType(IViewMember)):
                        l = layers.Lookup(o.Layer)
                        if l:
                            layer = l.Name
                    sCsv = sCsv + sComma + layer

                if self.alignPolyseg:
                    station = ""
                    offset = ""
                    rtn = self.alignPolyseg.FindPointFromPoint(p, None, None)
                    if rtn[0]:
                        station = self.staUnits.Format(rtn[4], self.hal)
                        offset =  self.ConvertToDisplay(rtn[5])
                    sCsv = sCsv + sComma + station + sComma + offset

                if self.refSurfaceModel:
                    surfaceElev = ""
                    deltaElev = ""
                    found,pSurface = self.refSurfaceModel.PickSurface(p)# get point from surface
                    if found:
                        surfaceElev = self.ConvertToDisplay(pSurface.Z)
                        if not Double.IsNaN(p.Z):
                            deltaElev = self.ConvertToDisplay(pSurface.Z - p.Z)
                    sCsv = sCsv + sComma + surfaceElev + sComma + deltaElev

                if self.incMF.IsChecked:
                    mediaPath = ""
                    observers = self.currentProject.Concordance.GetObserversOf(o.SerialNumber)
                    for oo in observers:
                        if isinstance(oo, self.mfType): # is it MediaFolder?
                            s = ""
                            for fp in oo.FileProperties:
                                mediaPath = mediaPath + s + fp.Name
                                s = "+" # character used if more than one image
                    sCsv = sCsv + sComma + mediaPath

                sCsv = sCsv +  "\n"
                self.file.write(sCsv)

        OptionsManager.SetValue("PointsToCvs:incFC", self.incFC.IsChecked)
        OptionsManager.SetValue("PointsToCvs:incD1", self.incD1.IsChecked)
        OptionsManager.SetValue("PointsToCvs:incD2", self.incD2.IsChecked)
        OptionsManager.SetValue("PointsToCvs:incMF", self.incMF.IsChecked)
        OptionsManager.SetValue("PointsToCvs:incLayer", self.incLayer.IsChecked)
        cmd.CloseUICommand()

