﻿"""
Sample TBC mcro to compute grid cells and create points from values in cell
"""
#region
import clr
clr.AddReference('IronPython.Wpf')
import wpf
from System import Math, Action, Func, DateTime
from System.Text import Encoding
from System.Collections.Generic import List
from System.Windows import Window, MessageBox, MessageBoxResult, MessageBoxButton, Visibility
from System.Windows.Controls import StackPanel, Button, CheckBox
from System.Windows.Input import Keyboard, AccessKeyManager, Key
from System import Type, Tuple, Array, Double, String
from System.IO import StreamReader, MemoryStream, Path, File
from System.Globalization import CultureInfo
from System.Windows.Threading import Dispatcher
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys
from System.Windows.Forms import DialogResult

from Trimble.Vce.Core import TransactMethodCall, ModelEvents
from Trimble.Vce.Core.Components import WorldView, Project, Layer, TextStyle
from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring

from Trimble.Vce.ForeignCad import TextEntity, MText, Text

from Trimble.Vce.Gem import Model3D

from Trimble.Vce.Coordinates import PointCollection, NEECoordinate, Point

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Geometry import Point3D, Plane3D, Arc, Matrix4D, Vector3D, Side
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import ISnapIn, IPolyseg, IHaveUcs, IMemberManagement, IName, IViewMember
from Trimble.Vce.Interfaces.Point import IPoint, IPointLocation
from Trimble.Vce.Interfaces.Units import CoordinateOrder, LinearType, VerticalAngleType, AzimuthType, AngularType
from Trimble.Vce.Interfaces.FeatureCoding import IFeatureCodeHost, IFeatureCode
from Trimble.Vce.Interfaces.DescriptionFields import IDescriptionFields
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces import ProgressBar
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
try:
    from Trimble.Sdk.UI import UIEventArgs
except:
    from Trimble.Sdk.Interfaces.UI import UIEventArgs
  
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
#endregion

# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "PointsToGrid"
    cmdData.CommandName = "PointsToGrid"
    cmdData.Caption = "_Points To Grid"
    cmdData.UIForm = "PointsToGrid"

    try:
        cmdData.Version = 1.02
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "Points To Grid"
        cmdData.ToolTipTitle = "Points To Grid"
        cmdData.ToolTipTextFormatted = "Create Points in a grid."

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class PointsToGrid(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\PointsToGrid.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption 
        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd
        self.objs.IsEntityValidCallback=self.IsValid
        optionMenu = SelectionContextMenuHandler()
        self.pType = clr.GetClrType(IPointLocation)
        
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

        self.gridSpacingCtl.Distance = OptionsManager.GetDouble("PointsToGrid:Grid", Double.NaN)
        self.cbLow.IsChecked = OptionsManager.GetBool("PointsToGrid:cbLow", True)
        self.cbHigh.IsChecked = OptionsManager.GetBool("PointsToGrid:cbHigh", False)
        self.cbAvg.IsChecked = OptionsManager.GetBool("PointsToGrid:cbAvg", False)
        self.cbEarly.IsChecked = OptionsManager.GetBool("PointsToGrid:cbEarly", False)
        self.cbLast.IsChecked = OptionsManager.GetBool("PointsToGrid:cbLast", False)

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def IsValid(self, serial):
        o=self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        # needs to be a point type
        if isinstance(obj, self.pType):
            # should be 3D point
            if Double.IsNaN(obj.Position.Z):
                return False
            return True
        return False

    def GridPoint(self, point, along, iSta, iOffset):
        if iSta in along:
            across = along[iSta]
        else:
            across = {}
        if iOffset in across:
            pts = across[iOffset]
        else:
            pts = []
        pts.append(point.SerialNumber)
        across[iOffset] = pts
        along[iSta] = across

    def Round(self, val, grid):
        if val > 0:
            v =  int((val + grid/2) / grid)
        else:
            v  = int((val - grid/2) / grid)
        return v

    def CreatePoint(self, id, layer, coord):
        p = self.pointCollection.Add(clr.GetClrType(Point))
        p.PointID = id
        p.AddCoordinate( NEECoordinate(coord))
        p.Layer = layer

    def DateFromText(self, fCode):
        try:
            # date formatted like: 20190417-144101.90//L
            year = int(fCode.Substring(0,4))
            month = int(fCode.Substring(4, 2))
            day = int(fCode.Substring(6, 2))
            hour = int(fCode.Substring(9, 2))
            min = int(fCode.Substring(11, 2))
            sec = int(fCode.Substring(13, 2))
            mSec = int(fCode.Substring(16, 2)) * 10
            return DateTime(year, month, day, hour, min, sec, mSec)
        except:
            return None

    def OkClicked(self, cmd, e):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        self.pointCollection = PointCollection.ProvidePointCollection(self.currentProject)
        h = self.refAlignment.SelectedSerialNumber()
        if not h:
            self.refAlignment.StatusMessage = "Select reference alignment"
            Keyboard.Focus(self.refAlignment)
            return False
        self.refAlignment.StatusMessage = ""

        gridSize = self.gridSpacingCtl.Distance
        if Double.IsNaN(gridSize) or gridSize <= 0.0:
            self.gridSpacingCtl.StatusMessage = "Enter valid scale factor"
            return False
        self.gridSpacingCtl.StatusMessage = ""

        self.gridSize = gridSize
        self.hal = self.refAlignment.SelectedEntity()
        self.alignPolyseg = self.hal.ComputePolySeg().Clone()
        self.alignPolyseg.Extend(10000.0)
        # compute cells
        along = {}
        for o in self.objs.SelectedMembers(self.currentProject):
            if not self.IsValidObj(o):
                continue
            p = o.Position

            rtn = self.alignPolyseg.FindPointFromPoint(p, None, None)
            if not rtn[0]:
                continue
            station = rtn[4]
            offset =  rtn[5]
            self.GridPoint(o, along, self.Round(station, gridSize), self.Round(offset, gridSize))

        numCells = 0
        maxCellPoints = 0
        minCellPoints = 10000000
        totalCellPoints = 0
        for va in along.itervalues():
            for vc in va.itervalues():
                l = len(vc)
                if l < minCellPoints:
                    minCellPoints = l
                if l > maxCellPoints:
                    maxCellPoints = l
                totalCellPoints += l
                numCells += 1
        info = "Cells Computed: " + str(numCells)
        if numCells:
            info = info + "\rMin: " + str(minCellPoints)
            info = info + "\rMax: " + str(maxCellPoints)
            info = info + "\rAve: " + str(int(totalCellPoints / numCells))
            info = info + "\r\rDo you want to Compute Grid Points?"
        r = VceMessageBox.ShowQuestionYesNo(None, info, "Grid cells computed")
        if r == DialogResult.Yes:
            UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())
            # pause the graphics so every change doesn't draw
            wv.PauseGraphicsCache(True)
            try:
                self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
                # the "with" statement will unroll any changes if something go wrong
                ProgressBar.TBC_ProgressBar.Title = "Creating Points"
                with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:   
                    concordance = self.currentProject.Concordance
                    nPts = 0
                    for ka in along:
                        sta = ka * gridSize
                        va = along[ka] # dictionary of values at offsets
                        for kc in va:
                            offset = kc * gridSize
                            seg = clr.Reference[Segment]()
                            t = clr.Reference[float]()
                            pcl = clr.Reference[Point3D]()
                            v = clr.Reference[Vector3D]()
                            if not self.alignPolyseg.FindPointFromStation(sta, seg, t, pcl, v):
                                continue
                            coord = pcl.Value + v.Value * offset
                            vc = va[kc] # list of point serial numbers
                            id = str(ka)+"-"+str(kc)
                            nPts += 1
                            if ProgressBar.TBC_ProgressBar.SetProgress(nPts * 100 // totalCellPoints):
                                break
                            if self.cbLow.IsChecked:
                                layer = self.layerLowCtl.SelectedSerialNumber
                                zLow = 1e10
                                for p in vc:
                                    point = concordance.Lookup(p)
                                    if point:
                                        if point.Position.Z < zLow:
                                            zLow = point.Position.Z
                                coord = Point3D(coord.X, coord.Y, zLow)
                                self.CreatePoint(id, layer, coord)

                            if self.cbHigh.IsChecked:
                                layer = self.layerHighCtl.SelectedSerialNumber
                                zHigh = -1e10
                                for p in vc:
                                    point = concordance.Lookup(p)
                                    if point:
                                        if point.Position.Z > zHigh:
                                            zHigh = point.Position.Z
                                coord = Point3D(coord.X, coord.Y, zHigh)
                                self.CreatePoint(id, layer, coord)

                            if self.cbAvg.IsChecked:
                                layer = self.layerAvgCtl.SelectedSerialNumber
                                z = 0
                                n = 0
                                for p in vc:
                                    point = concordance.Lookup(p)
                                    if point:
                                        n += 1
                                        z += point.Position.Z
                                if n:
                                    coord = Point3D(coord.X, coord.Y, z / n)
                                    self.CreatePoint(id, layer, coord)

                            if self.cbEarly.IsChecked:
                                layer = self.layerEarlyCtl.SelectedSerialNumber
                                _Point = None
                                _Date = DateTime(2999, 1, 1)
                                for p in vc:
                                    point = concordance.Lookup(p)
                                    if point and isinstance(point, clr.GetClrType(IFeatureCodeHost)):
                                        # we are only getting the first FC
                                        for fc in IFeatureCodeHost.FeatureCodes.GetValue(point):
                                            fCode = fc.Code
                                            date = self.DateFromText(fCode)
                                            if date < _Date:
                                                _Point = point
                                                _Date = date
                                if _Point:
                                    coord = Point3D(coord.X, coord.Y, _Point.Position.Z)
                                    self.CreatePoint(id, layer, coord)

                            if self.cbLast.IsChecked:
                                layer = self.layerLastCtl.SelectedSerialNumber
                                _Point = None
                                _Date = DateTime(1)
                                for p in vc:
                                    point = concordance.Lookup(p)
                                    if point and isinstance(point, clr.GetClrType(IFeatureCodeHost)):
                                        # we are only getting the first FC
                                        for fc in IFeatureCodeHost.FeatureCodes.GetValue(point):
                                            fCode = fc.Code
                                            date = self.DateFromText(fCode)
                                            if date > _Date:
                                                _Point = point
                                                _Date = date
                                if _Point:
                                    coord = Point3D(coord.X, coord.Y, _Point.Position.Z)
                                    self.CreatePoint(id, layer, coord)

                    failGuard.Commit()

            finally:
                OptionsManager.SetValue("PointsToGrid:Grid", self.gridSpacingCtl.Distance)
                OptionsManager.SetValue("PointsToGrid:cbLow", self.cbLow.IsChecked)
                OptionsManager.SetValue("PointsToGrid:cbHigh", self.cbHigh.IsChecked)
                OptionsManager.SetValue("PointsToGrid:cbAvg", self.cbAvg.IsChecked)
                OptionsManager.SetValue("PointsToGrid:cbEarly", self.cbEarly.IsChecked)
                OptionsManager.SetValue("PointsToGrid:cbLast", self.cbLast.IsChecked)
                self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
                UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
                wv.PauseGraphicsCache(False)

        cmd.CloseUICommand()

