﻿"""
TBC Macro to change the name of a block
"""
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, BlockView
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Array
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from System.IO import StreamReader
from System.Windows.Controls import StackPanel

from Trimble.Vce.Interfaces.Client import CommandGranularity

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "RenameBlock"
    cmdData.CommandName = "RenameBlock"
    cmdData.Caption = "_Rename Block"
    cmdData.UIForm = "RenameBlock"
    cmdData.Hidden = True

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class RenameBlock(StackPanel):
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\RenameBlock.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        self.block.AllowNone = False
        self.block.UseSelectionEngine=False
        self.block.SearchContainer = Project.FixedSerial.BlockCollection
        # we need an array of "types" that define what objects to include
        types = Array[Type]([clr.GetClrType(BlockView)])
        self.block.FilterByEntityTypes = types

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def OkClicked(self, sender, e):
        bCollection = self.currentProject[Project.FixedSerial.BlockCollection]
        blkName = self.blockName.Text
        if not blkName or bCollection.IsNameUsed(blkName):
            # empty name, need to report name error
            Keyboard.Focus(self.blockName)
            return False
        if self.block.SelectedSerial == 0:
            self.block.StatusMessage = "Select Block"
            return False

        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)      
        # the "with" statement will unroll any changes if something go wrong
        with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
            blk = self.block.SelectedEntity()
            blk.Name = blkName
            failGuard.Commit()
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
        sender.CloseUICommand()