﻿"""
Sample TBC macro to compute 3d slope intersect
"""
#region
import clr
# Reference the WPF assemblies
clr.AddReference('IronPython.Wpf')
import wpf
from System import Type, Tuple, Array, Double, Math, String
from System.IO import StreamReader
from System.Collections.Generic import List
from System.Windows.Controls import StackPanel
from System.Windows import  Visibility
from System.Windows.Input import Keyboard
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Point as GdiPoint

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core.Components.TextStyle import TextHeightMode, TextJustification
from Trimble.Vce.Core import TransactMethodCall

from Trimble.DiskIO import OptionsManager
from Trimble.Vce.Alignment import SuperelevationView, ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Intersections
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line

from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IViewMember, ISettableName, IName, IViewSite, ISnapInHelpers
from Trimble.Vce.Interfaces.Client import CommandGranularity
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import VceMessageBox, GestureDispatcher
try:
    from Trimble.Sdk.UI import InputMethod, I2DProjection
    from Trimble.Sdk.UI import CircleDrawingInstruction, LineDrawingInstruction, PolylineDrawingInstruction
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod, I2DProjection
    from Trimble.Sdk.Interfaces.UI import CircleDrawingInstruction, LineDrawingInstruction, PolylineDrawingInstruction
from Trimble.Sdk.UI import GdiCursorBase

from Trimble.Vce.UI.Controls.Adapters import InputSettingsAdapter
from Trimble.Vce.ForeignCad import TextEntity, Leader, MText

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice, UIEvents
clr.AddReference("Trimble.Vce.UI.Wrapper")
from Trimble.Vce.UI.Wrapper import CADWrapperHelper
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

#endregion
# define the command key and name
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "SlopeSlopeIntersect"
    cmdData.CommandName = "SlopeSlopeIntersect"
    cmdData.Caption = "_Slope-Slope Intersect"
    cmdData.UIForm = "SlopeSlopeIntersect"

    try:
        cmdData.Version = 0.92
        cmdData.MacroAuthor = "Trimble"
        #cmdData.MacroInfo =
        #r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2018/11/29/tml-assignnamefromlayer"
        #cmdData.HelpFile=cmdData.MacroInfo
    except:
        pass

    try:    
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "Surface"
        cmdData.ShortCaption = "Slope-Slope Intersect"
        cmdData.ToolTipTitle = "Slope-Slope Intersect"
        cmdData.ToolTipTextFormatted = "Compute 3D slope intersection between two lines"

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass 

class SlopeSlopeIntersect(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\SlopeSlopeIntersect.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroName = "Slope Slope Intersect"
        self.tmlFileFolder = macroFileFolder 

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.cmd = cmd
        self.Caption = cmd.Command.Caption
        self.lType = clr.GetClrType(IPolyseg)
        self.line1Ctl.AddTypeFilter(self.lType)
        self.line1RefCtl.AddTypeFilter(self.lType)
        self.line2Ctl.AddTypeFilter(self.lType)
        self.line2RefCtl.AddTypeFilter(self.lType)

        settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
        l = settings.GetUInt32("SlopeSlopeIntersect.Layer", 0)
        if self.currentProject.Concordance.Lookup(l):
            self.layerCtl.SelectedSerialNumber = l
        self.line1SlopeOptionCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line1SlopeOptionCtl", True)
        self.line2SlopeOptionCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line2SlopeOptionCtl", True)
        self.line1VOffsetCtl.Distance = OptionsManager.GetDouble("SlopeSlopeIntersect:line1VOffsetCtl", 0.0)
        self.line2VOffsetCtl.Distance = OptionsManager.GetDouble("SlopeSlopeIntersect:line2VOffsetCtl", 0.0)
        self.line1VOffsetTypeCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line1VOffsetTypeCtl", False)
        self.line2VOffsetTypeCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line2VOffsetTypeCtl", False)
        self.line1SlopeOptionCtl_Click(None, None)
        self.line2SlopeOptionCtl_Click(None, None)
        self.line1VOffsetMakeLineCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line1VOffsetMakeLineCtl", False)
        self.line2VOffsetMakeLineCtl.IsChecked = OptionsManager.GetBool("SlopeSlopeIntersect:line2VOffsetMakeLineCtl", False)

        self.slope1Ctl.Angle = OptionsManager.GetDouble("SlopeSlopeIntersect:slope1Ctl", Double.NaN)
        self.slope2Ctl.Angle = OptionsManager.GetDouble("SlopeSlopeIntersect:slope2Ctl", Double.NaN)
        self.samplingDistanceCtl.Distance = OptionsManager.GetDouble("SlopeSlopeIntersect:samplingDistanceCtl", Double.NaN)
        
        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd

        self.startStaCtl.ShowGdiCursor += self.DrawCustomCursor

    def CancelClicked(self, cmd, args):
        cmd.CloseUICommand()

    def line1Ctl_ValueChanged(self, sender, e):
        if not self.refAlignment.SelectedSerial:
            l = self.line1Ctl.Entity
            self.startStaCtl.StationProvider = l
            self.endStaCtl.StationProvider = l
        if self.line1SlopeOptionCtl.IsChecked and e.Cause == InputMethod.Mouse:
            Keyboard.Focus(self.line1RefCtl)

    def line2Ctl_ValueChanged(self, sender, e):
        if self.line2SlopeOptionCtl.IsChecked and e.Cause == InputMethod.Mouse:
            Keyboard.Focus(self.line2RefCtl)

    def refAlignment_ValueChanged(self, sender, e):        
        if self.refAlignment.SelectedSerial:
            l = self.refAlignment.SelectedEntity()
            self.startStaCtl.StationProvider = l
            self.endStaCtl.StationProvider = l

    def OkClicked(self, cmd, e):
        if self.ValidateInput(self.line1Ctl, self.line1SlopeOptionCtl, self.slope1Ctl, self.line1RefCtl, self.line1VOffsetCtl):
            return
        if self.ValidateInput(self.line2Ctl, self.line2SlopeOptionCtl, self.slope2Ctl, self.line2RefCtl, self.line2VOffsetCtl):
            return
        sampDist = self.samplingDistanceCtl.Distance
        if Double.IsNaN(sampDist):
            self.samplingDistanceCtl.StatusMessage = "Enter sampling distance"
            Keyboard.Focus(self.samplingDistanceCtl)
            return

        poly1 = self.line1Ctl.Entity.ComputePolySeg()
        if not poly1 or poly1.NumberOfNodes < 2:
            self.line1Ctl.StatusMessage = "Line not defined"
            return

        refHalPoly = None
        if self.refAlignment.SelectedSerial:
            refHalPoly = self.refAlignment.SelectedEntity().ComputePolySeg()

        begSta = self.startStaCtl.Distance
        endSta = self.endStaCtl.Distance
        if refHalPoly:
            if Double.IsNaN(begSta):
                begSta = refHalPoly.FirstNode.Station
            if begSta < refHalPoly.FirstNode.Station:
                self.startStaCtl.StatusMessage = "Enter start location"
                Keyboard.Focus(self.startStaCtl)
                return
            if Double.IsNaN(endSta):
                endSta = refHalPoly.LastNode.Station
            if endSta <= begSta:
                self.endStaCtl.StatusMessage = "Enter end location"
                Keyboard.Focus(self.endStaCtl)
                return
        else:
            if Double.IsNaN(begSta):
                begSta = poly1.FirstNode.Station
            if begSta < poly1.FirstNode.Station:
                self.startStaCtl.StatusMessage = "Enter start location"
                Keyboard.Focus(self.startStaCtl)
                return
            if Double.IsNaN(endSta):
                endSta = poly1.LastNode.Station
            if endSta <= begSta:
                self.endStaCtl.StatusMessage = "Enter end location"
                Keyboard.Focus(self.endStaCtl)
                return

        poly2 = self.line2Ctl.Entity.ComputePolySeg()
        if not poly2 or poly2.NumberOfNodes < 2:
            self.line2Ctl.StatusMessage = "Line not defined"
            return
        
        lineDeltaZ = self.line1VOffsetCtl.Distance
        lineMeasurePerpendicular = self.line1VOffsetTypeCtl.IsChecked
        if self.line1SlopeOptionCtl.IsChecked:
            e = self.line1RefCtl.Entity
            # if object in super view, then it defines the slope
            isSlope = isinstance(e.GetSite(), SuperelevationView)    
            polyRef = e.ComputePolySeg()
            polyRefVert = e.ComputeVerticalPolySeg()
            self.SlopeOptions1 = (True, lineDeltaZ, lineMeasurePerpendicular, isSlope, polyRef, polyRefVert)
        else:
            self.SlopeOptions1 = (False, lineDeltaZ, lineMeasurePerpendicular, Math.Tan(Math.PI / 2.0 - self.slope1Ctl.Angle))

        lineDeltaZ = self.line2VOffsetCtl.Distance
        lineMeasurePerpendicular = self.line2VOffsetTypeCtl.IsChecked
        if self.line2SlopeOptionCtl.IsChecked:
            e = self.line2RefCtl.Entity
            isSlope = isinstance(e.GetSite(), SuperelevationView)              
            polyRef = e.ComputePolySeg()
            polyRefVert = e.ComputeVerticalPolySeg()
            self.SlopeOptions2 = (True, lineDeltaZ, lineMeasurePerpendicular, isSlope, polyRef, polyRefVert)
        else:
            self.SlopeOptions2 = (False, lineDeltaZ, lineMeasurePerpendicular, Math.Tan(Math.PI / 2.0 - self.slope2Ctl.Angle))

        # tell undo manager we are starting to do something
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        try:
            self.debug = self.nameCtl.SelectedName.StartsWith("__@d")
            if self.debug:
                self.textBoxEntityInfo.Visibility = Visibility.Visible
                self.textBoxEntityInfo.Text = ""
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                self.ComputeSlopeSlopeIntersect(refHalPoly, poly1, self.line1Ctl.Entity.ComputeVerticalPolySeg(), 
                        poly2,self.line2Ctl.Entity.ComputeVerticalPolySeg(),begSta, endSta, sampDist)
                failGuard.Commit()
            settings = ConstructionCommandsSettings.ProvideObject(self.currentProject)
            settings.SetUInt32("SlopeSlopeIntersect.Layer", self.layerCtl.SelectedSerialNumber)

            OptionsManager.SetValue("SlopeSlopeIntersect:line1SlopeOptionCtl", self.line1SlopeOptionCtl.IsChecked)
            OptionsManager.SetValue("SlopeSlopeIntersect:line2SlopeOptionCtl", self.line2SlopeOptionCtl.IsChecked)   
            OptionsManager.SetValue("SlopeSlopeIntersect:line1VOffsetCtl", self.line1VOffsetCtl.Distance)
            OptionsManager.SetValue("SlopeSlopeIntersect:line2VOffsetCtl", self.line2VOffsetCtl.Distance)
            OptionsManager.SetValue("SlopeSlopeIntersect:line1VOffsetTypeCtl", self.line1VOffsetTypeCtl.IsChecked)
            OptionsManager.SetValue("SlopeSlopeIntersect:line2VOffsetTypeCtl", self.line2VOffsetTypeCtl.IsChecked)

            OptionsManager.SetValue("SlopeSlopeIntersect:line1VOffsetMakeLineCtl", self.line1VOffsetMakeLineCtl.IsChecked)
            OptionsManager.SetValue("SlopeSlopeIntersect:line2VOffsetMakeLineCtl", self.line2VOffsetMakeLineCtl.IsChecked)

            OptionsManager.SetValue("SlopeSlopeIntersect:slope1Ctl", self.slope1Ctl.Angle)
            OptionsManager.SetValue("SlopeSlopeIntersect:slope2Ctl", self.slope2Ctl.Angle)
            OptionsManager.SetValue("SlopeSlopeIntersect:samplingDistanceCtl", self.samplingDistanceCtl.Distance)

        finally:
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        Keyboard.Focus(self.line1Ctl)
   
    def DrawCustomCursor(self, sender, e):
        if not isinstance(e.MessagingView, clr.GetClrType(I2DProjection)):
            return
        if self.ValidateInput(self.line1Ctl, self.line1SlopeOptionCtl, self.slope1Ctl, self.line1RefCtl, self.line1VOffsetCtl):
            return
        if self.ValidateInput(self.line2Ctl, self.line2SlopeOptionCtl, self.slope2Ctl, self.line2RefCtl, self.line2VOffsetCtl):
            return
        poly1 = self.line1Ctl.Entity.ComputePolySeg()
        if not poly1 or poly1.NumberOfNodes < 2:
            return
        poly2 = self.line2Ctl.Entity.ComputePolySeg()
        if not poly2 or poly2.NumberOfNodes < 2:
            return
        lineDeltaZ = self.line1VOffsetCtl.Distance
        lineMeasurePerpendicular = self.line1VOffsetTypeCtl.IsChecked
        if self.line1SlopeOptionCtl.IsChecked:
            l = self.line1RefCtl.Entity
            # if object in super view, then it defines the slope
            isSlope = isinstance(l.GetSite(), SuperelevationView)    
            polyRef = l.ComputePolySeg()
            polyRefVert = l.ComputeVerticalPolySeg()
            self.SlopeOptions1 = (True, lineDeltaZ, lineMeasurePerpendicular, isSlope, polyRef, polyRefVert)
        else:
            self.SlopeOptions1 = (False, lineDeltaZ, lineMeasurePerpendicular, Math.Tan(Math.PI / 2.0 - self.slope1Ctl.Angle))
        lineDeltaZ = self.line2VOffsetCtl.Distance
        lineMeasurePerpendicular = self.line2VOffsetTypeCtl.IsChecked
        if self.line2SlopeOptionCtl.IsChecked:
            l = self.line2RefCtl.Entity
            isSlope = isinstance(l.GetSite(), SuperelevationView)              
            polyRef = l.ComputePolySeg()
            polyRefVert = l.ComputeVerticalPolySeg()
            self.SlopeOptions2 = (True, lineDeltaZ, lineMeasurePerpendicular, isSlope, polyRef, polyRefVert)
        else:
            self.SlopeOptions2 = (False, lineDeltaZ, lineMeasurePerpendicular, Math.Tan(Math.PI / 2.0 - self.slope2Ctl.Angle))

        refHalPoly = None
        if self.refAlignment.SelectedSerial:
            refHalPoly = self.refAlignment.SelectedEntity().ComputePolySeg()

        cursorLocation = e.MessagingView.ConvertPageToProject(Point3D(e.MousePosition.X, e.MousePosition.Y))
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        if refHalPoly:
            (found, pnt, staOnRefAlignment) = refHalPoly.FindPointFromPoint(cursorLocation)
            if not found: 
                return
            if not refHalPoly.FindPointFromStation(staOnRefAlignment, seg, t, pcl, v):
                return
            v1 = v.Value
            v1.To2D()
            v1.Normalize()
            pclOn1 = pcl.Value
            (pclOn1, staOn1) = self.VectorIntersectLine(pclOn1, v1, poly1, None)
            if pclOn1.IsUndefined:
                return
        else:
            (found, pnt, staOn1) = poly1.FindPointFromPoint(cursorLocation)
            if not found: 
                return
            if not poly1.FindPointFromStation(staOn1, seg, t, pcl, v):
                return
            v1 = v.Value
            v1.Normalize()
            pclOn1 = pcl.Value

        (pclOn2,s) = self.VectorIntersectLine(pclOn1, v1, poly2, None)
        if pclOn2.IsUndefined:
            return

        # create polyseg that connects 1 to 2
        across1Poly = PolySeg()
        across1Poly.Add(Array[Point3D]([pclOn1, pclOn2]))
        
        (found, tTest, pcl, side) =  across1Poly.FirstSegment.ProjectPoint(cursorLocation)
        (deltaElev, pnt, pcl1, pRef1, pcl2, pRef2) = self.ComputeElevDelta(staOn1, pclOn1, tTest, across1Poly, v1,
                     self.line1Ctl.Entity.ComputeVerticalPolySeg(), poly2, self.line2Ctl.Entity.ComputeVerticalPolySeg())
        if Double.IsNaN(deltaElev):
            return

        pagePt = e.MessagingView.ConvertProjectToPage(pcl2)
        gdiPoint = GdiPoint(pagePt.X, pagePt.Y)
        pcl2.X = pcl2.X + deltaElev # make radius show elev miss
        pagePtR = e.MessagingView.ConvertProjectToPage(pcl2)
        radius = Math.Abs(pagePtR.X - pagePt.X)
        cursor = MyCustomCursor(e.MousePosition, int(radius), gdiPoint)
        GestureDispatcher.PushVisibleCursor(cursor.Show(None, e.MessagingView, e.MousePosition));


    def ValidateInput(self, lineCtl, option, slopeCtl, refLineCtl, lineVOffsetCtl):
        line = lineCtl.Entity
        if not line:
            lineCtl.StatusMessage = "No line selected"
            Keyboard.Focus(lineCtl)
            return True
        lineCtl.StatusMessage = ""
        refLineCtl.StatusMessage = ""
        slopeCtl.StatusMessage = ""
        if option.IsChecked:
            line = refLineCtl.Entity
            if not line:
                refLineCtl.StatusMessage = "No line selected"
                Keyboard.Focus(refLineCtl)
                return True
        else:
            v = slopeCtl.Angle
            if Double.IsNaN(v):
                slopeCtl.StatusMessage = "Enter valid slope"
                Keyboard.Focus(slopeCtl)
                return True
        vo = lineVOffsetCtl.Distance
        if Double.IsNaN(vo):
            lineVOffsetCtl.StatusMessage = "Enter vertical offset distance"
            Keyboard.Focus(lineVOffsetCtl)
            return True
        return False

    def line1SlopeOptionCtl_Click(self, sender, e):
        if self.line1SlopeOptionCtl.IsChecked:
            self.slope1NameLabel.Visibility = Visibility.Collapsed
            self.slope1Ctl.Visibility = Visibility.Collapsed
            self.line1RefLabel.Visibility = Visibility.Visible
            self.line1RefCtl.Visibility = Visibility.Visible
        else:
            self.slope1NameLabel.Visibility = Visibility.Visible
            self.slope1Ctl.Visibility = Visibility.Visible
            self.line1RefLabel.Visibility = Visibility.Collapsed
            self.line1RefCtl.Visibility = Visibility.Collapsed
    
    def line2SlopeOptionCtl_Click(self, sender, e):
        if self.line2SlopeOptionCtl.IsChecked:
            self.slope2NameLabel.Visibility = Visibility.Collapsed
            self.slope2Ctl.Visibility = Visibility.Collapsed
            self.line2RefLabel.Visibility = Visibility.Visible
            self.line2RefCtl.Visibility = Visibility.Visible
        else:
            self.slope2NameLabel.Visibility = Visibility.Visible
            self.slope2Ctl.Visibility = Visibility.Visible
            self.line2RefLabel.Visibility = Visibility.Collapsed
            self.line2RefCtl.Visibility = Visibility.Collapsed
    
    def endStaCtl_ValueChanged(self, sender, e):
        if e.Cause == InputMethod.Mouse:     
            self.OkClicked(None, None)

# methods to calculate slope intersect
    def ComputeSlopeSlopeIntersect(self, refHalPoly, poly1, vPoly1, poly2, vPoly2, begSta, endSta, deltaSta):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        if refHalPoly:
            if begSta < refHalPoly.FirstNode.Station:
                begSta = refHalPoly.FirstNode.Station
            if endSta > refHalPoly.LastNode.Station:
                endSta = refHalPoly.LastNode.Station
        else:
            if begSta < poly1.FirstNode.Station:
                begSta = poly1.FirstNode.Station
            if endSta > poly1.LastNode.Station:
                endSta = poly1.LastNode.Station
        pts = List[Point3D]() 
        pts1 = List[Point3D]() 
        pts1R = List[Point3D]() 
        pts2 = List[Point3D]() 
        pts2R = List[Point3D]() 
        if deltaSta > 0.01:
            sta = begSta
            while sta <= endSta:
                # if near end, set to end
                if Math.Abs(endSta - sta) < deltaSta / 2:
                    sta = endSta
                (p, pcl1, pRef1, pcl2, pRef2) = self.ComputeSlopeSlopeIntersectAtStation(refHalPoly, sta, poly1, vPoly1, poly2, vPoly2)
                if p.Is3D:
                    if p.Z < -1.0e10:
                        break
                    pts.Add(p)
                    pts1.Add(pcl1)
                    pts1R.Add(pRef1)
                    pts2.Add(pcl2)
                    pts2R.Add(pRef2)
                elif pts and not self.debug:
                    # if we can't find solution and we did before, stop
                    break
                sta += deltaSta
                # if near end, set to end
                if Math.Abs(endSta - sta) < deltaSta / 2:
                    sta = endSta

        else:
            # compute at poly1 nodes
            for node in poly1.Nodes():
                if refHalPoly:
                    (found, pnt, sta)=refHalPoly.FindPointFromPoint(node.Point)
                    if not found:
                        if pts:
                            break
                        continue
                else:
                    sta = node.Station
                if sta < begSta - Point3D.Tolerance:
                    continue
                if sta > endSta + Point3D.Tolerance:
                    break
                (p, pcl1, pRef1, pcl2, pRef2) = self.ComputeSlopeSlopeIntersectAtStation(refHalPoly, sta, poly1, vPoly1, poly2, vPoly2)
                if p.Is3D:
                    if p.Z < -1.0e10:
                        break
                    pts.Add(p)
                    pts1.Add(pcl1)
                    pts1R.Add(pRef1)
                    pts2.Add(pcl2)
                    pts2R.Add(pRef2)
                elif pts and not self.debug:
                    # if we can't find solution and we did before, stop
                    break    

        # check if any points computed
        if not pts:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Could not compute intersection. Check elevations.")
            return False
        polyseg = PolySeg()
        polyseg.Add(pts)

        ls = wv.Add(clr.GetClrType(Linestring))
        ls.Layer = self.layerCtl.SelectedSerialNumber 
        ls.Name = self.nameCtl.SelectedName
        ls.Append(polyseg, None, False, False)
        # check if we offset the segment
        if self.SlopeOptions1[1] and self.line1VOffsetMakeLineCtl.IsChecked:
            polyseg = PolySeg()
            polyseg.Add(pts1)
            ls = wv.Add(clr.GetClrType(Linestring))
            ls.Layer = self.layerCtl.SelectedSerialNumber 
            ls.Name = self.nameCtl.SelectedName
            ls.Append(polyseg, None, False, False)

            if self.SlopeOptions1[0]:
                polyseg = PolySeg()
                polyseg.Add(pts1R)
                ls = wv.Add(clr.GetClrType(Linestring))
                ls.Layer = self.layerCtl.SelectedSerialNumber 
                ls.Name = self.nameCtl.SelectedName
                ls.Append(polyseg, None, False, False)

        if self.SlopeOptions2[1] and self.line2VOffsetMakeLineCtl.IsChecked:
            polyseg = PolySeg()
            polyseg.Add(pts2)
            ls = wv.Add(clr.GetClrType(Linestring))
            ls.Layer = self.layerCtl.SelectedSerialNumber 
            ls.Name = self.nameCtl.SelectedName
            ls.Append(polyseg, None, False, False)

            if self.SlopeOptions2[0]:
                polyseg = PolySeg()
                polyseg.Add(pts2R)
                ls = wv.Add(clr.GetClrType(Linestring))
                ls.Layer = self.layerCtl.SelectedSerialNumber 
                ls.Name = self.nameCtl.SelectedName
                ls.Append(polyseg, None, False, False)

        return True

    def ComputeSlopeSlopeIntersectAtStation(self, refHalPoly, staOnRefAlignment, poly1, vPoly1, poly2, vPoly2):
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v = clr.Reference[Vector3D]()
        if refHalPoly:
            if not refHalPoly.FindPointFromStation(staOnRefAlignment, seg, t, pcl, v):
                return (Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
            v1 = v.Value
            v1.To2D()
            v1.Normalize()
            pclOn1 = pcl.Value
            (pclOn1, staOn1) = self.VectorIntersectLine(pclOn1, v1, poly1, vPoly1)
        else:
            staOn1 = staOnRefAlignment
            if not poly1.FindPointFromStation(staOn1, seg, t, pcl, v):
                return (Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
            v1 = v.Value
            v1.To2D()
            v1.Normalize()
            pclOn1 = pcl.Value
            (pclOn1, staOn1) = self.VectorIntersectLine(pclOn1, v1, poly1, vPoly1)
        (pclOn2,s) = self.VectorIntersectLine(pclOn1, v1, poly2, vPoly2)

        # create polyseg that connects 1 to 2
        across1Poly = PolySeg()
        across1Poly.Add(Array[Point3D]([pclOn1, pclOn2]))

        (found, pnt, sta2) = poly2.FindPointFromPoint(pclOn1)
        if not found: 
            return (Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
        # first guess at sta is mid point of line
        tTest = 0.5
        prevDeltaElev = 0
        if self.debug:
            self.textBoxEntityInfo.Text = "Sta: {0:.3}".format(staOn1)
        for i in range(40):  
            (deltaElev, pnt, pcl1, pRef1, pcl2, pRef2) = self.ComputeElevDelta(staOn1, pclOn1, tTest, across1Poly, v1, vPoly1, poly2, vPoly2)
            if Double.IsNaN(deltaElev):
                return (Point3D.Undefined, None, None, None, None)

            if self.debug:
                txt = '\n"t"{0:.3}'.format(tTest) + ' deltaElev {0:.3}'.format(deltaElev)
                self.textBoxEntityInfo.Text += txt
            if abs(deltaElev) < 0.0001:
                return (pnt, pcl1, pRef1, pcl2, pRef2)

            if prevDeltaElev == 0:
                prevDeltaElev = deltaElev
                preT = tTest
                # just use a small adj the first time
                tTest = tTest + 0.1
            else:
                t = self.Interpolate(preT, tTest, prevDeltaElev, deltaElev, 0.0)
                if False:
                    if Math.Abs(tTest - t) > 0.1:
                        if t > tTest:
                            t = tTest + 0.1
                        else:
                            t = tTest - 0.1
                prevDeltaElev = deltaElev
                preT = tTest
                tTest = t

        VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, "Did not converge within tolerance")
        return (Point3D(0, 0, -1.1e10), Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)

    def Interpolate(self, x1, x2, y1, y2, y):
        r = (y2 - y1) / (x2 - x1)
        x = (y - y1) / r 
        x = x + x1
        return x

    def ComputeElevDelta(self, staOn1, pclOn1, tTest, acrossPoly, v1, vPoly1, poly2, vPoly2):
        (found, pTest) = acrossPoly.FirstSegment.ComputePoint(tTest)
        seg = clr.Reference[Segment]()
        t = clr.Reference[float]()
        staOn2 = clr.Reference[float]()
        offset = clr.Reference[float]()
        pcl = clr.Reference[Point3D]()
        v2 = clr.Reference[Vector3D]()
        side = clr.Reference[Side]()
        rtn = poly2.FindPointFromPoint(pTest, seg, t, pcl, staOn2, v2, offset, side)
        if seg.Value == None:
            return (Double.NaN, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined, Point3D.Undefined)
        (found, v2) = seg.Value.Slope(t.Value)
        v2.To2D()
        v2.Rotate90(Side.Right)
        v2.Normalize()
        (e2, pcl2, pRef2) = self.ComputeOffsetElev(True, staOn2.Value, pcl.Value, offset.Value, vPoly2, self.SlopeOptions2, v2, pclOn1, pTest)
        offset = Math.Abs(acrossPoly.FirstSegment.Length2D * tTest)
        # we only allow the fixed slopes to switch direction
        doSlopeFix = self.SlopeOptions2[0] and not self.SlopeOptions2[3]
        (e1, pcl1, pRef1) = self.ComputeOffsetElev(doSlopeFix, staOn1, pclOn1, offset, vPoly1, self.SlopeOptions1, v1, pcl.Value, pTest)
        pTest.Z = e1
        deltaElev = e2 - e1
        return (deltaElev, pTest, pcl1, pRef1, pcl2, pRef2)

    def ComputeOffsetElev(self, fixSlope, sta, pcl, offset, polyVert, slopeOptions, v, otherPnt, pTest):
        z = pcl.Z
        pRef = Point3D.Undefined
        if polyVert:
            (found, segment, p, isOn) = polyVert.LocateX(Point3D(sta, 0.0))
            if found:
                z = p.Y
        if Double.IsNaN(z):
            return (z, pcl, pRef)
        slope = Double.NaN
        zAdj = slopeOptions[1]
        adjPerp = slopeOptions[2]
        if slopeOptions[0]:

            if slopeOptions[3]:
                # slope table used
                (found, segment, p, isOn) = slopeOptions[4].LocateX(Point3D(sta, 0.0))
                if found:
                    slope = p.Y / 100
            else:
                # find intersection of vector from line to ref line
                fixSlope = False
                acrossPoly = PolySeg()
                acrossPoly.Add(Array[Point3D]([pcl - 10000.0 * v, pcl + 10000.0 * v]))
                ints = Intersections()

                if not slopeOptions[4].Intersect(acrossPoly, True, ints):
                    return (Double.NaN, pcl, pRef)
                # we could get more than one intersection.  Find nearest to middle (t
                # value of 0.5)
                bestT = 1e10
                bestI = 0
                for i in range(ints.Count):
                    if Math.Abs(ints[i].T2 - 0.5) < bestT:
                        bestT = ints[i].T2
                        bestI = i
                i = ints[i]
                staRef = i.Segment1.Station(i.T1)
                (ret, pRef) = i.Segment1.ComputePoint(i.T1)
                # check if ref line has vertical polyseg
                if slopeOptions[5]:
                    (found, segment, p, isOn) = slopeOptions[5].LocateX(Point3D(staRef, 0.0))
                    if found:
                        pRef.Z = p.Y
                (dist, p, p) = Line2D.Length(pcl, pRef)
                if dist:
                    slope = (pRef.Z - z) / dist
                    (t, p, p, p) = Line2D.TAtPoint(pcl, pRef, pTest)
                    if t < 0:
                        slope = -slope
        else:
            slope = slopeOptions[3]
           
        if Double.IsNaN(slope):
            return (z, pcl, pRef)
        if fixSlope:
            # check if intersection is on oppsite side
            (t, p, p, p) = Line2D.TAtPoint(pcl, otherPnt, pTest)
            if t < 0:
                slope = -slope
        
        deltaZ = offset * slope
        # check if we should ofset the line
        pcl_ = Point3D(pcl)
        pcl_.Z  = z
        if zAdj:
            if adjPerp:
                zAdj = zAdj / Math.Cos(Math.Atan(slope))
            deltaZ += zAdj
            pcl_.Z += zAdj
            if pRef.Is3D:
                pRef.Z += zAdj
        z = z + deltaZ
        return (z, pcl_, pRef)

    def VectorIntersectLine(self, fromPoint, directionVector, polyToIntersect, polyToIntersectVert):
        acrossPoly = PolySeg()
        acrossPoly.Add(Array[Point3D]([fromPoint - 10000.0 * directionVector, fromPoint + 10000.0 * directionVector]))
        ints = Intersections()
        if not polyToIntersect.Intersect(acrossPoly, True, ints):
            return (Point3D.Undefined, Double.NaN)
        # we could get more than one intersection.  
        # Find nearest to middle (t value of 0.5)
        bestT = 1e10
        bestI = 0
        for i in range(ints.Count):
            if Math.Abs(ints[i].T2 - 0.5) < bestT:
                bestT = ints[i].T2
                bestI = i
        i = ints[bestI]
        staRef = i.Segment1.Station(i.T1)
        (ret, pRef) = i.Segment1.ComputePoint(i.T1)
        # check if ref line has vertical polyseg
        if polyToIntersectVert:
            (found, segment, p, isOn) = polyToIntersectVert.LocateX(Point3D(staRef, 0.0))
            if found:
                pRef.Z = p.Y
        return (pRef, staRef)

class MyCustomCursor(GdiCursorBase):
    def __new__(self, cursorLocation, radius, poly2Pol):
        return GdiCursorBase.__new__(self, InputSettingsAdapter())

    def __init__(self, cursorLocation, radius, poly2Pol):
        self.cursorLocation = cursorLocation
        self.radius = radius
        self.poly2Pol = poly2Pol

    def GenerateDrawingInstructions(self, viewCollection, messagingView, mousePosition):
        width = int(self.m_InputSettings.PickAperture * 0.66)
        if self.radius:
            instru = CircleDrawingInstruction(messagingView, self.cursorLocation, self.radius)
            instru.SetComputedColor(self.CursorColor)
            instru.Width = width
            self.PushInstruction(instru.Xor())
        instru = LineDrawingInstruction(messagingView, self.cursorLocation, self.poly2Pol)
        instru.SetComputedColor(self.CursorColor)
        instru.Width = width
        self.PushInstruction(instru.Xor())
        