﻿"""
TBC Macro to compare one surface to another
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap, Color
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings, GemVertexType, GemCorner

from Trimble.Vce.Alignment import ProfileView

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
        
from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox

from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "SurfaceImport"
    cmdData.CommandName = "SurfaceImport"
    cmdData.Caption = "_Surface Import"
    cmdData.UIForm = "SurfaceImport"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"
    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class SurfaceImport(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\SurfaceImport.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        types = Array[Type](SurfaceTypeLists.AllWithCutFillMap)
        self.surface.FilterByEntityTypes = types

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def OkClicked(self, sender, e):
        doImport = self.importOption.IsChecked
        if not doImport:
            if self.surface.SelectedSerial == 0:
                self.surface.StatusMessage = "Enter a surface"
                return False
            self.SurfaceModel = self.surface.SelectedEntity()

        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.surface.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.surface.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)

        fileName = Path.ChangeExtension(fileName, ".xyz")
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        ProgressBar.TBC_ProgressBar.Title = "Computing Surface"
           
        try:
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                if doImport:
                    self.ReadSurfaceFile(fileName)
                else:
                    self.WriteSurfaceFile(fileName)
                failGuard.Commit()

        except Exception as ex:
            VceMessageBox.ShowError(str(ex), "Error:" + fileName)

        self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        sender.CloseUICommand()

    def WriteSurfaceFile(self, fileName):
        if File.Exists(fileName):
            File.Delete(fileName)
        sComma = ","
        with open(fileName, 'w') as f: 
            nVertices = self.SurfaceModel.NumberOfVertices
            nTri = self.SurfaceModel.NumberOfTriangles
            f.write(str(nVertices) + "\n")
            f.write(str(nTri) + "\n")
            for i in range(nVertices):
                p = self.SurfaceModel.GetVertexPoint(i)
                s = str(i) + sComma + format(p.X, "0.5f") + sComma + format(p.Y, "0.5f") + \
                    sComma + format(p.Z, "0.5f") + "\n"
                f.write(s)
            for i in range(nTri):  
                iVertex0 = self.SurfaceModel.GetTriangleIVertex(i, 0)
                iVertex1 = self.SurfaceModel.GetTriangleIVertex(i, 1)
                iVertex2 = self.SurfaceModel.GetTriangleIVertex(i, 2)

                if self.SurfaceModel.GetTriangleOuterSide(i, 0):
                    iTriangleLeft0 = -1
                    corner0 = 0
                else:
                    iTriangleLeft0, corner0 = self.SurfaceModel.GetTriangleLeftCorner(i, 0)

                if self.SurfaceModel.GetTriangleOuterSide(i, 1):
                    iTriangleLeft1 = -1
                    corner1 = 0
                else:
                    iTriangleLeft1, corner1 = self.SurfaceModel.GetTriangleLeftCorner(i, 1)

                if self.SurfaceModel.GetTriangleOuterSide(i, 2):
                    iTriangleLeft2 = -1
                    corner2 = 0
                else:
                    iTriangleLeft2, corner2 = self.SurfaceModel.GetTriangleLeftCorner(i, 2)

                #c = sComma + str(corner0)+str(corner1)+str(corner2)
                s = str(i) + sComma + str(iVertex0) + sComma + str(iVertex1) + sComma + \
                    str(iVertex2) + sComma + \
                    str(iTriangleLeft0) + sComma + \
                    str(iTriangleLeft1) + sComma + \
                    str(iTriangleLeft2) + "\n"
                f.write(s)

    def CornerFromVertex(self, triangles, tri,  iv):
        for c in range(3):
            if triangles.VertexIndex(tri, c) == iv:
                return c
        raise Exception("Vertex not in triangle")

    def ReadSurfaceFile(self, fileName):
        wv = self.currentProject[Project.FixedSerial.WorldView]
        surName = Path.GetFileNameWithoutExtension(fileName)
        surName = Model3D.GetUniqueName(surName, None, wv) #make sure name is unique
        newSurface = wv.Add(clr.GetClrType(Model3D))
        newSurface.Name = surName
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.SubStep, self.Caption)

        with StreamReader(fileName) as s:
            nVertices = int(s.ReadLine())
            nTri = int(s.ReadLine())
            for i in range(nVertices):
                splits = s.ReadLine().Split(',')
                x = float(splits[1])
                y = float(splits[2])
                z = float(splits[3])
                newSurface.Gem.AddVertex(False, GemVertexType.Control, 0, Point3D(x,y,z))

            triangles = newSurface.Gem.GetTriangleList()
            vertexList = newSurface.Gem.GetVertexList()           
            for i in range(nTri):
                splits = s.ReadLine().Split(',')
                vertexes = (int(splits[1]), int(splits[2]), int(splits[3]))
                leftTriangles = (int(splits[4]), int(splits[5]), int(splits[6]))
                t = newSurface.Gem.AddTriangle(0, False, vertexes[0], vertexes[1], vertexes[2])
                for c in range(3):
                    triLeft = leftTriangles[c]
                    triangles.LeftTriangle(t, c, triLeft, 0)
                    if triLeft == -1:
                        triangles.OuterSide(t, c, True)

            # we now need to fix the left triangle property
            edges = {}
            for t in range(nTri):
                for c in range(3):
                    triLeft, cornerLeft = triangles.LeftTriangle(t, c)
                    if triLeft == -1:
                        triangles.LeftTriangle(t, c, 0, 0)
                        v = triangles.VertexIndex(t,GemCorner.Next(c))
                        edges[v] = (t, c)
                    else:
                        v = triangles.VertexIndex(t,GemCorner.Next(c))
                        corner = self.CornerFromVertex(triangles, triLeft,  v)
                        triangles.LeftTriangle(t, c, triLeft, GemCorner.Next(corner))

            for item in edges:
                # the key is "to" vertex from triangle edge
                (t,c) = edges[item]
                v = triangles.VertexIndex(t,c)
                (tLeft, cLeft) = edges[v]
                vertexList.LastTriangle(v, tLeft, cLeft)

        newSurface.Color = Color.Red
        newSurface.Gem.IsConstructed = True