﻿"""
TBC Macro to compare one surface to another
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
clr.AddReference("Trimble.Vce.UI.Controls")
        
from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox

from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd

#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "SurfaceToGrid"
    cmdData.CommandName = "SurfaceToGrid"
    cmdData.Caption = "_Surface To Grid"
    cmdData.UIForm = "SurfaceToGrid"
    try:
        cmdData.Version = 1.53
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r"https://community.trimble.com/groups/trimble-business-center-hce/blog/2018/11/29/tml-surfacetogrid"
        cmdData.HelpFile = cmdData.MacroInfo
    except:
        pass   

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class SurfaceToGrid(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\SurfaceToGrid.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject        
        self.macroName = "Surface To Grid"
        self.tmlFileFolder = macroFileFolder


    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        types = Array[Type](SurfaceTypeLists.AllWithCutFillMap)
        self.surface.FilterByEntityTypes = types
        self.refSurface.FilterByEntityTypes = types   
        self.refSurface.SelectIndex(0)
        self.boundaryObj.AddTypeFilter(clr.GetClrType(IPolyseg))
        self.doDiffCalcs.Checked += self.calcModeChanged
        self.doDiffCalcs.Unchecked += self.calcModeChanged
        self.reportTolCB.Unchecked += self.reportModeChanged
        self.reportTolCB.Checked += self.reportModeChanged

        self.brgCtl.Direction = Math.PI / 2.0
        wv = self.currentProject[Project.FixedSerial.WorldView]
        self.coordCtl.SetCoordinate(Point3D(0.0, 0.0), self.currentProject, wv.CoordinateSystemDefinition)
        self.gridSpacingCtl.Distance = OptionsManager.GetDouble("SurfaceToGrid:Grid", Double.NaN)
        self.gridSpacingAltCtl.Distance = OptionsManager.GetDouble("SurfaceToGrid:GridAlt", Double.NaN)
        self.posTolCtl.DisplayUnit = LinearType.DisplaySmall
        #self.posTolCtl.FormatProperty.AddSuffix = ControlBoolean.True
        self.negTolCtl.DisplayUnit = LinearType.DisplaySmall
        #self.negTolCtl.FormatProperty.AddSuffix = ControlBoolean.True

        self.posTolCtl.Distance = OptionsManager.GetDouble("SurfaceToGrid:PosTol", 0.005)
        self.negTolCtl.Distance = OptionsManager.GetDouble("SurfaceToGrid:NegTol", 0.005)

        self.doDiffCalcs.IsChecked = OptionsManager.GetBool("SurfaceToGrid:DoDiff", True)
        self.calcModeChanged(None, None)
        self.reportTolCB.IsChecked = OptionsManager.GetBool("SurfaceToGrid:ReportTol", True)
        self.reportModeChanged(None, None)

        self.refAlignment.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.refAlignment.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd

        self.algnBased.IsChecked = OptionsManager.GetBool("SurfaceToGrid:AlgGrid", False)
        self.algnBased.Checked += self.algnModeChanged
        self.algnBased.Unchecked += self.algnModeChanged
        self.algnModeChanged(None, None)
        self.createXYZ.IsChecked = OptionsManager.GetBool("SurfaceToGrid:createXYZ", True)
        self.createCSV.IsChecked = OptionsManager.GetBool("SurfaceToGrid:createCSV", False)

    def algnModeChanged(self, sender, args):
        if self.algnBased.IsChecked:
            self.algnControls.Visibility = Visibility.Visible
            self.gridControls.Visibility = Visibility.Collapsed
        else:
            self.algnControls.Visibility = Visibility.Collapsed
            self.gridControls.Visibility = Visibility.Visible

    def calcModeChanged(self, sender, args):
        if self.doDiffCalcs.IsChecked:
            self.calcControls.Visibility = Visibility.Visible
        else:
            self.calcControls.Visibility = Visibility.Collapsed

    def reportModeChanged(self, sender, args):
        if self.reportTolCB.IsChecked:
            self.negTolCtl.IsEnabled = True
            self.posTolCtl.IsEnabled = True
        else:
            self.negTolCtl.IsEnabled = False
            self.posTolCtl.IsEnabled = False

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def GetLimitFromAlignment(self, poly):
        top = int.MinValue
        right = top
        bottom = int.MaxValue
        left = bottom
        startSta = self.alignPolyseg.BeginStation
        endSta = self.alignPolyseg.LastNode.Station
        baseline = self.alignPolyseg.Clone()
        baseline.Extend(1000000000.0) # extend a lot so we don't miss when computing offset

        for p in poly.Point3Ds():
            rtn = baseline.FindPointFromPoint(p, None, None)
            if not rtn[0]:
                continue
            station = rtn[4]
            if station < startSta:
                station = startSta
            if station > endSta:
                station = endSta
            offset = rtn[5]

            n = int(station / self.gridSize)
            if n < bottom:
                bottom = n
            if n > top:
                top = n

            n = int(offset / self.gridSizeAlt)
            if n < left:
                left = n
            if n > right:
                right = n

        return (left, right, bottom, top)

    def GetLimitFromPolyseg(self, poly):
        # compute the left/right and from/to limits.  The return values
        # are the number of gridSize lines from the origin
        top = int.MinValue
        right = top
        bottom = int.MaxValue
        left = bottom
        baseline = PolySeg()
        baseline.AddLine(self.pOrigin, self.pOrigin + self.brg, 0.0, 1.0)
        baseline.Extend(1000000000.0) # extend a lot so we don't miss when computing offset

        for p in poly.Point3Ds():
            rtn = baseline.FindPointFromPoint(p, None, None)
            station = rtn[4]
            offset = rtn[5]

            n = int(station / self.gridSize)
            if n < bottom:
                bottom = n
            if n > top:
                top = n

            n = int(offset / self.gridSizeAlt)
            if n < left:
                left = n
            if n > right:
                right = n

        return (left, right, bottom, top)

    def ComputePointsFromLine(self, linePolyseg):
        for node in linePolyseg.Nodes():
            if not node.Next: # check if last node
                continue
            if not node.Visible: # we only want segments inside polyseg
                continue
            staEnd = node.Next.Station + Point3D.Tolerance      
            sta = node.Station
            sta = Math.Ceiling((sta - Point3D.Tolerance) / self.gridSizeAlt) * self.gridSizeAlt
            while sta <= staEnd:
                rtn = linePolyseg.FindPointFromStation(sta)
                sta += self.gridSizeAlt
                p = rtn[1] # get point x, y from polyseg
                found,p = self.SurfaceModel.PickSurface(p)# get point from surface
                toleranceState = "N/A"
                if found:
                    originalZ = p.Z
                    referenceZ = 0
                    if self.refSurfaceModel:
                        toleranceState = "OK"
                        # we have a second surface.  Compute diff
                        # save elevation from surface
                        z = p.Z

                        # get point from reference surface
                        found,p = self.refSurfaceModel.PickSurface(p)
                        referenceZ = p.Z
                        if not found:
                            continue

                        # calculate difference from surface and reference surface
                        z = p.Z - z
                        # count total above and below so we can do a rough volume calc (based on grid size)
                        if z > 0:
                            self.TotalAbove = self.TotalAbove + z
                        if z < 0:
                            self.TotalBelow = self.TotalBelow - z

                        # construct point to report out
                        p = Point3D(p.X, p.Y, z)

                        if z < self.negTol:
                            self.NumPointsBelow = self.NumPointsBelow + 1
                            toleranceState = "BELOW"
                        elif z > self.posTol:
                            self.NumPointsAbove = self.NumPointsAbove + 1
                            toleranceState = "ABOVE"
                        elif self.skipPointInTol:
                            continue # ignore points that are in tolerance

                    self.NumPoints = self.NumPoints + 1
                    sComma = str(self.cultureInfo) # could be period in some regions
                    if self.file:
                        s = format(p.X * self.unitFactor, "0.5f") + sComma + format(p.Y * self.unitFactor, "0.5f") + \
                           sComma + format(p.Z * self.unitFactor, "0.5f") + "\n"
                        self.file.write(s)
                    if self.csvFile:
                        sCsv = format(p.X * self.unitFactor, "0.5f") + sComma + format(p.Y * self.unitFactor, "0.5f") + \
                           sComma + format(originalZ * self.unitFactor, "0.5f") + sComma + \
                           format(referenceZ * self.unitFactor, "0.5f") + sComma + format(p.Z * self.unitFactor, "0.5f") + \
                           sComma + toleranceState + "\n"
                        self.csvFile.write(sCsv)

    def ComputePointsInPolyseg(self, poly):
        if self.algnBased.IsChecked:
            left, right, bottom, top = self.GetLimitFromAlignment(poly)
            if left > right:
                return

            percentStep = 100.0 / (top - bottom + 1)
            percent = 0
            for l in range(bottom, top + 1): 
                station = l * self.gridSize
                p = PolySeg()
                # there are several method with same name but different "out" values
                # so we need to use the Reference to tell python which one to use
                seg = clr.Reference[Segment]()
                t = clr.Reference[float]()
                pcl = clr.Reference[Point3D]()
                v = clr.Reference[Vector3D]()
                if self.alignPolyseg.FindPointFromStation(station, seg, t, pcl, v):
                    pt1 = pcl.Value + v.Value * self.gridSizeAlt * left
                    pt2 = pcl.Value + v.Value * self.gridSizeAlt * right
                    line = PolySeg(Line(pt1, pt2))
                    line.BeginStation = 0
                    line.ComputeStationing()
                    success = clr.Reference[bool]()
                    line.Hide(poly, Side.Out, success)
                    self.ComputePointsFromLine(line)
                    percent = percent + percentStep
                    if ProgressBar.TBC_ProgressBar.SetProgress(int(percent)):
                        break
        else:
            left, right, bottom, top = self.GetLimitFromPolyseg(poly)
            if left > right:
                return
            vRight = Vector3D(self.brg)
            vRight.Rotate90(Side.Right)            
            percentStep = 100.0 / (right - left + 1)
            percent = 0
            for l in range(bottom, top + 1):  
                pt = self.pOrigin + self.brg * self.gridSize * l
                pt1 = pt + vRight * self.gridSizeAlt * left
                pt2 = pt + vRight * self.gridSizeAlt * right
                line = PolySeg(Line(pt1, pt2))
                line.BeginStation = 0
                line.ComputeStationing()
                success = clr.Reference[bool]()
                line.Hide(poly, Side.Out, success)
                if False:
                    # debug code
                    wv = self.currentProject[Project.FixedSerial.WorldView]
                    ls = wv.Add(clr.GetClrType(Linestring))
                    e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                    e.Position = line.FirstNode.Point
                    ls.AppendElement(e)  
                    e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                    e.Position = line.LastNode.Point
                    ls.AppendElement(e) 

                self.ComputePointsFromLine(line)
                percent = percent + percentStep
                if ProgressBar.TBC_ProgressBar.SetProgress(int(percent)):
                    break

    def ComputePointsInRegion(self, region, polySurface1, polySurface2, polyBoundary):
        # check if any regions inside this one
        for r in region.ChildRegions():
            self.ComputePointsInRegion(r, polySurface1, polySurface2, polyBoundary)

        point = region.InsidePoint()
        if polySurface1.PointInPolyseg(point) == Side.Out:
            return
        if polySurface2:
            if polySurface2.PointInPolyseg(point) == Side.Out:
                return
        if polyBoundary:
            if polyBoundary.PointInPolyseg(point) == Side.Out:
                return
        # this region is valid
        poly = region.AreaPolyseg()
        self.ComputePointsInPolyseg(poly)

    def ComputePointsMultipleBoundaries(self):
        rb = RegionBuilder()
        rb.PerformSelfIntersectionTest = False
        rb.Add(self.SurfaceBoundary)

        surface2Poly = None
        if self.refSurfaceModel:
            surface2Poly = DtmIsopachs.FormDtmBoundary(self.refSurfaceModel.Gem)
            rb.Add(surface2Poly)

        if self.bndyPoly:
            rb.Add(self.bndyPoly)

        rb.Build()
        for r in rb.Regions():
            self.ComputePointsInRegion(r, self.SurfaceBoundary, surface2Poly, self.bndyPoly)

    def OkClicked(self, sender, e):
        if self.surface.SelectedSerial == 0:
            self.surface.StatusMessage = "Enter a surface"
            return False

        if self.algnBased.IsChecked:
            if self.refAlignment.SelectedSerial == 0:
                self.refAlignment.StatusMessage = "Enter an alignment"
                return False
        else:
            ctrl = self.coordCtl
            p1 = self.coordCtl.Coordinate
            if p1.IsUndefined :
                self.coordCtl.StatusMessage = "No coordinate defined"
                return False
            self.coordCtl.StatusMessage = ""
            self.pOrigin = p1

            rotation = self.brgCtl.Direction
            if Double.IsNaN(rotation):
                self.brgCtl.StatusMessage = "Enter valid grid direction"
                return False
            self.brgCtl.StatusMessage = ""

            v = Vector3D()
            v.Angle = rotation
            self.brg = v

        gridSize = self.gridSpacingCtl.Distance
        if Double.IsNaN(gridSize) or gridSize <= 0.0:
            self.gridSpacingCtl.StatusMessage = "Enter valid scale factor"
            return False
        self.gridSpacingCtl.StatusMessage = ""
        self.gridSize = gridSize

        gridSize = self.gridSpacingAltCtl.Distance
        if Double.IsNaN(gridSize) or gridSize <= 0.0:
            self.gridSpacingAltCtl.StatusMessage = "Enter valid scale factor"
            return False
        self.gridSpacingAltCtl.StatusMessage = ""
        self.gridSizeAlt = gridSize

        if self.doDiffCalcs.IsChecked:
            if self.refSurface.SelectedSerial == 0:
                self.refSurface.StatusMessage = "Enter a surface"
                return False
            self.posTol = 0
            self.negTol = 0
            if self.reportTolCB.IsChecked:
                # will compute diff so we need tolerances
                posTol = self.posTolCtl.Distance
                if Double.IsNaN(posTol):
                    self.posTolCtl.StatusMessage = "Enter valid tolerance"
                    return False
                self.posTolCtl.StatusMessage = ""
                self.posTol = Math.Abs(posTol)

                negTol = self.negTolCtl.Distance
                if Double.IsNaN(negTol):
                    self.negTolCtl.StatusMessage = "Enter valid tolerance"
                    return False
                self.negTolCtl.StatusMessage = ""
                self.negTol = -Math.Abs(negTol)

        self.bndyPoly = None
        if self.boundaryObj.Entity:
            poly = self.boundaryObj.Entity.ComputePolySeg()
            if not poly or poly.NumberOfNodes < 2:
                self.boundaryObj.StatusMessage = "Enter valid boundary"
                return False
            self.boundaryObj.StatusMessage = ""
            # we need to remove arcs
            settings = Model3DCompSettings.ProvideSettingsObject(self.currentProject)
            poly = poly.Linearize(settings.HorzProximityTolerance, settings.VertProximityTolerance,
                        self.gridSize, None, False)
            self.bndyPoly = poly

        fileName = self.fileNameCtl.Text
        if not fileName:
            # empty name, need to report name error
            self.surface.StatusMessage = "Enter valid file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if fileName.IndexOfAny(Path.GetInvalidPathChars()) != -1:
            self.surface.StatusMessage = "Invalid char in file name"
            Keyboard.Focus(self.fileNameCtl)
            return False
        if not Path.IsPathRooted(fileName):
            fileName = Path.Combine(SearchPath.ExportAppDataPath,fileName)

        if not self.createXYZ.IsChecked and not self.createCSV.IsChecked:
            self.surface.StatusMessage = "Select output file type"
            Keyboard.Focus(self.createXYZ)
            return False

        fileName = Path.ChangeExtension(fileName, ".xyz")
        csvFileName = Path.ChangeExtension(fileName, ".csv")
        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        ProgressBar.TBC_ProgressBar.Title = "Computing Grid"

        try:
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
                self.refSurfaceModel = self.refSurface.SelectedEntity()
                self.SurfaceModel = self.surface.SelectedEntity()
                boundaryOriginal = DtmIsopachs.FormDtmBoundary(self.SurfaceModel.Gem)
                boundaryOriginal.ComputeStationing()
                self.SurfaceBoundary = boundaryOriginal

                self.alignPolyseg = None
                if self.algnBased.IsChecked:
                    self.alignPolyseg = self.refAlignment.SelectedEntity().ComputePolySeg()

                # points are stored in meters, get conversion factor
                self.unitFactor = self.currentProject.Units.Linear.Convert(1.0, LinearType.Display)
                self.cultureInfo = CultureInfo.CurrentCulture.TextInfo.ListSeparator

                self.NumPoints = 0
                self.NumPointsAbove = 0
                self.NumPointsBelow = 0
                self.TotalAbove = 0
                self.TotalBelow = 0
                if self.reportTolCB.IsChecked:
                    self.skipPointInTol = self.tolReport.IsChecked
                else:
                    self.skipPointInTol = False
                if self.createXYZ.IsChecked and not self.createCSV.IsChecked:
                    if File.Exists(fileName):
                        File.Delete(fileName)
                    with open(fileName, 'w') as f:   
                        self.file = f
                        self.csvFile = None
                        if self.refSurface.SelectedSerial != 0 or self.boundaryObj.Entity:
                            self.ComputePointsMultipleBoundaries()
                        else:
                            self.ComputePointsInPolyseg(boundaryOriginal)

                elif not self.createXYZ.IsChecked and self.createCSV.IsChecked:
                    if File.Exists(csvFileName):
                        File.Delete(csvFileName)
                    with open(csvFileName, 'w') as f:
                        self.csvFile = f  
                        self.file = None
                        if self.refSurface.SelectedSerial != 0 or self.boundaryObj.Entity:
                            self.ComputePointsMultipleBoundaries()
                        else:
                            self.ComputePointsInPolyseg(boundaryOriginal)

                elif self.createXYZ.IsChecked and self.createCSV.IsChecked:
                    if File.Exists(fileName):
                        File.Delete(fileName)
                    if File.Exists(csvFileName):
                        File.Delete(csvFileName)
                    with open(csvFileName, 'w') as f2:
                        self.csvFile = f2
                        with open(fileName, 'w') as f:   
                            self.file = f
                            if self.refSurface.SelectedSerial != 0 or self.boundaryObj.Entity:
                                self.ComputePointsMultipleBoundaries()
                            else:
                                self.ComputePointsInPolyseg(boundaryOriginal)

                failGuard.Commit()

        except Exception as ex:
            VceMessageBox.ShowError(str(ex), "Error:" + fileName)
            f = None

        self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        if f:
            OptionsManager.SetValue("SurfaceToGrid:Grid", self.gridSpacingCtl.Distance)
            OptionsManager.SetValue("SurfaceToGrid:GridAlt", self.gridSpacingAltCtl.Distance)
            OptionsManager.SetValue("SurfaceToGrid:PosTol", self.posTolCtl.Distance)
            OptionsManager.SetValue("SurfaceToGrid:NegTol", self.negTolCtl.Distance)
            OptionsManager.SetValue("SurfaceToGrid:DoDiff",  self.doDiffCalcs.IsChecked)
            OptionsManager.SetValue("SurfaceToGrid:ReportTol", self.reportTolCB.IsChecked)
            OptionsManager.SetValue("SurfaceToGrid:AlgGrid", self.algnBased.IsChecked)
            OptionsManager.SetValue("SurfaceToGrid:createXYZ", self.createXYZ.IsChecked)
            OptionsManager.SetValue("SurfaceToGrid:createCSV", self.createCSV.IsChecked)

            f = str(self.NumPoints) + " points exported to file\r" + fileName
            if self.doDiffCalcs.IsChecked:
                area = self.gridSize * self.gridSizeAlt
                cut = self.TotalBelow * area
                fill = self.TotalAbove * area
                cut = self.currentProject.Units.Volume.Format(cut)
                fill = self.currentProject.Units.Volume.Format(fill)
                f = f + "\rCut Volume: " + cut + "   Fill Volume: " + fill
                if self.reportTolCB.IsChecked:
                    f = f + "\rNumber below tolerance: " + str(self.NumPointsBelow) + "\rNumber above tolerance: " + str(self.NumPointsAbove)
                        
            if self.createXYZ.IsChecked:
                f = f + "\r\rDo you want to Import this file?"
                r = VceMessageBox.ShowQuestionYesNo(None, f, "Grid points computed")
                if r == DialogResult.Yes:
                    TrimbleOffice.TheOffice.CommandProcessor.ExecuteCommand("Import", fileName)
            else:
                VceMessageBox.ShowWarning(None, f, "Grid points computed")

        sender.CloseUICommand()