﻿"""
TBC Macro to test macro callback feature
"""
import clr
# Reference the WPF assemblies
clr.AddReference('IronPython.Wpf')
import wpf
from System import Guid
from System.IO import StreamReader
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap

from Trimble.Vce.Core import TransactMethodCall
from Trimble.Vce.Interfaces import Client
from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IViewMember, ISettableName, IName, IViewSite, ISnapInHelpers
from Trimble.Vce.ForeignCad import MText, Text
from Trimble.Vce.Alignment.Linestring import Linestring
from Trimble.Vce.Macros import GenericContainer

from Trimble.DiskIO import OptionsManager

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Triangle3D, Triangle2D
from Trimble.Vce.Geometry.PolySeg import PolySeg
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import EditTextCmd
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.Core import GlobalSelection
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
clr.AddReference("Trimble.Vce.UI.Controls")
try:
    from Trimble.Sdk.UI import InputMethod
except:
    from Trimble.Sdk.Interfaces.UI import InputMethod
      
def UpdateOnChange(entity):
    list = entity.IsObservedBy()
    for txt in list:
        if isinstance(txt, clr.GetClrType(MText)):
            break
    else:
        return
    for o in entity:
        if isinstance(o, clr.GetClrType(Linestring)):
            n = o.ElementCount
            for i in range(n):
                o.RemoveElementAt(0) # just delete the first one each time
            poly = PolySeg()
            pts = txt.GetTextBoxBounds()
            poly.Add(pts)
            poly.Add(pts[0]) # add first point again
            o.Append(poly, None, False, False)
            break

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "TextBox"
    cmdData.CommandName = "TextBox"
    cmdData.Caption = "_TextBox"
    cmdData.UIForm = "TextBox"
    cmdData.HelpFile="Macros.chm"
    cmdData.HelpTopic="22206"

    try:
        cmdData.DefaultTabKey = "Macros"
        cmdData.DefaultTabGroupKey = "CAD"
        cmdData.ShortCaption = "TextBox"
        cmdData.Version = 1.0
        cmdData.MacroAuthor = "Trimble"
    except:
        pass
    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

MacroCode = """
import clr
clr.AddReference("Trimble.Vce.ForeignCad")
from Trimble.Vce.ForeignCad import MText, Text
clr.AddReference("Trimble.Vce.Alignment")
from Trimble.Vce.Alignment.Linestring import Linestring
clr.AddReference("Trimble.Vce.Geometry")
from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D, Triangle3D, Triangle2D
from Trimble.Vce.Geometry.PolySeg import PolySeg

def UpdateOnChange(entity):
    list = entity.IsObservedBy()
    for txt in list:
        if isinstance(txt, clr.GetClrType(MText)):
            break
    else:
        return
    for o in entity:
        if isinstance(o, clr.GetClrType(Linestring)):
            n = o.ElementCount
            for i in range(n):
                o.RemoveElementAt(0) # just delete the first one each time
            poly = PolySeg()
            pts = txt.GetTextBoxBounds()
            poly.Add(pts)
            poly.Add(pts[0]) # add first point again
            o.Append(poly, None, False, False)
            break
"""

class TextBox(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\TextBox.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject
        self.macroId = "1cce2230-51e0-45af-880b-c6c632de9fe3"
        self.macroVersion = 1.03

    def OnLoad(self, cmd, buttons, event):
        self.Caption = cmd.Command.Caption
        self.tObj1.AddTypeFilter(clr.GetClrType(MText))
        self.tObj1.ValueChanged += self.TextChanged
 
    def TextChanged(self, ctrl, e):
        # set global selection to selected text
        if e.Cause == InputMethod.Mouse:   
            self.OkClicked(None, None)

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def OkClicked(self, sender, e):
        txt = self.tObj1.Entity
        if not txt:
            self.tObj1.StatusMessage = "Select a Text Object"
            return False

        self.currentProject.TransactionManager.AddBeginMark(Client.CommandGranularity.Command, self.Caption)
                    
        # the "with" statement will unroll any changes if something go wrong
        with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:   
            sn, container = ISnapInHelpers.GetMemberManagement(txt)
            observers = self.currentProject.Concordance.GetObserversOf(txt.SerialNumber)
            l = None
            for o in observers:
                if isinstance(o, clr.GetClrType(GenericContainer)):
                    macroContainer = o
                    #macroContainer.MacroGuid = Guid.Empty # set to empty to disable caching
                    macroContainer.MacroName = "TextBox"
                    if macroContainer.GetCallbackVersion() != self.macroVersion:
                        macroContainer.SetCallback(self.macroVersion, MacroCode)
                    for si in macroContainer:
                        if isinstance(si, clr.GetClrType(Linestring)):
                            l = si
                            break
                    break
            else:
                macroContainer = container.Add(clr.GetClrType(GenericContainer)) 
                macroContainer.MacroGuid = Guid(self.macroId)
                macroContainer.MacroName = "TextBox"
                macroContainer.SetCallback(self.macroVersion, MacroCode)
                
            if not l:
                l = macroContainer.Add(clr.GetClrType(Linestring))
            macroContainer.AddDependancy(txt.SerialNumber)
            macroContainer.CallCallback()
            #UpdateOnChange(macroContainer)            
            
            failGuard.Commit()

            self.currentProject.TransactionManager.AddEndMark(Client.CommandGranularity.Command)
