﻿"""
Command to allow quickly toggle a list of ribbon tabs
"""
import clr
clr.AddReference('IronPython.Wpf')
import wpf
clr.AddReference("PresentationFramework")
clr.AddReference("PresentationCore")
clr.AddReference("windowsbase")
from System.Windows import  Window
from System.Windows.Controls import CheckBox
from System.Collections.Generic import List
from System.IO import StreamReader
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import Control, Keys
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from Trimble.DiskIO import OptionsManager
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import * # so we can get the ribbon
HiddenToggleGroup = "HiddenToggleGroup"

def Setup(cmdData, macroFileFolder):
    cmdData.Key = "ToggleTabs"
    cmdData.CommandName = "ToggleRibbonTabs"
    cmdData.Caption = "_Toggle Ribbon Tabs"

    try:
        cmdData.Version = 1.01
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""

        cmdData.ToolTipTitle = "Toggle Tabs"
        cmdData.ToolTipTextFormatted = "Command to toggle tabs which are not often used. Use Ctrl key to configure"
        cmdData.QuickAccessMenuLocation = "___" # using undefined key will append cmd
        cmdData.EnableNoProject = True
        cmdData.HelpFile="Macros.chm"
        cmdData.HelpTopic="22207"

        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

def Execute(cmd, currentProject, macroFileFolder, parameters): 
    def HasRibbonBeenConfigured():
        # since this method is "inside" Execute, we have access to "Excute's" properties
        tab = ribbonTabs[0] # get the first tab
        if tab.Groups.IndexOf(HiddenToggleGroup) == -1:
            # if new ribbon, clear out old options
            OptionsManager.SetValue("ToggleRibbonTabs.Tabs", 0)
            return False
        if not OptionsManager.GetInt("ToggleRibbonTabs.Tabs", 0):
            return False
        return True

    hasControl = Control.ModifierKeys & Keys.Control
    ribbon = TrimbleOffice.TheOffice.MainWindow.AppToolbar.Ribbon
    ribbonTabs = ribbon.Tabs
    if not HasRibbonBeenConfigured() or hasControl:
        form = ToggleRibbonDialog(ribbonTabs, macroFileFolder).ShowDialog()
        return

    tabs = OptionsManager.GetInt("ToggleRibbonTabs.Tabs",0)
    ConfigureRibbon(ribbonTabs, tabs, False)

def ConfigureRibbon(ribbonTabs, tabToggleBits, fromDialog):
    # we will add a hidden group to the first tab
    tab = ribbonTabs[0] # get the first tab
    if tab.Groups.IndexOf(HiddenToggleGroup) == -1:
        g = tab.Groups.Add(HiddenToggleGroup)
        # empty groups will be removed so we need to add a command
        g.Tools.AddTool("Delete")
        g.Visible = False

    prevTabs = OptionsManager.GetInt("ToggleRibbonTabs.Tabs",0)
    for t in ribbonTabs:
        index = t.Index
        showToggle = tabToggleBits & 1 << index
        wasToggle = prevTabs & 1 << index
        if showToggle:
            if fromDialog:
                t.Visible =False
            else:
                t.Visible = not t.Visible
        elif wasToggle:
            t.Visible = True

    return True

class ToggleRibbonDialog(Window): # this inherits from the WPF window control
    def __init__(self, ribbonTabs, macroFileFolder):
        with StreamReader(macroFileFolder + r"\ToggleTabs.xaml") as s:
            wpf.LoadComponent(self, s) 
        self.ribbonTabs = ribbonTabs # save the ribbon
        self.Ok.Click += self.Button_Click_Ok
        self.Cancel.Click += self.Button_Click_Cancel
        self.Items =[]# List[str]() # we want .net strings
        self.Keys=[]
        
        tabs = OptionsManager.GetInt("ToggleRibbonTabs.Tabs",0)
        for r in ribbonTabs:
            thisTabToggles = tabs & 1 << r.Index
            if not r.Visible:                
                if not thisTabToggles:
                    continue
            c = CheckBox()
            c.Content=r.Caption
            c.IsChecked = thisTabToggles != 0
            self.Items.Add(c)
            self.Keys.Add(r.Key)
        self.ListOfTabs.ItemsSource = self.Items  
        pass

    def Button_Click_Ok(self, sender, e):
            # save tab visible status
            bits = 0
            k=-1
            for i in self.Items:
                k=k+1
                # skip tabs not visible
                if  not i.IsChecked:
                    continue
                index = self.ribbonTabs.IndexOf(self.Keys[k])
                if index >= 0:
                    bits = bits | 1 << index
            ConfigureRibbon(self.ribbonTabs, bits, True)
            OptionsManager.SetValue("ToggleRibbonTabs.Tabs", bits)
            self.DialogResult = True
            self.Close()
            pass

    def Button_Click_Cancel(self, sender, e):
        self.Close()
        pass