﻿"""
Macro to show floating toolbar
"""
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf
clr.AddReference("PresentationFramework")
clr.AddReference("PresentationCore")
clr.AddReference("windowsbase")
from System.Windows import  Window, MessageBox, Visibility, WindowStyle, Int32Rect, Thickness, SystemParameters, Application
from System.Windows.Interop import Imaging, WindowInteropHelper
from System.Windows.Controls import (Button, Image, ListBoxItem, Label, StackPanel, Orientation, 
    ContextMenu, MenuItem)
from System.Windows.Media.Imaging import BitmapSizeOptions
from System import Environment, IntPtr
from System.IO import StreamReader, File, FileMode, FileStream, FileAccess, FileShare
from System.Runtime.Serialization.Formatters.Binary import BinaryFormatter
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap

from Trimble.Vce.Interfaces import *
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands.CommandBase import CleanCaption
clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.Controls")
from Trimble.Vce.UI.Controls import VceMessageBox


def Setup(cmdData, macroFileFolder):
    cmdData.Key = "Toolbar"
    cmdData.CommandName = "Toolbar"
    cmdData.Caption = "_Show My Toolbar"
    cmdData.EnableNoProject = True
    cmdData.HasStartup = True
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "22208"

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

def CreateWindow(macroFileFolder, cmds, autoStart, windowStyle):
    form = FloatingToolbar(cmds, macroFileFolder)
    form.Topmost=False
    #form.ShowInTaskbar=False # set to False if you don't want to show in TaskBar
    form.autoStart = autoStart
    form.WindowStyle = windowStyle
    # since TBC is a WinForm app and we are using WPF, we need to do a
    # few tricks to get the window parented correctly.
    interop =  WindowInteropHelper(form)
    interop.EnsureHandle() # Make sure we have a handle (creates one if its missing)  
    interop.Owner = TrimbleOffice.TheOffice.MainWindow.Form.Handle #make TBC our parent   
    return form

def StartupExecute(macroFileFolder):
    data = ReadToolbarLayout(macroFileFolder) # this will be an empty list first time
    if data:
        f = data[0]
        if f[4]: #autostart set
            cmds = data[1]
            form = CreateWindow(macroFileFolder, cmds, f[4], f[5])
            SetWindowLocation(form, f[0],f[1],f[2],f[3])
            form.Show()

def Execute(cmd, currentProject, macroFileFolder, parameters): 
    # check if we are already running
    if Application.Current:
        windows = Application.Current.Windows
        for w in windows:
            if isinstance(w, FloatingToolbar):
                MessageBox.Show("Toolbox already open.", "Toolbox")
                return
    data = ReadToolbarLayout(macroFileFolder) # this will be an empty list first time
    if data:
        cmds = data[1]
    else:
        cmds = []
    if data:
        f = data[0]
        form = CreateWindow(macroFileFolder, cmds, f[4], f[5])
        SetWindowLocation(form, f[0],f[1],f[2],f[3])
    else:
        form = CreateWindow(macroFileFolder, cmds, True, WindowStyle.ToolWindow)
    form.Show()

    return

class FloatingToolbar(Window): # this inherits from the WPF window control
    def __init__(self, cmds, macroFileFolder):
        with StreamReader(macroFileFolder + r"\Toolbar.xaml") as s:
            wpf.LoadComponent(self, s) 
        # add buttons
        cmdProcessor = TrimbleOffice.TheOffice.CommandProcessor
        cmdList = cmdProcessor.CommandList
        for cmd in cmds:
            self.AddButton(cmdList, cmd)

        self.panel.UpdateLayout()# panel defined in XAML
        self.UpdateLayout()
        self.MouseDoubleClick += self.ConfigureToolbar
        self.Closing += self.ToolbarClosing
        self.SizeChanged += self.ToolbarChanged
        self.LocationChanged += self.ToolbarChanged

        self.Loaded += self.window_Loaded

        contextmenu = ContextMenu()
        self.ContextMenu = contextmenu
        mi = MenuItem()
        mi.Header = "Customize My Toolbar"
        mi.Click+=self.ConfigureToolbar
        contextmenu.Items.Add(mi)
        mi = MenuItem()
        mi.Header = "Toggle Border Style"
        mi.Click+=self.ContextMenu_ToggleBorder
        contextmenu.Items.Add(mi)
        mi = MenuItem()
        mi.Header = "Save Layout"
        mi.Click+=self.ContextMenu_SaveLayout
        contextmenu.Items.Add(mi)
        mi = MenuItem()
        mi.Header = "Close Window"
        mi.Click+=self.ContexMenu_CloseWindow
        contextmenu.Items.Add(mi)

    def window_Loaded(self, sender,  e):
        self.Changed = False

    def ContextMenu_ToggleBorder(self, sender, e):
        if self.WindowStyle == WindowStyle['None']:
            self.WindowStyle = WindowStyle.ToolWindow
        else:
            self.WindowStyle = WindowStyle['None']

    def ContexMenu_CloseWindow(self, sender, e):
        self.Close()

    def ContextMenu_SaveLayout(self, sender, e):
        if self.Changed:
            cmds = []
            for btn in self.panel.Children:
                c = btn.Tag
                cmds.Add(c)

            WriteToolbarLayout(self, cmds)
        self.Changed = False

    def ToolbarChanged(self, sender, e):
        self.Changed = True

    def ToolbarClosing(self, sender, e):
        # this is NOT called when the parent window (TBC) closes (and then this
        # window closes)
        if self.Changed:
            cmds = []
            for btn in self.panel.Children:
                c = btn.Tag
                cmds.Add(c)

            WriteToolbarLayout(self, cmds)

    def ButtonClicked(self, sender, e):
        cmdProcessor = TrimbleOffice.TheOffice.CommandProcessor
        cmdProcessor.ExecuteCommand(sender.Tag) # "Tag" holds the command key
        pass

    def AddButton(self, cmdList, cmdKey):
        c = cmdKey.lower()
        if cmdList.ContainsKey(c):
            cmd = cmdList[c]
            b = Button()
            b.Width = 30
            b.Margin = Thickness(3)
            b.Tag = cmd.Key.lower()
            caption = CleanCaption(cmd.Caption)
            b.ToolTip = caption
            b.Click+= self.ButtonClicked
            # button shows command icon
            # since command images are defined as bitmaps, we need to jump
            # thru
            # a few hoops to create a wpf image
            bitmap = cmd.ImageSmall
            if bitmap:
                interopImage = Imaging.CreateBitmapSourceFromHBitmap(bitmap.GetHbitmap(), IntPtr.Zero,
                                Int32Rect.Empty, BitmapSizeOptions.FromWidthAndHeight(bitmap.Width, bitmap.Height))
                image = Image()
                image.Source = interopImage
                b.Content = image
            else:
                b.Content = caption

            self.panel.Children.Add(b)

    def ConfigureToolbar(self, sender, e):
        form = ToolbarConfigureDialog(self)
        try:
            form.ShowDialog()
        except Exception as ex:
            VceMessageBox.Show(TrimbleOffice.TheOffice.MainWindow.Form, str(ex))
            return

def ReadToolbarLayout(macroFileFolder):
    file = macroFileFolder + r"\Toolbar.Layout"
    try:
        if not File.Exists(file):
            return None
        
        formatter = BinaryFormatter()
        with  FileStream(file, FileMode.Open,
                FileAccess.Read, FileShare['None']) as stream:
            height = formatter.Deserialize(stream)
            width = formatter.Deserialize(stream)
            left = formatter.Deserialize(stream)
            top = formatter.Deserialize(stream)
            autoStart = formatter.Deserialize(stream)
            winStyle = formatter.Deserialize(stream)
            cmds = formatter.Deserialize(stream)
            stream.Close()
 
        return ((height,width,left,top,autoStart,winStyle), cmds)

    except:
        File.Delete(file)
        return None

def WriteToolbarLayout(form, cmds):
    file = MacroFileFolder + r"\Toolbar.Layout"
    try:
        if File.Exists(file):
            File.Delete(file)

        formatter = BinaryFormatter()
        with  FileStream(file, FileMode.OpenOrCreate,
                FileAccess.Write, FileShare['None']) as stream:
            formatter.Serialize(stream, form.Height)
            formatter.Serialize(stream, form.Width)
            formatter.Serialize(stream, form.Left)
            formatter.Serialize(stream, form.Top)
            formatter.Serialize(stream, form.autoStart)
            formatter.Serialize(stream, form.WindowStyle)

            formatter.Serialize(stream, cmds)
            stream.Close()

    except:
        File.Delete(file)
  
def SetWindowLocation(form, height,width,left,top):
    if height > SystemParameters.VirtualScreenHeight:
        height = SystemParameters.VirtualScreenHeight
    if width > SystemParameters.VirtualScreenWidth:
        width = SystemParameters.VirtualScreenWidth
    form.Width = width
    form.Height = height
    form.Top = top
    form.Left = left

class ToolbarConfigureDialog(Window):
    def __init__(self, toolBarWindow):
        self.toolBarWindow = toolBarWindow # save toolbar window so we can update it later
        with StreamReader(MacroFileFolder + r"\ToolbarConfigure.xaml") as s:
            wpf.LoadComponent(self, s) 

        cmdProcessor = TrimbleOffice.TheOffice.CommandProcessor
        self.cbStartup.IsChecked = self.toolBarWindow.autoStart
        self.commands = cmdProcessor.CommandList
        for btn in self.toolBarWindow.panel.Children:
            self.AddCmdToList(btn.Tag)
    
        # build list of commands so we can sort
        g = {}      
        for c in self.commands.Values:
            if not c.Hidden:
                s = CleanCaption(c.Caption)
                k = c.Key.lower()
                g[s]=k

        validCmds = []
        self.validKeys = []
        for key, value in sorted(g.items()):
             validCmds.Add(key)
             self.validKeys.Add(value)
        self.cmdList.ItemsSource = validCmds
        self.CloseBtn.Click+=self.Button_Click_Close
        self.OkBtn.Click+=self.Button_Click_Add
        self.DelBtn.Click+=self.Button_Click_Delete

    def Button_Click_Close(self, sender, e):
        self.toolBarWindow.panel.Children.Clear() # remove any existing toolbar items
        self.toolBarWindow.autoStart = self.cbStartup.IsChecked
        self.toolBarWindow.Changed = True # set this to save updates
        for i in self.listBox.Items:
            self.toolBarWindow.AddButton(self.commands, i.Tag)
        self.Close()
        pass

    def AddCmdToList(self, cmdKey):
        l = ListBoxItem()
        cmd = self.commands[cmdKey]
        l.Tag = cmdKey
        sp = StackPanel()
        sp.Orientation = Orientation.Horizontal
        label = Label()
        label.Content = CleanCaption(cmd.Caption)
        #sp.Children.Add(label)

        l.Content = label
        self.listBox.Items.Add(l)

    def Button_Click_Add(self, sender, e):
        k = self.cmdList.SelectedIndex
        if k >= 0:
            self.AddCmdToList(self.validKeys[k])

    def Button_Click_Delete(self, sender, e):
        if self.listBox.SelectedItem:
            list = []
            for i in self.listBox.SelectedItems:
                list.Add(i)
            for i in list:
                self.listBox.Items.Remove(i)