﻿"""
TBC Macro to compute boundary lines from selection of lines
"""
#region
# Reference the WPF assemblies
import clr
clr.AddReference('IronPython.Wpf')
import wpf

from Trimble.Vce.Core.Components import WorldView, Project, Layer
from Trimble.Vce.Core import TransactMethodCall
from System import Type, Tuple, Array, Double, Math, String
from System.Globalization import CultureInfo
clr.AddReference("System.Windows.Forms")
from System.Windows.Forms import DialogResult
clr.AddReference("System.Drawing")
from System.Drawing import Bitmap
from System.IO import StreamReader, Path, File
from System.Collections.Generic import List
from System.Windows import  Window, ResourceDictionary, Visibility
from System.Windows.Controls import StackPanel
from System.Windows.Input import Keyboard

from Trimble.Vce.Interfaces.Units import LinearType
from Trimble.Vce.Interfaces.SnapIn import IPolyseg
from Trimble.Vce.Interfaces.Client import CommandGranularity
from Trimble.Vce.Interfaces import ProgressBar

from Trimble.DiskIO import SearchPath, OptionsManager

from Trimble.Vce.Gem import Model3D, DtmIsopachs, Model3DCompSettings

from Trimble.Vce.Alignment import ProfileView
from Trimble.Vce.Alignment.Linestring import Linestring,VerticalElementFactory, IVerticalNoCurve, IStationVerticalLocation
from Trimble.Vce.Alignment.Linestring import ElementFactory, IStraightSegment, IXYZLocation,ITangentArcSegment, ArcTangentType

from Trimble.Vce.Geometry import Point3D, Line2D, Side, Vector3D
from Trimble.Vce.Geometry.PolySeg import PolySeg
from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line
from Trimble.Vce.Geometry.Region import Region, RegionBuilder
clr.AddReference("Trimble.Vce.UI.Construction")
from Trimble.Vce.UI.Construction import ThicknessUtilities

clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents
clr.AddReference("Trimble.Vce.UI.Controls")     
from Trimble.Vce.Core import GlobalSelection
from Trimble.Vce.UI.Controls import SurfaceTypeLists, VceMessageBox

from Trimble.Vce.UI.UIManager import TrimbleOffice
clr.AddReference("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import CreateHorizontalAlignmentCmd
clr.AddReference("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import SelectionContextMenuHandler, ViewHelper
#endregion
def Setup(cmdData, macroFileFolder):
    cmdData.Key = "TrackLineEdge"
    cmdData.CommandName = "TrackLineEdge"
    cmdData.Caption = "_Track Line Edge"
    cmdData.UIForm = "TrackLineEdge"
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "xxx"
    try:
        cmdData.Version = 1.62
        cmdData.MacroAuthor = "Trimble"
        cmdData.MacroInfo = r""
    except:
        pass   

    try:
        b = Bitmap(macroFileFolder + "\\" + cmdData.Key + ".png")
        cmdData.ImageSmall = b
    except:
        pass

class TrackLineEdge(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader(macroFileFolder + r"\TrackLineEdge.xaml") as s:
            wpf.LoadComponent(self, s)
        self.currentProject = currentProject

    def OnLoad(self, cmd, buttons, event):
        buttons[0].Content = "Apply"
        buttons[1].Content = "Close"
        self.Caption = cmd.Command.Caption
        self.objs.IsEntityValidCallback = self.IsValid
        optionMenu = SelectionContextMenuHandler()
        # remove options that don't apply here
        optionMenu.ExcludedCommands = "SelectObservations | SelectPoints | SelectDuplicatePoints"
        self.objs.ButtonContextMenu = optionMenu

        self.edgeLinesCB.IsChecked = OptionsManager.GetBool("TrackLineEdge:edgeLinesCB", True)
        self.edgeLayerMaskCtl.Text = OptionsManager.GetString("TrackLineEdge:edgeLayerMaskCtl", "{L}-Edge")
        self.bndyCB.IsChecked = OptionsManager.GetBool("TrackLineEdge:bndyCB", False)
        self.bndyLayerMaskCtl.Text = OptionsManager.GetString("TrackLineEdge:bndyLayerMaskCtl", "{L}-Bndy")

        self.polyType = clr.GetClrType(IPolyseg)
        self.lType = clr.GetClrType(Linestring)
        self.Profile = clr.GetClrType(ProfileView)

        self.alignmentCtl.SetEntityType(clr.GetClrType(ProfileView), self.currentProject)
        self.alignmentCtl.NewCmdCallback = CreateHorizontalAlignmentCmd.RunThisCmd

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        return self.IsValidObj(o)

    def IsValidObj(self, obj):
        if isinstance(obj, self.Profile):
            return False
        if isinstance(obj, self.polyType):
            return True
        return False

    def CancelClicked(self, sender, args):
        sender.CloseUICommand()

    def UpdateLayer(self, layerSerial, layers, layerString):
        layerObj = layers.Lookup(layerSerial)
        if not layerObj:
            return layers.Lookup(Project.FixedSerial.LayerZero)
        newLayerName = layerString.Replace("{L}", layerObj.Name)
        layerNew = layers.GetLayer(newLayerName)
        if layerNew:
            return layerNew
        # layer doesn't exist
        newLayer = layers.Add(clr.GetClrType(Layer))
        newLayer.Name = newLayerName
        newLayer.DefaultColor = layerObj.DefaultColor
        newLayer.LineWeight = layerObj.LineWeight
        newLayer.LineStyle = layerObj.LineStyle
        newLayer.LayerGroupSerial = layerObj.LayerGroupSerial
        return newLayer

    def CreateLinestring(self, container, layer, polyseg, nameSuffix):
        ls = container.Add(self.lType)
        ls.Layer = layer.SerialNumber 
        ls.Name = layer.Name + nameSuffix
        try:
            ls.Append(polyseg, None, False, False)
            return ls
        except:
            # some objects with funny UCS throw when trying to append to
            # linestring
            container.Remove(ls.SerialNumber)
            return None

    def OkClicked(self, sender, e):
        layerCollection = self.currentProject[Project.FixedSerial.LayerContainer]
        if self.alignmentCtl.SelectedSerial == 0:
            self.alignmentCtl.StatusMessage = "An alignment is required"
            return False
        self.alignmentCtl.StatusMessage = ""

        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)

        with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:
            self.alignment = self.alignmentCtl.SelectedEntity()
            alignPolyseg = self.alignment.ComputePolySeg()
            seg = clr.Reference[Segment]()
            t = clr.Reference[float]()
            staRef = clr.Reference[float]()
            offsetRef = clr.Reference[float]()
            pcl = clr.Reference[Point3D]()
            v = clr.Reference[Vector3D]()
            side = clr.Reference[Side]()
            staLocationsLeft = {}
            staLocationsRight = {}
            foundLayer = False
            for o in self.objs.SelectedMembers(self.currentProject):
                if not self.IsValidObj(o):
                    continue
                polySeg = o.ComputePolySeg()
                if not polySeg:
                    continue
                polyVert=o.ComputeVerticalPolySeg()
                if not foundLayer:
                    layerSerial = o.Layer
                    container = o.GetSite()
                    foundLayer = True
                for node in polySeg.Nodes():
                    coord = node.Point
                    if coord.Is2D and polyVert:
                         (found, segment, p, isOn) = polyVert.LocateX(Point3D(node.Station, 0.0))
                         if found:
                            coord.Z = p.Y
                    rtn = alignPolyseg.FindPointFromPoint(coord, seg, t, pcl, staRef, v,offsetRef, side)
                    if not rtn:
                        continue
                    sta = round(staRef.Value, 0)
                    offset = offsetRef.Value
                    if side.Value == Side.Left:
                        offset = -offset
                    if sta in staLocationsLeft:
                        (sn, c, off) = staLocationsLeft[sta]
                        if offset < off:
                            staLocationsLeft[sta] = (o.SerialNumber, coord, offset)
                    else:
                        staLocationsLeft.Add(sta, (o.SerialNumber, coord, offset))
                    if sta in staLocationsRight:
                        (sn, c, off) = staLocationsRight[sta]
                        if offset > off:
                            staLocationsRight[sta] = (o.SerialNumber, coord, offset)
                    else:
                        staLocationsRight.Add(sta, (o.SerialNumber, coord, offset))

            if foundLayer:
                ptsListLeft = List[Point3D]() 
                leftStart = 0
                leftEnd = 0
                for key, (sn, coord, offset) in sorted(staLocationsLeft.items()):
                    if not leftStart:
                        leftStart = sn
                    leftEnd = sn
                    ptsListLeft.Add(coord)
                
                polySegLeft = PolySeg()
                polySegLeft.Add(ptsListLeft)
                if self.edgeLinesCB.IsChecked:
                    l = self.UpdateLayer(layerSerial, layerCollection, self.edgeLayerMaskCtl.Text)
                    self.CreateLinestring(container, l, polySegLeft, "-L"+ self.nameSuffix.Text)

                ptsListRight = List[Point3D]()   
                rightStart = 0
                rightEnd = 0
                for key, (sn, coord, offset) in sorted(staLocationsRight.items()):
                    if not rightStart:
                        rightStart = sn
                    rightEnd = sn
                    ptsListRight.Add(coord)
         
                polySegRight = PolySeg()
                polySegRight.Add(ptsListRight.ToArray())
                if self.edgeLinesCB.IsChecked:
                    l = self.UpdateLayer(layerSerial, layerCollection, self.edgeLayerMaskCtl.Text)
                    self.CreateLinestring(container, l, polySegRight, "-R"+ self.nameSuffix.Text)

                if self.bndyCB.IsChecked and leftStart == rightStart and leftEnd == rightEnd:
                    polysegs = List[PolySeg]()  
                    polysegs.Add(polySegLeft)
                    polysegs.Add(polySegRight)
                    lStart = self.currentProject.Concordance.Lookup(leftStart)
                    lEnd = self.currentProject.Concordance.Lookup(leftEnd)
                    if lStart and lEnd:
                        polysegs.Add(lStart.ComputePolySeg())
                        polysegs.Add(lEnd.ComputePolySeg())
                        mergeCount = PolySeg.JoinTouchingPolysegs(polysegs)
                        if mergeCount == 3:
                            # find polyseg left over from merge
                            for p in polysegs:
                                if p:
                                    break
                            l = self.UpdateLayer(layerSerial, layerCollection, self.bndyLayerMaskCtl.Text)
                            self.CreateLinestring(container, l, p, self.nameSuffix.Text)
            failGuard.Commit()

        self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)

        OptionsManager.SetValue("TrackLineEdge:edgeLinesCB", self.edgeLinesCB.IsChecked)
        OptionsManager.SetValue("TrackLineEdge:edgeLayerMaskCtl", self.edgeLayerMaskCtl.Text)
        OptionsManager.SetValue("TrackLineEdge:bndyCB", self.bndyCB.IsChecked)
        OptionsManager.SetValue("TrackLineEdge:bndyLayerMaskCtl", self.bndyLayerMaskCtl.Text)
        GlobalSelection.Clear()
        Keyboard.Focus(self.objs)